/*******************************************************************************
* FILE NAME: ifilstrm.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*     IFileStream                                                              *                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <ifilstrm.hpp>
#include <iexcbase.hpp>
#include <iexcept.hpp>
#include <icconsta.h>

#include <stdio.h>
#include <fstream.h>
#include <io.h>
#include <errno.h>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IFILSTRM_CPP_
  #include <ipagetun.h>
#endif



/*------------------------------------------------------------------------------
| IFileStream::IFileStream                                                     |
|                                                                              |
| Constructor from a file name.                                                |
------------------------------------------------------------------------------*/
IFileStream::IFileStream(const char *fileName, IBase::Boolean writeable)
{
    fName = fileName;
    fWriteable = writeable;
    int omode = (writeable ? (ios::in | ios::out) : ios::in) | ios::bin;
    if (! fBuffer.open(fileName, omode)) {
        ITHROWLIBRARYERROR(IC_STREAM_FILE_OPEN_ERROR,
                      IBaseErrorInfo::accessError,
                      IException::recoverable);
    }
}

/*------------------------------------------------------------------------------
| IFileStream::~IFileStream                                                    |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IFileStream::~IFileStream()
{
    fBuffer.close();
}

/*------------------------------------------------------------------------------
| IFileStream::flush                                                           |
|                                                                              |
| Flush the buffer.                                                            |
------------------------------------------------------------------------------*/
IFileStream&
IFileStream::flush()
{
   fBuffer.sync();
   return *this;
}

/*------------------------------------------------------------------------------
| IFileStream::seek                                                            |
|                                                                              |
| Seek to a new position in the file if within the limits of the file,         |
| otherwise throw an exception.                                                |
------------------------------------------------------------------------------*/
IFileStream&
IFileStream::seek (IBaseStream::Position position)
{
    //** Check if the new position is within the file.
    if (position > logicalEndOfStream())
    {
        setLogicalEndOfStream(position);
    }
    else {
        if (fBuffer.seekoff(position, ios::beg, 0) == EOF) {
           ITHROWLIBRARYERROR(IC_STREAM_FILE_SEEK_POS_INVALID,
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);
        }
    }
    return *this;
}

/*------------------------------------------------------------------------------
| IFileStream::position                                                        |
|                                                                              |
| Return the current read/write pointer for the file.                          |
------------------------------------------------------------------------------*/
IBaseStream::Position
IFileStream::position () const
{
    filebuf*    fbtmp = (filebuf*) /*NonConst*/ (&fBuffer);
    return fbtmp->seekoff(0, ios::cur, 0);
}

/*------------------------------------------------------------------------------
| IFileStream::isWriteable                                                     |
|                                                                              |
| Return whether the file can be written to.                                   |
------------------------------------------------------------------------------*/
IBase::Boolean
IFileStream::isWriteable () const
{
    return fWriteable;
}

/*------------------------------------------------------------------------------
| IFileStream::setLogicalEndOfStream                                           |
|                                                                              |
| Set the logical size of the file.                                            |
------------------------------------------------------------------------------*/
IFileStream&
IFileStream::setLogicalEndOfStream (IBaseStream::Position endOfStream)
{
    setPhysicalEndOfStream(endOfStream);
    return *this;
}

/*------------------------------------------------------------------------------
| IFileStream::logicalEndOfStream                                              |
|                                                                              |
| Return the logical size of the file.                                         |
------------------------------------------------------------------------------*/
IBaseStream::Position
IFileStream::logicalEndOfStream () const
{
    return physicalEndOfStream();
}

/*------------------------------------------------------------------------------
| IFileStream::setPhysicalEndOfStream                                          |
|                                                                              |
| Set the physical size of the file.                                           |
------------------------------------------------------------------------------*/
IFileStream&
IFileStream::setPhysicalEndOfStream(IBaseStream::Position endOfStream)
{
        int fd = fBuffer.detach();

    if (fd == -1) {
        // file not opened
        return *this;
    }

    int errstatus = 0;

    if (_chsize(fd, endOfStream) != 0) {
        errstatus = errno;
    }

    if (fBuffer.attach(fd) == 0) {
       ITHROWLIBRARYERROR(IC_STREAM_FILE_OPEN_ERROR,
                      IBaseErrorInfo::accessError,
                      IException::recoverable);
    }

    fBuffer.seekoff(0, ios::end, 0);


    switch (errstatus) {
        case 0:       return *this;
        case ENOSPC:
           ITHROWLIBRARYERROR(IC_STREAM_FILE_TOO_SMALL,
                      IBaseErrorInfo::accessError,
                      IException::recoverable);
        case EACCESS:
        case EBADF:
        default:
           ITHROWLIBRARYERROR(IC_STREAM_FILE_OPEN_ERROR,
                      IBaseErrorInfo::accessError,
                      IException::recoverable);

    }

    return *this;
}

/*------------------------------------------------------------------------------
| IFileStream::physicalEndOfStream                                             |
|                                                                              |
| Return the physical size of the file.                                        |
------------------------------------------------------------------------------*/
IBaseStream::Position
IFileStream::physicalEndOfStream () const
{
    filebuf*    fbtmp = (filebuf*) /*NonConst*/ (&fBuffer);
    streampos origPos = fbtmp->seekoff(0, ios::cur, 0);
    IBaseStream::Position endPos = fbtmp->seekoff(0, ios::end, 0);
    fbtmp->seekoff(origPos, ios::beg, 0);
    return endPos;
}

/*------------------------------------------------------------------------------
| IFileStream::handleWriteBufferFull                                           |
|                                                                              |
| Write the data to disk.                                                      |
------------------------------------------------------------------------------*/
void
IFileStream::handleWriteBufferFull (const void* source, size_t sourceSize)
{
    if (fBuffer.sputn((char *)source, (int)sourceSize) != sourceSize) {
       ITHROWLIBRARYERROR(IC_STREAM_FILE_WRITE_ERROR,
                      IBaseErrorInfo::accessError,
                      IException::recoverable);
    }
}

/*------------------------------------------------------------------------------
| IFileStream::handleReadBufferEmpty                                           |
|                                                                              |
| Read the data from disk.                                                     |
------------------------------------------------------------------------------*/
void
IFileStream::handleReadBufferEmpty (void* destination, size_t destinationSize)
{
    if (fBuffer.sgetn((char *)destination, (int)destinationSize) != destinationSize) {
       ITHROWLIBRARYERROR(IC_STREAM_FILE_INSUFFICIENT_DATA,
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
    }
}
