/*******************************************************************************
* FILE NAME: istring1.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the testing functions of          *
*   the class IString.                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #include <string.h>
  }

#include <istring.hpp>

#include <ibuffer.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISTRING1_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IString::isAlnum                                                             |
| IString::isAlpha                                                             |
| IString::isAscii                                                             |
| IString::isCntrl                                                             |
| IString::isDigits                                                            |
| IString::isHexDigits                                                         |
| IString::isLowerCase                                                         |
| IString::isPrintable                                                         |
| IString::isPunctuation                                                       |
| IString::isUpperCase                                                         |
| IString::isWhiteSpace                                                        |
|                                                                              |
| Each of these functions is delegated to the associated IBuffer holding the   |
| contents of the receiving IString.                                           |
|                                                                              |
| Notes:                                                                       |
|   1. The default (SBCS) implementation is to utilize the                     |
|      corresponding standard C library function to test the string            |
|      contents.                                                               |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::isBinaryDigits                                                      |
|                                                                              |
| Examine each character of the receiver in turn, returning false if a non '0' |
| or '1' is found.  Otherwise, return true.                                    |
|                                                                              |
| The test for '0'/'1' is made by exclusive-or-ing with '1'; this yields all   |
| the bits that differ from '1'.  If any but the least significant vary, then  |
| the test fails.                                                              |
|                                                                              |
| Notes:    (Why ?) - it should be false (defect 5935 in icl2)                                                                  |
|   1. If the string is null (of length 0) then 1 (true) is returned.          |
------------------------------------------------------------------------------*/
IBase::Boolean IString :: isBinaryDigits ( ) const
  {
  const char *p   = this->data();
  unsigned    pos = this->length();

  if (!pos)  return false;      // return false for null string

  while ( pos-- )
    if ( ( *p++ ^ '1' ) > 1 )
      return false;
  return true;
  }

/*------------------------------------------------------------------------------
| IString::isLike                                                              |
|                                                                              |
| This function is implemented by invoking the protected version that accepts  |
| a char* pointer and length.                                                  |
------------------------------------------------------------------------------*/
IBase::Boolean IString :: isLike ( const char *pPattern,
                                   unsigned    patternLen,
                                   char        zeroOrMore,
                                   char        anyChar ) const
  {
  // Answer true if the receiver and pattern are equal.
  if ( this->buffer()->compare( pPattern, patternLen ) == IBuffer::equal )
    return true;
  // Index into strings:
  unsigned iText    = 0,
           iPattern = 0,
           lastStar = 0,
           len      = this->length();
  // Continue till we get to the end of the pattern...
  while ( iPattern < patternLen )
    {
    char p = pPattern[ iPattern++ ];
    // See if the pattern character matches "zeroOrMore" characters...
    if ( p == zeroOrMore )
      // We have reached a * in the pattern.
      // If at the end, then the receiver has matched it (since
      // the last * matches the rest of the text).
      // Otherwise, the scan now starts here (just beyond the *)
      // since the previous portion of the pattern has been
      // matched.
      if ( iPattern >= patternLen )
        return true;
      else
        lastStar = iPattern;
    else
      // Not a *.  If we are at the end of the text, then the
      // result is false (since there is no character to match
      // the next pattern character).
      if ( iText >= len )
        return false;
      else
        {
        // Compare text/pattern characters...
        char t = this->data()[ iText++ ];
        if ( p == anyChar || p == t )
          // Characters match.  If we had a * somewhere, are at
          // the end of the pattern, and not at the end of the
          // text, then we need to restart the scan from the last
          // * while using it to swallow one more character of
          // the text.
          //
          // This situation arises in cases like the following:
          //     pattern = *abc
          //     text    = 123abcabc
          //
          // If these conditions aren't met, then we simply go
          // back to the top of the while loop and process the
          // next pattern character.
          if ( lastStar > 0
               &&
               iPattern >= patternLen
               &&
               iText < len )
            ;
          else
            continue;
        else
          // Characters don't match.  If we haven't seen a
          // "zeroOrMore" character in the pattern, then we've
          // failed to match it.
          if ( lastStar == 0 )
            return false;
        // We get here if we need to "reset" the scan to the
        // last "zeroOrMore" wildcard in the pattern.  First,
        // calculate how many characters were matched after the
        // last such wildcard in the pattern:
        unsigned matched = iPattern - lastStar - 1;
        // Reset pattern position to just past it:
        iPattern = lastStar;
        // Back up in the text stream to rescan all but
        // the first matched character:
        iText -= matched;
        }
    }
  // At end of pattern.  If also at the end of the text, then they match,
  // otherwise, they don't:
  return iText >= len;
  }

/*------------------------------------------------------------------------------
| IString::isAbbrevFor                                                         |
|                                                                              |
| This function is implemented via the protected member of the same name.      |
| That function accepts as arguments a full specification of the "full"        |
| string (pointer *and* length) and the minimimum abbreviation length.  It     |
| determines whether the receiver is an acceptable abbreviation by comparing   |
| the appreviation to beginning of the full string and if equal, comparing the |
| length of the abbreviation to the minimum.                                   |
------------------------------------------------------------------------------*/
IBase::Boolean IString :: isAbbrevFor ( const char *pFullString,
                                        unsigned    fullStringLen,
                                        unsigned    minLen ) const
  {
  if ( minLen == 0 )
    minLen = this->length();
  if ( this->length() <= fullStringLen
       &&
       this->length() >= minLen
       &&
       this->buffer()->compare(pFullString, this->length()) == IBuffer::equal )
    return true;
  else
    return false;
  }

/*------------------------------------------------------------------------------
| IString::includes                                                            |
|                                                                              |
| Simply invoke the equivalent indexOf function.                               |
|                                                                              |
| Notes:                                                                       |
|   1. These funcions are essentially just shorthand for the corresponding     |
|      indexOf functions.  It provides a more meaningful function name         |
|      in cases where invoked to obtain a Boolean result (rather than the      |
|      specific index within the string).                                      |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::isInternationalized()                                               |
|                                                                              |
| Check if Internationalization sementics  (I18N) are ON                       |
|                                                                              |
------------------------------------------------------------------------------*/
// See istring.inl


/*------------------------------------------------------------------------------
| IString::enableInternationalization(Boolean enable = true)                   |
|                                                                              |
| Turn on Internationalization (I18N) semantics                                |
|                                                                              |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::disableInternationalization()                                       |
|                                                                              |
| Turn off Internationalization (I18N) semantics                               |
|                                                                              |
------------------------------------------------------------------------------*/
// See istring.inl

