/*******************************************************************************
* FILE NAME: istring5.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the "accessing" functions for     *
*   the class IString.                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <istring.hpp>

#include <istream.hpp>

#include <ibuffer.hpp>

#include <iexcbase.hpp>

#include <icconsta.h>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISTRING5_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IString::length                                                              |
|                                                                              |
| Simply return the length field from the associated IBuffer.                  |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::size                                                                |
|                                                                              |
| This function is simply a synonym for IString::length.                       |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::subString(unsigned startPos) const                                  |
|                                                                              |
| - extract string from 'startPos' to the end of string.                       |
|                                                                              |
| Notes: In MBCS environment, if 'startPos' is not pointing to the beginning   |
|        of the character, the partial MBCS character is replaced by blank     |
|        character(s) until the next valid character or end of string.         |
|                                                                              |
------------------------------------------------------------------------------*/
IString IString :: subString ( unsigned startPos ) const
{
  if ( startPos <= this->length() ) {
#ifdef IC_NLS
    if (IString::isInternationalized()) {
      return IString( this->buffer()->subString( startPos ) );
    } else
#endif
      return this->subString( startPos, this->length() - startPos + 1 );
  } else {
    return IString::null;
  }
}

/*------------------------------------------------------------------------------
| IString::subString(unsigned startPos, unsigned len, char padCharacter) const |
|                                                                              |
| - extract 'len' bytes from the original string starting from 'startPos'.     |
|   If 'len' is greater than the length of 'startPos' to the end of string,    |
|   'padCharacter' will be used to pad the rest of the resulting string.       |
|                                                                              |
| Notes: In MBCS environment, if 'startPos' is not pointing to the beginning   |
|        of the character, the partial MBCS character is replaced by           |
|        'padCharacter' until the next valid character or end of string.       |
|        The resulting string is then filled with characters from the original |
|        extracted 'len' bytes until 'len' bytes are filled. If there is not   |
|        enough room to contain "full" character at the end (in EBCDIC,        |
|        shift-out and shift-in characters might have been inserted to         |
|        maintain the resuting string as a valid string) or there are still    |
|        spaces left after all the characters are filled from the original     |
|        extracted 'len' bytes (in EBCDIC, the original string might have      |
|        redundant shift-out/shift-in characters), 'padCharacter' will be used |
|        for padding.                                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IString IString :: subString ( unsigned startPos,
                               unsigned len,
                               char     padCharacter ) const
  {
  IString result( this->buffer()->subString( startPos,
                                            len,
                                            padCharacter ) );
  return result;
  }

/*------------------------------------------------------------------------------
| IString::operator []                                                         |
|                                                                              |
| The const and non-const version of these functions differ in the way         |
| indexing off the end of the string is handled.  For the const case, an       |
| Invalid request exception is thrown. In the non-const case, the string is    |
| extended with blanks (and a reference to the last blank added is returned).  |
|                                                                              |
| Otherwise, return the specified element of the receiver's buffer.            |
|                                                                              |
| Notes:                                                                       |
|   1. The indexing is 1-based.  For convenience, index zero                   |
|      is treated like index 1 (this is consistent with the treatment          |
|      of index zero by other member functions; e.g., for searching).          |
|   2. This function presents some problems for the integrity of IString       |
|      contents (as does operator char*).  While some precautions are          |
|      taken (ensuring the receiver has the only reference to its              |
|      buffer) subsequent assignment of the IString and manipulation via       |
|      the char& returned by this function may cause unexpected result.        |
------------------------------------------------------------------------------*/
const char& IString :: operator [] ( unsigned index ) const
   {
   index -= ( index > 0 ); // Get 0-based index.
   if ( index >= this->length() )
      {
      IInvalidRequest
        exception(IBase::messageText(IC_ISTRING_INDEX_ERROR),
                  IC_ISTRING_INDEX_ERROR,
                  IException::recoverable );
      ITHROW( exception );
      }
   return this->data()[ index ];
   }

/*------------------------------------------------------------------------------
| IString::operator []                                                         |
------------------------------------------------------------------------------*/
char& IString :: operator [] ( unsigned index )
   {
   IString old( this->buffer() );
   index -= ( index > 0 ); // Get 0-based index.
   // Test whether buffer has to grow:
   if ( index >= old.length() )
     { // It does,
     this->initBuffer( old.data(), old.length(),
                      0, index - old.length() + 1,
                      0, 0,
                      ' ' );
     }
   else
     { // It doesn't, but make sure we've got our own copy.
     if ( this->buffer()->useCount() == 1 )
       { // Keep it, instead.
       this->buffer() -> addRef();
       }
     else
       { // Need to make copy.
       this->initBuffer( old.data(), old.length(),
                         0, 0,
                         0, 0,
                         0 );
       }
     }
   return this->data()[ index ];
   }
const char& IString :: operator [] ( signed index ) const
   {
   if (index < 0)
      {
      IInvalidRequest
        exception(IBase::messageText(IC_ISTRING_INDEX_ERROR),
                  IC_ISTRING_INDEX_ERROR,
                  IException::recoverable );
      ITHROW( exception );
      }

   index -= ( index > 0 ); // Get 0-based index.
   if ( index >= this->length() )
      {
      IInvalidRequest
        exception(IBase::messageText(IC_ISTRING_INDEX_ERROR),
                  IC_ISTRING_INDEX_ERROR,
                  IException::recoverable );
      ITHROW( exception );
      }
   return this->data()[ index ];
   }

char& IString :: operator [] ( signed index )
   {
   if (index < 0)
      {
      IInvalidRequest
        exception(IBase::messageText(IC_ISTRING_INDEX_ERROR),
                  IC_ISTRING_INDEX_ERROR,
                  IException::recoverable );
      ITHROW( exception );
      }

   IString old( this->buffer() );
   index -= ( index > 0 ); // Get 0-based index.
   // Test whether buffer has to grow:
   if ( index >= old.length() )
     { // It does,
     this->initBuffer( old.data(), old.length(),
                      0, index - old.length() + 1,
                      0, 0,
                      ' ' );
     }
   else
     { // It doesn't, but make sure we've got our own copy.
     if ( this->buffer()->useCount() == 1 )
       { // Keep it, instead.
       this->buffer() -> addRef();
       }
     else
       { // Need to make copy.
       this->initBuffer( old.data(), old.length(),
                         0, 0,
                         0, 0,
                         0 );
       }
     }
   return this->data()[ index ];
   }


