/*******************************************************************************
* FILE NAME: igrabhdl.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the classes:                                             *
*     IMGrabbable                                                              *
*     IGrabHandler                                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include "igrabhdl.hpp"

#include <iwindow.hpp> 
#include <ihandle.hpp> 
#include <imphdr.hpp>
#include <imousevt.hpp>
#include <igrect.hpp>
#include <igrafctx.hpp>

#include <windows.h>
#include <assert.h>

static const unsigned kHandleSize = 4;

/*******************************************************************************
*                                                                              *
*   Class IHandlerDeleter                                                      *
*                                                                              *
*******************************************************************************/

class IHandlerDeleter {
public:
                            IHandlerDeleter() : fpHandler( NULL ) {}
                            ~IHandlerDeleter();

    void                    operator=( IGrabHandler* hdlr );

private:
    IGrabHandler*           fpHandler;
};

void
IHandlerDeleter::operator=( IGrabHandler* hdlr )
{
    delete fpHandler;
    fpHandler = hdlr;
}

IHandlerDeleter::~IHandlerDeleter()
{
    delete fpHandler;
}


/*******************************************************************************
*                                                                              *
*   Class IGrabHandler                                                         *
*                                                                              *
*   Handler for moving and sizing an object in a window.                       *
*                                                                              *
*******************************************************************************/

class IGrabHandler : public IMousePointerHandler
{
public:
                            IGrabHandler( IMGrabbable& grabee,
                                          IGrabHandles& handles,
                                          Boolean dragMove );
    virtual                 ~IGrabHandler();
    
    virtual void            mouseDown( const IPoint& position );
    virtual Boolean         isMouseDown() const;

    static  void            deleteHandler( IGrabHandler* );
    
protected:
    virtual Boolean         mouseClicked( IMouseClickEvent& mouseClickEvent );
    virtual Boolean         mouseMoved( IMouseEvent& mouseEvent );
    virtual Boolean         mousePointerChange( IMousePointerEvent& mousePointerEvent );
    

private:
    void                    setStatus( const IPoint& point );
    virtual void            changeTo( const IRectangle& newRect );
    void                    xorOutline();
    
    IMGrabbable&            fGrabbedObject;
    IWindow*                fpObjectWindow;
    IGrabHandles*           fpGrabHandles;
    IPointerHandle          fhCursor;
    Boolean                 fbDragMove;

    // moving and sizing
    IPoint                  fStartMouse;
    IRectangle              frectStart;
    Boolean                 fbMouseDown;
    Boolean                 fbSizing;
    Boolean                 fbMoving;

    // the current state
    Boolean                 fbForUs;
    Boolean                 fbOverAHandle;
    Boolean                 fbLeft;
    Boolean                 fbTop;
    Boolean                 fbRight;
    Boolean                 fbBottom;
    Boolean                 fbOkToDelete;

    static IHandlerDeleter  fgHandlerDeleter;

friend class IGrabHandles;   
};


IHandlerDeleter IGrabHandler::fgHandlerDeleter;

IGrabHandler::IGrabHandler( IMGrabbable & grabee,
                            IGrabHandles& handles,
                            Boolean dragMove ) :
              IMousePointerHandler(),
              fGrabbedObject( grabee ),
              fpObjectWindow( &grabee.inWindow() ),
              fpGrabHandles( &handles ),
              fhCursor( 0 ),
              fbDragMove( dragMove ),
              fStartMouse( 0, 0 ),
              frectStart( 0, 0, 0, 0 ),
              fbForUs( false ),
              fbMouseDown( false ),
              fbSizing( false ),
              fbMoving( false ),
              fbOverAHandle( false ),
              fbLeft( false ),
              fbTop( false ),
              fbRight( false ),
              fbBottom( false ),
              fbOkToDelete( true )
{
    handleEventsFor( fpObjectWindow );
}

IGrabHandler::~IGrabHandler()
{
    stopHandlingEventsFor( fpObjectWindow );
}

void
IGrabHandler::changeTo( const IRectangle & newRect )
{
    // change our presentation
    fpGrabHandles->changeTo( newRect );
}

void
IGrabHandler::mouseDown( const IPoint& position )
{
    fbMouseDown = true;
    fbMoving = false;
    fbSizing = false;

    setStatus( position );
    assert( fbForUs );

    fbSizing = fbOverAHandle;
    fbMoving = !fbOverAHandle;
    fStartMouse = position;
    frectStart = fpGrabHandles->getOutline();

}

Boolean
IGrabHandler::isMouseDown() const
{
    return fbMouseDown;
}

void
IGrabHandler::deleteHandler( IGrabHandler* pHandler )
{
    if ( pHandler->fbOkToDelete )
        delete pHandler;
    else
        fgHandlerDeleter = pHandler;
}

void
IGrabHandler::setStatus( const IPoint& point )
{
    // see which handle we are over (if any)
    fbOverAHandle = true;
    fbLeft = false;
    fbTop = false;
    fbRight = false;
    fbBottom = false;
    fbForUs = true;

    IGrabHandles::ELocation where = fpGrabHandles->getLocation( point );
    
    switch( where ) {
    case IGrabHandles::kInside:
#ifdef ORIGIN_BOTTOM_LEFT
        fhCursor =  ISystemPointerHandle( ISystemPointerHandle::move );
#else        
        fhCursor = IPointerHandle( (HCURSOR)(void*)LoadCursor( NULL,
                                                               IDC_SIZEALL ) );
#endif        
        fbOverAHandle = false;
        break;
    case IGrabHandles::kOutside:
        fhCursor =  ISystemPointerHandle( ISystemPointerHandle::arrow );
        fbOverAHandle = false;
        fbForUs = false;
        break;
    case IGrabHandles::kTopLeft:
        fhCursor =  ISystemPointerHandle( ISystemPointerHandle::sizeNWSE );
        fbLeft = true;
        fbTop = true;
        break;
    case IGrabHandles::kTopMiddle:
        fhCursor =  ISystemPointerHandle( ISystemPointerHandle::sizeVertical );
        fbTop = true;
        break;
    case IGrabHandles::kTopRight:
        fhCursor =  ISystemPointerHandle( ISystemPointerHandle::sizeNESW );
        fbTop = true;
        fbRight = true;
        break;
    case IGrabHandles::kMiddleLeft:
        fhCursor =  ISystemPointerHandle( ISystemPointerHandle::sizeHorizontal );
        fbLeft = true;
        break;
    case IGrabHandles::kMiddleRight:
        fhCursor =  ISystemPointerHandle( ISystemPointerHandle::sizeHorizontal );
        fbRight = true;
        break;
    case IGrabHandles::kBottomLeft:
        fhCursor =  ISystemPointerHandle( ISystemPointerHandle::sizeNESW );
        fbLeft = true;
        fbBottom = true;
        break;
    case IGrabHandles::kBottomMiddle:
        fhCursor =  ISystemPointerHandle( ISystemPointerHandle::sizeVertical );
        fbBottom = true;
        break;
    case IGrabHandles::kBottomRight:
        fhCursor =  ISystemPointerHandle( ISystemPointerHandle::sizeNWSE );
        fbRight = true;
        fbBottom = true;
        break;
    }
}

Boolean
IGrabHandler::mouseClicked( IMouseClickEvent& mouseClickEvent )
{
    Boolean result = false;

    // reset flags to false no matter what
    fbMoving = false;
    fbSizing = false;

    setStatus( mouseClickEvent.mousePosition() );

    // see if this event is for us (set on last mouse move)
    if ( fbForUs )
    {
        // we are only interested in button 1 events
        if ( mouseClickEvent.mouseButton() == IMouseClickEvent::button1 )
        {
            // see which action they are doing
            switch( mouseClickEvent.mouseAction() ) {
            case IMouseClickEvent::down:
              {
                result = true;
                fbMouseDown = true;
                fbSizing = fbOverAHandle;
                fbMoving = !fbOverAHandle;
                fStartMouse = mouseClickEvent.mousePosition();
                frectStart = fpGrabHandles->getOutline();
                xorOutline();
                break;
              }
            case IMouseClickEvent::up:
              {
                if ( fbMouseDown )
                {
                    result = true;
                    xorOutline();
                    fbMouseDown = false;
                    // signal the object that moved or sized
                    fGrabbedObject.setRect( fpGrabHandles->getOutline() );
                }
                break;
              }
            }

        }
    }

    return result;
}

Boolean
IGrabHandler::mouseMoved( IMouseEvent& mouseEvent )
{
    Boolean result = false;

    // compute the delta change from the original mouse position
    IPoint delta( mouseEvent.mousePosition() - fStartMouse );

    // were we moving or sizing?
    if ( fbMoving )
    {
        if ( fbDragMove )
        {
            xorOutline();
            
            fbOkToDelete = false; // delay our deletion
            Boolean deleted = fGrabbedObject.doDragDrop( fStartMouse );
            if ( !deleted ) {
                fbOkToDelete = true;
                fpGrabHandles->fpHandler = this;
                changeTo( fGrabbedObject.rect() );
                fbMoving = false;
                fbMouseDown = false;
            }
        }
        else
        {
            IRectangle rect( frectStart );
            xorOutline();
            changeTo( rect.moveBy( delta ) );
            xorOutline();
        }
        return true;

    }
    else if ( fbSizing )
    {
        int l = frectStart.minX();
        int r = frectStart.maxX();
        
        // following is sensitive to the coord system
#ifdef ORIGIN_BOTTOM_LEFT
        int t = frectStart.maxY();
        int b = frectStart.minY();
#else
        int t = frectStart.minY();
        int b = frectStart.maxY();
#endif

        if ( fbLeft )
            l += delta.x();
        if ( fbTop )
            t += delta.y();
        if ( fbRight )
            r += delta.x();
        if ( fbBottom )
            b += delta.y();
     
        xorOutline();

        // the following is sensitive to the coord system
        // (l, b), (r, t) for 0,0 bottom left
        // (l, t), (r, b) for 0,0 top left
#ifdef ORIGIN_BOTTOM_LEFT
        changeTo( IRectangle( IPoint(l, b), IPoint(r, t) ) );
#else
        changeTo( IRectangle( IPoint(l, t), IPoint(r, b) ) );
#endif
        xorOutline();
        return true;
    }
    else
    {
        setStatus( mouseEvent.mousePosition() );
        result = fbForUs;
    }
     
    return result;
}

Boolean
IGrabHandler::mousePointerChange( IMousePointerEvent& mousePointerEvent )
{
    Boolean result = false;

    // see if this event is for us (set on last mouse move)
    if ( fbForUs )
    {
        result = true;
        mousePointerEvent.setMousePointer(fhCursor);
    }

    return result;
}

void
IGrabHandler::xorOutline()
{
    IPresSpaceHandle ps = fpObjectWindow->presSpace();
    fpGrabHandles->draw( ps, false );
}


/*******************************************************************************
*                                                                              *
*   Class IMGrabbable                                                          *
*                                                                              *
*******************************************************************************/

Boolean
IMGrabbable::doDragDrop( const IPoint& )
{
    // no-op by default
    return false;
}


/*******************************************************************************
*                                                                              *
*   Class IGrabHandles                                                         *
*                                                                              *
*   A class to draw and test grab handles                                      *
*                                                                              *
*******************************************************************************/

IGrabHandles::IGrabHandles( IMGrabbable& grabee, Boolean dragMove ) :
              fpHandler( new IGrabHandler( grabee, *this, dragMove ) ),
              fOutline( grabee.rect() )
{
    moveHandles();

    IPresSpaceHandle ps = grabee.inWindow().presSpace();
    draw( ps );
}

IGrabHandles::~IGrabHandles()
{
    fpHandler->stopHandlingEventsFor( fpHandler->fpObjectWindow );
    IGrabHandler::deleteHandler( fpHandler );
}

void
IGrabHandles::changeTo( const IRectangle& newRect )
{
    fOutline = newRect;
    moveHandles();
}

void
IGrabHandles::moveTo( const IPoint& newPoint )
{
    fOutline.moveTo(newPoint);
    moveHandles();
}

IRectangle&
IGrabHandles::getOutline() const
{
    return (IRectangle&)fOutline;
}

void
IGrabHandles::moveHandles()
{
    IPoint tp1, tp2;
    tp1 = fOutline.minXMaxY() + IPoint(0, -kHandleSize);
    tp2 = fOutline.minXMaxY() + IPoint(kHandleSize, 0);
#ifdef ORIGIN_BOTTOM_LEFT
    fTL = IRectangle( tp1, tp2 );
#else
    fBL = IRectangle( tp1, tp2 );
#endif
    tp1 = fOutline.centerXMaxY() + IPoint(-(kHandleSize / 2), -kHandleSize);
    tp2 = fOutline.centerXMaxY() + IPoint((kHandleSize / 2), 0);
#ifdef ORIGIN_BOTTOM_LEFT
    fTM = IRectangle( tp1, tp2 );
#else
    fBM = IRectangle( tp1, tp2 );
#endif
    tp1 = fOutline.maxXMaxY() + IPoint(-kHandleSize, -kHandleSize);
    tp2 = fOutline.maxXMaxY() + IPoint(0, 0);
#ifdef ORIGIN_BOTTOM_LEFT
    fTR = IRectangle( tp1, tp2 );
#else
    fBR = IRectangle( tp1, tp2 );
#endif
    tp1 = fOutline.minXCenterY() + IPoint(0, -(kHandleSize / 2));
    tp2 = fOutline.minXCenterY() + IPoint(kHandleSize, (kHandleSize / 2));
    fML = IRectangle( tp1, tp2 );
    tp1 = fOutline.maxXCenterY() + IPoint(-kHandleSize, -(kHandleSize / 2));
    tp2 = fOutline.maxXCenterY() + IPoint(0, (kHandleSize / 2));
    fMR = IRectangle( tp1, tp2 );
    tp1 = fOutline.minXMinY() + IPoint(0, 0);
    tp2 = fOutline.minXMinY() + IPoint(kHandleSize, kHandleSize);
#ifdef ORIGIN_BOTTOM_LEFT
    fBL = IRectangle( tp1, tp2 );
#else
    fTL = IRectangle( tp1, tp2 );
#endif
    tp1 = fOutline.centerXMinY() + IPoint(-(kHandleSize / 2), 0);
    tp2 = fOutline.centerXMinY() + IPoint((kHandleSize / 2), kHandleSize);
#ifdef ORIGIN_BOTTOM_LEFT
    fBM = IRectangle( tp1, tp2 );
#else
    fTM = IRectangle( tp1, tp2 );
#endif
    tp1 = fOutline.maxXMinY() + IPoint(-kHandleSize, 0);
    tp2 = fOutline.maxXMinY() + IPoint(0, kHandleSize);
#ifdef ORIGIN_BOTTOM_LEFT
    fBR = IRectangle( tp1, tp2 );
#else
    fTR = IRectangle( tp1, tp2 );
#endif
    
}

IGrabHandles::ELocation
IGrabHandles::getLocation( const IPoint& position )
{
    IGrabHandles::ELocation result = kOutside;

    Boolean insideOutline = fOutline.contains( position );
    if ( insideOutline )
    {
        result = kInside;
        if ( fTL.contains( position ) )
        {
            result = kTopLeft;
        }
        else if( fTM.contains( position ) )
        {
            result = kTopMiddle;
        }
        else if( fTR.contains( position ) )
        {
            result = kTopRight;
        }
        else if( fML.contains( position ) )
        {
            result = kMiddleLeft;
        }
        else if( fMR.contains( position ) )
        {
            result = kMiddleRight;
        }
        else if( fBL.contains( position ) )
        {
            result = kBottomLeft;
        }
        else if( fBM.contains( position ) )
        {
            result = kBottomMiddle;
        }
        else if( fBR.contains( position ) )
        {
            result = kBottomRight;
        }
    }               

    return result;
}

void
IGrabHandles::draw( IPresSpaceHandle& ps, Boolean bWithHandles ) const
{
    IGraphicContext gc( ps );
    gc.setPenWidth( 1 );

    if ( bWithHandles )
    {
        gc.setPenColor( IColor::black );
        gc.setFillColor( IColor::black );
        gc.setDrawOperation( IGraphicBundle::fillAndFrame );
        IGRectangle grtl( fTL );
        grtl.drawOn( gc );
        IGRectangle grtm( fTM );
        grtm.drawOn( gc );
        IGRectangle grtr( fTR );
        grtr.drawOn( gc );
        IGRectangle grml( fML );
        grml.drawOn( gc );
        IGRectangle grmr( fMR );
        grmr.drawOn( gc );
        IGRectangle grbl( fBL );
        grbl.drawOn( gc );
        IGRectangle grbm( fBM );
        grbm.drawOn( gc );
        IGRectangle grbr( fBR );
        grbr.drawOn( gc );
    }
    else
    {
        gc.setPenColor( IColor::white );
        gc.setPenType( IGraphicBundle::alternate );
        gc.setMixMode( IGraphicBundle::xor );
    }

    gc.setDrawOperation( IGraphicBundle::frame );
    IGRectangle border( fOutline );
    border.drawOn( gc );

    if ( bWithHandles && fpHandler->isMouseDown() )
        draw( ps, false );

}

void
IGrabHandles::mouseDown( const IPoint& position )
{
    fpHandler->mouseDown( position );
}

