/*******************************************************************************
* FILE NAME: icanvas.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icanvas.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512  (from iwindow.cpp)
#pragma priority( -2147482112 )

#ifndef _IBASE_
  #include <ibase.hpp>
#endif

extern "C" {
  #ifdef IC_PM
  #define INCL_WINDIALOGS           // DLGC_xxx
  #define INCL_WINERRORS            // PMERR_xxx
  #define INCL_WINFRAMEMGR          // WM_QUERYFOCUSCHAIN
  #define INCL_WINSTATICS           // SS_BKGNDRECT
  #define INCL_WINSYS               // PP_BACKGROUNDCOLOR
  #define INCL_WINWINDOWMGR         // QWL_xxx
  #endif
  #include <iwindefs.h>
}

#include <icanvas.hpp>
#include <icanvprv.hpp>
#include <icconst.h>
#include <icmnfun.hpp>
#include <icoordsy.hpp>
#include <icolor.hpp>
#include <icvhdr.hpp>
#include <iexcept.hpp>
#include <ifont.hpp>
#include <irecohdr.hpp>
#include <itrace.hpp>
#include <iwposbuf.hpp>
#include <iwcname.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _ICANVAS_CPP_
  #include <ipagetun.h>
#endif

#ifdef IC_WIN
extern unsigned long propagationInProgress;
#endif

/*------------------------------------------------------------------------------
| Public canvas styles.                                                        |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const ICanvas::Style ICanvas::classDefaultStyle = WS_VISIBLE;
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
ICanvas::Style ICanvas::currentDefaultStyle     = WS_VISIBLE;

/*------------------------------------------------------------------------------
| The static object that ICanvas and all of its derived classes use to obtain  |
| the common canvas handler.                                                   |
------------------------------------------------------------------------------*/
static ICanvasStatics CanvasStatics;
#pragma data_seg()

/*------------------------------------------------------------------------------
| ICanvasData::ICanvasData                                                     |
|                                                                              |
| Constructor for the private canvas data class.                               |
------------------------------------------------------------------------------*/
ICanvasData::ICanvasData ( )
  : ICanvasData::Inherited()
  , fShowWindow ( true )
  , fChildCount ( 0 )
{ }

/*------------------------------------------------------------------------------
| ICanvas::ICanvas                                                             |
|                                                                              |
| Constructor to create a canvas window.                                       |
| Notes: The canvas window is created as a WC_STATIC, SS_BGNDRECT window       |
|        (fill box control).                                                   |
------------------------------------------------------------------------------*/
ICanvas::ICanvas ( unsigned long windowIdentifier,
                   IWindow* parent,
                   IWindow* owner,
                   const IRectangle& initial,
                   const Style& style )
  : sizClLayout(), hwndClDefaultButton( 0 ),
    bClChildrenReversed( false ), bClDefaultHandlerAdded( false ),
#ifdef IC_WIN
    fFont( 0 ),
#endif
    fCanvasData( new ICanvasData() )
{
  // Save the extended style to make sure we have a copy of it stored.
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

  // Create the presentation system window.
  this->initialize( windowIdentifier,
                    parent,
                    owner,
                    initial,
                    this->convertToGUIStyle( style ),
                    this->convertToGUIStyle( style, true ) );

  /*****************************************************************/
  /* Implementation notes:                                         */
  /* - The WS_TABSTOP style is added in later by layout() only if  */
  /*   the canvas has a tab-able child window.  This window style  */
  /*   is needed to allow a frame or canvas parent to identify the */
  /*   canvas as a tab-able child window with WinEnumDlgItem       */
  /*   during Tab key processing (PM does this for a frame window, */
  /*   and the ICanvasHandler class does this for a parent         */
  /*   canvas), and for the ICanvasHandler to give the input focus */
  /*   to the appropriate child window during processing of a      */
  /*   WM_SETFOCUS message.                                        */
  /* - Like all other controls, the initial rectangle is only used */
  /*   if the canvas is not a frame control and not a child of     */
  /*   another canvas window (if it is a child of another canvas,  */
  /*   it will size itself to be able to hold all of its child     */
  /*   windows in calcMinimumSize()).                              */
  /*****************************************************************/
  CanvasStatics.canvasHandler().handleEventsFor( this );
  CanvasStatics.keyboardHandler().handleEventsFor( this );
                                            // Add private handler.
  bClDefaultHandlerAdded = true;            // Flag as added.

  // Add the default recoordination handler.  This handler is needed
  // for the canvas classes that do not do their own positioning of
  // child windows.   Those that do child window positioning need to
  // later remove the handler.
  IRecoordHandler::defaultHandler()->handleEventsFor( this );

  if ( IWindow::defaultOrdering() == IWindow::behindSiblings )
  {                                    // No need to reverse children.
     bClChildrenReversed = true;
  }
}

/*------------------------------------------------------------------------------
| ICanvas::ICanvas                                                             |
------------------------------------------------------------------------------*/
ICanvas::ICanvas ( )
  : sizClLayout(), hwndClDefaultButton( 0 ),
    bClChildrenReversed( false ), bClDefaultHandlerAdded( false ),
#ifdef IC_WIN
    fFont( 0 ),
#endif
    fCanvasData( new ICanvasData() )
{
  CanvasStatics.canvasHandler().handleEventsFor( this );
  CanvasStatics.keyboardHandler().handleEventsFor( this );
                                            // Add private handler.
  bClDefaultHandlerAdded = true;            // Flag as added.

  // Assume that derived classes that need the recoordination handler
  // will add one themselves (canvases that do not do their own
  // positioning of child windows).

  if ( IWindow::defaultOrdering() == IWindow::behindSiblings )
  {                                    // No need to reverse children.
     bClChildrenReversed = true;
  }
}

/*------------------------------------------------------------------------------
| ICanvas::initialize                                                          |
|                                                                              |
| Create the presentation system window.                                       |
------------------------------------------------------------------------------*/
ICanvas& ICanvas::initialize ( unsigned long     windowIdentifier,
                               IWindow*          parent,
                               IWindow*          owner,
                               const IRectangle& initialRect,
                               unsigned long     style,
                               unsigned long     extendedStyle )
{
  IASSERTPARM( parent != 0 );

  IWindowHandle canvasHandle =
                   this->create( windowIdentifier,
                                 0,
                                 style,
                                 WC_STATIC,
                                 parent->handle(),
                                 owner ?
                                   owner->handle() : IWindowHandle( 0 ),
                                 initialRect,
                                 0,
                                 0,
                                 IWindow::defaultOrdering(),
                                 extendedStyle );
  this->startHandlingEventsFor( canvasHandle );

  // Remove styles that were not specified by the caller, but which
  // get added via class styles.
  unsigned long
    currentStyle = this->style(),
    newStyle     = currentStyle;
  if ( ! ( style & IWindow::clipToParent.asUnsignedLong() ) )
  {           // Don't let canvas paint over its parent.
     newStyle &= ~IWindow::clipToParent.asUnsignedLong();
  }
  if ( ! ( style & IWindow::synchPaint.asUnsignedLong() ) )
  {           // Disable synchronous painting.
     newStyle &= ~IWindow::synchPaint.asUnsignedLong();
  }
  if ( newStyle != currentStyle )
  {
     this->setStyle( newStyle );
  }

  return *this;
}

/*------------------------------------------------------------------------------
| ICanvas::~ICanvas                                                            |
------------------------------------------------------------------------------*/
ICanvas::~ICanvas ( )
{
  // Remove the recoordination handler.  May be a noop if never added,
  // or a derived class already removed it in its ctor.
  IRecoordHandler::defaultHandler()->stopHandlingEventsFor( this );

  if ( bClDefaultHandlerAdded )
  {                               // Remove the private handler.
     CanvasStatics.keyboardHandler().stopHandlingEventsFor( this );
     CanvasStatics.canvasHandler().stopHandlingEventsFor( this );
     bClDefaultHandlerAdded = false;        // Flag as removed.
  }
#ifdef IC_WIN
  if (fFont)
    delete fFont;
#endif
  if (fCanvasData)
    delete fCanvasData;
}

/*------------------------------------------------------------------------------
| ICanvas::defaultStyle                                                        |
|                                                                              |
| Return the current default style for the class.                              |
------------------------------------------------------------------------------*/
ICanvas::Style ICanvas::defaultStyle ( )
{
  return currentDefaultStyle;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| ICanvas::setDefaultStyle                                                     |
|                                                                              |
| Sets the current default style for the class.                                |
------------------------------------------------------------------------------*/
void ICanvas::setDefaultStyle ( const ICanvas::Style& style )
{
  currentDefaultStyle = style;
}
#endif

/*------------------------------------------------------------------------------
| ICanvas::convertToGUIStyle                                                   |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long ICanvas::convertToGUIStyle ( const IBitFlag& guiStyle,
                                           Boolean bExtOnly ) const
{
  // Obtain the style from the class (IControl) that we inherit from.
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if ( bExtOnly )
  {
    // Use mask to only return extended styles in the user defined range.
    ulStyle |= guiStyle.asExtendedUnsignedLong() & IS_EXTMASK;
  }
  else
  {
    // No styles are defined for the base canvas, so we do not mask.
    ulStyle |= guiStyle.asUnsignedLong() |
#ifdef IC_WIN
               WS_CHILD |
#endif
    // Add the static background rectangle and group styles.
               SS_BKGNDRECT | IControl::group.asUnsignedLong();
  }

  return ulStyle;
}

/*------------------------------------------------------------------------------
| ICanvas::backgroundColor                                                     |
|                                                                              |
| Returns the background color of the canvas.                                  |
------------------------------------------------------------------------------*/
IColor ICanvas::backgroundColor ( ) const
{
   return Inherited::color( PP_BACKGROUNDCOLOR,
                            defaultbackgroundcolor( handle() ));
              // The above accounts for dialog vs window background.
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| ICanvas::setFont                                                             |
|                                                                              |
| Sets the font for the canvas.  Windows does not allow a font to be set for   |
| a static control that is not of style text.  Since canvas is SS_WHITERECT,   |
| we must save the font for our own usage.                                     |
------------------------------------------------------------------------------*/
IWindow& ICanvas::setFont( const IFont& fm )
{
  // First save this font in our window data before propagating.
  // Note that storing the font in the ICanvas::fFont member data is needed
  // to support the separate requirement for constructing an IFont from the
  // window handle of a canvas.  Calling IWindow::setFont is needed to
  // propagate the font to owned windows.  This order is important, so that
  // the font in ICanvas::fFont is stored before IWindow::setLayoutDistorted
  // is called with IWindow::fontChanged.
  if ( this->fFont )
  {                // Clean up the prior font.
     delete this->fFont;
  }
  this->fFont = new IFont( fm );
  Inherited::setFont( fm );

  // Bump in propagation static by 1 (since nested calls could occur).
  propagationInProgress++;

  // Now lets propagate this font to all of the children of the canvas.
  // This is to give it similar behavior to OS/2 pres-params support.
  IWindowHandle hwndChild = 0;
  IWindow::ChildCursor cursor( *(IWindow*)this );

  // Enumerate all child windows.
  for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
  {
    // Get the IWindow object for each child found
    Boolean bNewChild = false;
    IWindowHandle hwndChild = this->childAt( cursor );
    IWindow* pwndChild = IWindow::windowWithHandle( hwndChild );

    // If no IWindow object exists, create one temporarily
    if ( !pwndChild )
    {
      pwndChild = new IWindow( hwndChild );
      bNewChild = true;
    }

    // Set the font for that window
    pwndChild->setFont( fm );

    // Delete any temporary IWindow object created (if one was needed)
    if ( bNewChild )
       delete pwndChild;
  }
  propagationInProgress--;
  return *this;
}


/*------------------------------------------------------------------------------
| ICanvas::resetFont                                                           |
|                                                                              |
| Resets the font for the canvas.  Windows does not allow a font to be set for |
| a static control that is not of style text.  Since canvas is SS_WHITERECT,   |
| we must perform the font actions ourselves.                                  |
------------------------------------------------------------------------------*/
IWindow& ICanvas::resetFont( )
{
  // First delete any existing font in our private data.
  if ( this->fFont )
  {                // Clean up the prior font.
     delete this->fFont;
     this->fFont = 0;
  }
  Inherited::resetFont();

  // No need to propagate since Windows will automatically update any
  // windows which had a dependency on the stored font.
  return *this;
}
#endif //IC_WIN

/*------------------------------------------------------------------------------
| ICanvas::isTabStop                                                           |
|                                                                              |
| Returns true if the canvas can be tabbed to based on whether any of its      |
| child windows can be tabbed to.                                              |
------------------------------------------------------------------------------*/
Boolean ICanvas::isTabStop ( ) const
{
  Boolean bTabStop = false;       // Assume no child windows are tab stops.
  IWindowHandle hwndChild = 0;
  IWindow::ChildCursor cursor( *(IWindow*)this );

  for ( cursor.setToFirst();
        cursor.isValid()  &&  bTabStop == false;
        cursor.setToNext() )
  {                               // Enumerate all child windows.
     Boolean bNewChild = false;
     IWindowHandle hwndChild = this->childAt( cursor );
     IWindow* pwndChild = IWindow::windowWithHandle( hwndChild );
     if ( !pwndChild )
     {                            // Child window object found.
        pwndChild = new IWindow( hwndChild );
        bNewChild = true;
     }
     bTabStop = pwndChild->isTabStop();
              // Note: Can't just check style directly, since if the
              // child is another canvas, its isTabStop() function
              // needs to be called.
     if ( bNewChild )
     {                                 // Memory clean-up needed.
        delete pwndChild;
     }
  }                               // End for all child windows.

  ((ICanvas*)this)->enableTabStop( bTabStop );
                 // Make non-const to store appropriate tabStop style.
  return bTabStop;
}

/*------------------------------------------------------------------------------
| ICanvas::origDefaultButtonHandle                                             |
|                                                                              |
| Return the original default push button for the frame window.                |
------------------------------------------------------------------------------*/
IWindowHandle ICanvas::origDefaultButtonHandle ( ) const
{
  return ICanvasStatics::origDefaultButtonHandle( this->handle() );
}

/*------------------------------------------------------------------------------
| ICanvas::defaultPushButton                                                   |
|                                                                              |
| Return the first child window that is a default push button, if one exists.  |
------------------------------------------------------------------------------*/
IWindowHandle ICanvas::defaultPushButton ( ) const
{
  return ICanvasStatics::defaultPushButtonHandle( this );
}

/*------------------------------------------------------------------------------
| ICanvas::matchForMnemonic                                                    |
|                                                                              |
| Returns the first child window that uses the specified character as a        |
| mnemonic.                                                                    |
------------------------------------------------------------------------------*/
IWindowHandle ICanvas::matchForMnemonic ( unsigned short character ) const
{
  return ICanvasStatics::matchForMnemonic( this, character );
}

/*------------------------------------------------------------------------------
| ICanvas::calcMinimumSize                                                     |
|                                                                              |
| Calculate the minimum screen size needed by the canvas.  This is the size    |
| needed to display all child windows (as determined by the layout routine).   |
------------------------------------------------------------------------------*/
ISize ICanvas::calcMinimumSize ( ) const
{
  ((ICanvas*)this)->layout();        // Calculate minimum size.
  ISize sizMin = layoutSize();       // Use size from layout().
  return sizMin;
}

/*------------------------------------------------------------------------------
| ICanvas::setLayoutDistorted                                                  |
|                                                                              |
| Provide common canvas processing for various distortion notifications.       |
------------------------------------------------------------------------------*/
ICanvas& ICanvas::setLayoutDistorted( unsigned long layoutAttributeOn,
                                      unsigned long layoutAttributeOff )
{
  unsigned long ulFlagsOn = layoutAttributeOn;

  if ( layoutAttributeOn & IWindow::childWindowCreated )
  {                                    // New child window.
     ulFlagsOn &= (unsigned long)(~IWindow::childWindowCreated);
     ulFlagsOn |= IWindow::layoutChanged;
  }                                    // Need new layout.
  if ( layoutAttributeOn & IWindow::childWindowDestroyed )
  {                                    // Lost a child window.
     ulFlagsOn &= (unsigned long)(~IWindow::childWindowDestroyed);
     ulFlagsOn |= IWindow::layoutChanged;
  }                                    // Need new layout.

  Inherited::setLayoutDistorted( ulFlagsOn, layoutAttributeOff );

  if ( layoutAttributeOn  &&
       this->isLayoutDistorted( IWindow::immediateUpdate )  &&
       this->isVisible() )
  {                                    // Need to update canvas appearance.
     Inherited::setLayoutDistorted( 0, IWindow::immediateUpdate );
     if ( this->size() > ISize() )
     {                                 // Truly visible window.
        this->refresh();               // Don't bypass paint handlers.
     }
     else                              // Height or width is 0.
     {                                 // Will never get a paint.
        this->layout();                // Circumvent WM_PAINT.
     }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| ICanvas::layout                                                              |
|                                                                              |
| This method should be overridden by derived classes to position/size child   |
| windows.                                                                     |
------------------------------------------------------------------------------*/
ICanvas& ICanvas::layout ( )
{
  IMODTRACE_DEVELOP( "ICanvas::layout()" );

  if ( this->isLayoutDistorted( IWindow::layoutChanged ))
  {
     ITRACE_DEVELOP( "isLayoutDistorted(layoutChanged)==true" );
     /**************************************************************/
     /* The layout is distorted so see if this canvas needs        */
     /* updating.                                                  */
     /*                                                            */
     /* The user is responsible for positioning child windows and  */
     /* sizing the canvas large enough so that all child windows   */
     /* are contained by the canvas.                               */
     /*                                                            */
     /* Some of this responsibility will be easied if the canvas   */
     /* is a child of a specialized canvas (one that sizes and     */
     /* positions children according to minimumSize()).            */
     /* minimumSize() calls ICanvas::calcMinimumSize() if the user */
     /* has not specified a minimum size himself.  This function   */
     /* will return the size stored by the layout() routine, one   */
     /* large enough to contain all the child windows.             */
     /**************************************************************/
     ISize sizLayout = this->layoutSize(); // Save original size for children.

     IWindowPosBuffer wposbuf = this->fixupChildren();
     wposbuf.apply( false );           // Reorder z-order of children.

     this->setLayoutDistorted( 0, IWindow::layoutChanged );
     if ( sizLayout != this->layoutSize() )  // New minimum size for children.
     {                                 // Cause layout in parent canvas.
        this->setLayoutDistorted( IWindow::minimumSizeChanged, 0 );
     }
  }   /* end isDistorted */

  return *this;
}

/*------------------------------------------------------------------------------
| ICanvas::fixupChildren                                                       |
|                                                                              |
| Generates a list of all child windows, determines whether any can be tabbed  |
| to, and allows for reversing the order in which they are iterated.           |
------------------------------------------------------------------------------*/
IWindowPosBuffer ICanvas::fixupChildren ( )
{
  IMODTRACE_DEVELOP( "ICanvas::fixupChildren()" );
  /*****************************************************************/
  /* Dynamically set the WS_TABSTOP style based on whether the     */
  /* canvas has any tab-able children (WS_TABSTOP is needed so     */
  /* PM's frame window procedure and the ICanvasHandler will       */
  /* consider giving the input focus to the canvas during Tab key  */
  /* processing, and so the canvas can receive WM_SETFOCUS and     */
  /* give the input focus to an appropriate child window).         */
  /*                                                               */
  /* Additionally, reverse the z-order of the canvas' child        */
  /* windows, since the order that windows are constructed and     */
  /* expected to be layed out is typically the opposite of the     */
  /* z-order.  Because of complications involving tabbing and      */
  /* cursor movement with autoselect radio buttons, it became far  */
  /* easier to reverse the z-order of the child windows than to    */
  /* compensate in the tab and cursor movement code to fix this.   */
  /*****************************************************************/
  IWindowPosBuffer wposbuf( this );
  unsigned long childCount = 0;
  Boolean bTabControl = false;         // Assume no tab-able children.
  IPoint ptMax;
  IWindowHandle hwndChild = 0;
  IWindowHandle hwndLastChild = 0;
  Boolean
    defaultPBFound = ( this->origDefaultButtonHandle() == 0 ) ?
                          false : true;

  IWindow::ChildCursor cursor( *(IWindow*)this );
  for ( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
  {                                    // Enumerate all child controls.
     IPoint ptChild;
     childCount++;
     hwndChild = this->childAt( cursor );
#ifdef IC_PM
     Boolean bNewChild = false;
     IWindow* pwndChild = IWindow::windowWithHandle( hwndChild );
     if ( ! pwndChild )
     {                                 // A non-wrappered window.
        pwndChild = new IWindow( hwndChild );
        bNewChild = true;
     }

     ptChild = pwndChild->rect().topRight();
     if ( bTabControl == false )
     {                                 // No tab-able children found yet.
        if ( pwndChild->isTabStop() )
        {                         // Child window that can be tabbed to.
           bTabControl = true;         // Flag as such.
        }
        /*******************************************************/
        /* Note: if child is a canvas, isTabStop() will allow  */
        /* it to check all its children for a tab-able control)*/
        /*******************************************************/
     }
     if ( bNewChild )
     {                                 // Memory clean-up needed.
        delete pwndChild;
     }
#endif // IC_PM
     //********************* New code for defect 7915 ******************//
     //********************* contained in this ifdef  ******************//
#ifdef IC_WIN
     // Need to do this all natively without using IWindow wrappers since
     // the canvas children could be in different processes
     unsigned long ulStyle = ISTYLEOF(hwndChild);
     HWND parentHwnd = IPARENTOF( hwndChild );

     // Get window rectangle in coordinates relative to parent
     RECT          rect;
     GetWindowRect( hwndChild, &rect );
     if ( ulStyle & WS_POPUP )
       parentHwnd = IWindow::desktopWindow()->handle();
     IMAPWINDOWPOINTS( HWND_DESKTOP, parentHwnd, (POINT*)&rect, 2);
     IRectangle childRect(rect);

     // If conversion is needed, then calculate size of parent rectangle
     // and adjust coordinates accordingly
     if ( ICoordinateSystem::isConversionNeeded() )
     {
       RECT rectl;
       GetClientRect( parentHwnd, &rectl );
       ISize parentSize( rectl.right - rectl.left, rectl.bottom - rectl.top );
       childRect = ICoordinateSystem::convertToApplication( childRect,
                                                            parentSize );
     }
     ptChild = childRect.topRight();

     // If no tab-able children have been found yet AND
     //  this child window can be tabbed to, then flag as found
     if (( bTabControl == false ) && (ulStyle & WS_TABSTOP))
       bTabControl = true;

#endif // IC_WIN

     ptMax = ptMax.maximum( ptChild );

     if ( bClChildrenReversed == false )
     {                              // Need to reverse z-order of children.
        if ( hwndLastChild )
        {                           // Have a preceding child.
           wposbuf.orderAfter( hwndLastChild, hwndChild );
        }                           // Push preceding towards bottom.
        hwndLastChild = hwndChild;
     }                              // End reverse children.

     if ( ! defaultPBFound )
     {             // Frame doesn't have a default push button yet.
        if ( hwndChild.sendEvent( WM_QUERYDLGCODE, 0, 0 ).asUnsignedLong() &
             DLGC_DEFAULT)
        {                           // Found a default push button.
#ifdef IC_PMWIN
           // Save it away with the frame.
           IEventResult mr =        // Find the frame window.
                      hwndChild.sendEvent( WM_QUERYFOCUSCHAIN,
                                           IEventParameter1( QFC_FRAME ),
                                           IEventParameter2( 0 ));
#ifdef IC_WIN
           IWindowHandle hwndFrame( (HWND)mr.asUnsignedLong() );
           if (( hwndFrame ) &&
               (IWindowClassName(hwndFrame).asString().indexOf(WC_FRAME) == 1) )
           {
              ISETWINDOWULONG( hwndFrame, IGWL_DEFBUTTON, hwndChild.asUnsigned() );
           }
#else
           IWindowHandle hwndFrame( mr.asUnsignedLong() );
           if ( hwndFrame )
           {                        // Store the default button.
              ISETWINDOWULONG( hwndFrame, QWL_DEFBUTTON, hwndChild );
           }  // (Can't use IWindow::setWindowData, because it's protected.)
#endif
#else   // !IC_PMWIN
           hwndClDefaultButton = hwndChild;  // Save it away.
#endif  // !IC_PMWIN
        }                           // End found a default push button.
     }                              // End default push button search.
  }                                 // End enumerate all children.

  fCanvasData->fChildCount = childCount;
  if ( bClChildrenReversed == false  &&  hwndLastChild )
  {                                 // Need to reverse z-order of children.
     wposbuf.orderAsFirst( hwndLastChild );  // Re-order last one.
     bClChildrenReversed = true;
  }                                 // Make topmost child.
  this->enableTabStop( bTabControl );
  this->setLayoutSize( ISize( ptMax ));  // Store size used by layout.

  return wposbuf;
}
