/*******************************************************************************
* FILE NAME: iframext.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iframext.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #define INCL_WINFRAMEMGR
  #define INCL_WINSYS
  #define INCL_GPIPRIMITIVES
  #define INCL_GPILOGCOLORTABLE
  #include <iwindefs.h>
  #include <math.h>
  }

#include <ihandle.hpp>
#include <iframext.hpp>
#include <iswp.hpp>
#include <iplatfrm.hpp>
#include <ipoint.hpp>
#include <itrace.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IFRAMEXT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IFrameExtension::attachedToId                                                |
|                                                                              |
| Return id of attached-to frame, based on location data member value.         |
------------------------------------------------------------------------------*/
unsigned long IFrameExtension :: attachedToId ( ) const
  {
  switch ( this->extLocation )
    {
    case IFrameWindow::rightOfTitleBar:
    case IFrameWindow::leftOfTitleBar:
#ifdef IC_WIN
      return 0;          // not supported ... causes extension to be hidden
#else
      return FID_TITLEBAR;
#endif
    case IFrameWindow::rightOfMenuBar:
    case IFrameWindow::leftOfMenuBar:
#ifdef IC_WIN
      return 0;          // not supported ... causes extension to be hidden
#else
      return FID_MENU;
#endif
    default:
      return FID_CLIENT;
    }
  }

/*------------------------------------------------------------------------------
| IFrameExtension::attachTo                                                    |
|                                                                              |
| Calculate the size and position of the extension from the argument base      |
| control size/position, according to extension location.  Adjust the          |
| size and position of the attached-to frame control.                          |
------------------------------------------------------------------------------*/
ISWP IFrameExtension :: attachTo ( ISWP &baseSWP )
  {
  IMODTRACE_DEVELOP( "IFrameExtension::attachTo" );

  ITRACE_ALL( IString( "input baseSWP=" )
              +
              baseSWP.position().asString()
              +
              baseSWP.size().asString() );

  ISWP
    extSWP;

  // Set extension size...
  extSWP.setSize( this->sizeTo( baseSWP.size() ) );

  // Set extension position...
  switch ( this->location() )
    {
#ifdef IC_ORIGINUPPERLEFT       // 0,0 upper left
    case IFrameWindow::aboveClient:
#else                           // 0,0 lower left
    case IFrameWindow::belowClient:
#endif
    case IFrameWindow::leftOfClient:
    case IFrameWindow::leftOfTitleBar:
    case IFrameWindow::leftOfMenuBar:
      extSWP.setPosition( baseSWP.position() );
      break;
#ifdef IC_ORIGINUPPERLEFT       // 0,0 upper left
    case IFrameWindow::belowClient:
#else                           // 0,0 lower left
    case IFrameWindow::aboveClient:
#endif
    case IFrameWindow::rightOfClient:
    case IFrameWindow::rightOfTitleBar:
    case IFrameWindow::rightOfMenuBar:
      extSWP.setPosition( baseSWP.position()
                          +
                          baseSWP.size()
                          -
                          extSWP.size() );
      break;
    }

  // Remember size/position for use during painting...
  this->extPosition = extSWP.position();
  this->extSize     = extSWP.size();

  // Adjust base control position...
  switch ( this->location() )
    {
    case IFrameWindow::aboveClient:
#ifdef IC_ORIGINUPPERLEFT     //  0,0 upper left
      baseSWP
        .setPosition( baseSWP.position()
                      +
                      IPoint( 0, extSWP.size().height()
                                 +
                                 this->separatorWidth() ) );
#endif
      break;
    case IFrameWindow::rightOfClient:
    case IFrameWindow::rightOfTitleBar:
    case IFrameWindow::rightOfMenuBar:
      break;
    case IFrameWindow::belowClient:
#ifdef  IC_ORIGINLOWERLEFT    //  0,0 lower left
      baseSWP
        .setPosition( baseSWP.position()
                      +
                      IPoint( 0, extSWP.size().height()
                                 +
                                 this->separatorWidth() ) );
#endif
      break;
    case IFrameWindow::leftOfClient:
    case IFrameWindow::leftOfTitleBar:
    case IFrameWindow::leftOfMenuBar:
      baseSWP
        .setPosition( baseSWP.position()
                      +
                      IPoint( extSWP.size().width()
                              +
                              this->separatorWidth(), 0 ) );
      break;
    }

  // Adjust base control size...
  if ( this->location() == IFrameWindow::aboveClient
       ||
       this->location() == IFrameWindow::belowClient )
    baseSWP.setSize( baseSWP.size() - IPoint( 0, extSWP.size().height()
                                                 +
                                                 this->separatorWidth() ) );
  else
    baseSWP.setSize( baseSWP.size() - IPoint( extSWP.size().width()
                                              +
                                              this->separatorWidth(), 0 ) );

  ITRACE_ALL( IString( "output baseSWP=" )
              +
              baseSWP.position().asString()
              +
              baseSWP.size().asString() );

  ITRACE_ALL( IString( "output extSWP=" )
              +
              extSWP.position().asString()
              +
              extSWP.size().asString() );

  return extSWP;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::sizeTo                                                      |
|                                                                              |
| Calculates and returns the size this extension would have when attached to   |
| a control of the argument size.                                              |
| Notes: If top/bottom, then the width is the same, otherwise the height is    |
|          the same.                                                           |
|        If fixed size, then the other dimension is fixed, otherwise it is     |
|          calculated relative to the base dimension.                          |
------------------------------------------------------------------------------*/
ISize IFrameExtension :: sizeTo ( const ISize &baseSize ) const
  {
  IMODTRACE_DEVELOP( "IFrameExtension::sizeTo" );

  ITRACE_ALL( IString( "input=" ) + baseSize.asString() );

  ISize
    result = baseSize;
  switch ( this->location() )
    {
    case IFrameWindow::aboveClient:
    case IFrameWindow::belowClient:
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() < result.height() )
          result.setHeight( this->fixedSize() );
        }
      else if ( this->type() == minimumSize )
        {
        result.setHeight( this->control()->minimumSize().height() );
        }
      else
        {
        result.setHeight( (long)(result.height() * this->relativeSize()) );
        }
      break;

    default:
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() < result.width() )
          result.setWidth( this->fixedSize() );
        }
      else if ( this->type() == minimumSize )
        {
        result.setWidth( this->control()->minimumSize().width() );
        }
      else
        {
        result.setWidth( (long)(result.width() * this->relativeSize()) );
        }
      break;
    }

  ITRACE_ALL( IString( "output=" ) + result.asString() );

  return result;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::totalRectFor                                                |
|                                                                              |
| Calculates and returns a rectangle from a base rectangle plus                |
| the frame extensions size and position.                                      |
------------------------------------------------------------------------------*/
IRectangle IFrameExtension::totalRectFor( const IRectangle& baseRect ) const
  {
  IMODTRACE_DEVELOP( "IFrameExtension::totalRectFor" );

  ITRACE_ALL( IString( "input=" ) + baseRect.asString() );

  IRectangle result = baseRect;
  long delta = 0;

  switch ( this->location() )
    {
#ifdef  IC_ORIGINUPPERLEFT   // 0,0 upper left
    // If the extension is below the client, increase the height of the
    // resulting rectangle by the height of the frame extension.
    case IFrameWindow::belowClient:
#else                       // 0,0 lower left
    // If the extension is above the client, increase the height of the
    // resulting rectangle by the height of the frame extension.
    case IFrameWindow::aboveClient:
#endif
      delta = result.height();
      if ( this->type() == fixed )
        {
        delta += this->fixedSize();
        }
      else if ( this->type() == minimumSize )
        {
        delta += this->control()->minimumSize().height();
        }
      else
        {
        delta = (long)ceil( delta / (1.0 - this->relativeSize()) );
        }

      result.sizeTo( IPair( result.width(),
                            delta + this->separatorWidth() ));
      break;

#ifdef  IC_ORIGINUPPERLEFT      //  0,0 upper left
    // If the extension is above the client, increase the height of the
    // resulting rectangle by the height of the frame extension and move it
    // up by the same amount.
    case IFrameWindow::aboveClient:
#else
    // If the extension is below the client, increase the height of the
    // resulting rectangle by the height of the frame extension and move it
    // down by the same amount.
    case IFrameWindow::belowClient:
#endif
      if ( this->type() == fixed )
        {
        delta = this->fixedSize();
        }
      else if ( this->type() == minimumSize )
        {
        delta = this->control()->minimumSize().height();
        }
      else
        {
        delta = (long)ceil( result.height() * this->relativeSize() /
                            (1.0 - this->relativeSize()) );
        }

      delta += this->separatorWidth();
      result
       .sizeTo( IPair( result.width(), result.height() + delta ))
       .moveTo( IPair( result.left(), result.bottom() - delta ));
      break;

    // If the extension is to the left, increase the width of the resulting
    // rectangle by the width of the frame extension and move it to the left
    // by the same amount.
    case IFrameWindow::leftOfClient:
    case IFrameWindow::leftOfTitleBar:
    case IFrameWindow::leftOfMenuBar:
      if ( this->type() == fixed )
        {
        delta = this->fixedSize();
        }
      else if ( this->type() == minimumSize )
        {
        delta = this->control()->minimumSize().width();
        }
      else
        {
        delta = (long)ceil( result.width() * this->relativeSize() /
                            (1.0 - this->relativeSize()) );
        }

      delta += this->separatorWidth();
      result
       .sizeTo( IPair( result.width() + delta, result.height() ))
       .moveTo( IPair( result.left() - delta, result.bottom() ));
      break;

    // If the extension is to the right, increase the width of the resulting
    // rectangle by the width of the frame extension.
    case IFrameWindow::rightOfClient:
    case IFrameWindow::rightOfTitleBar:
    case IFrameWindow::rightOfMenuBar:
      delta = result.width();
      if ( this->type() == fixed )
        {
        delta += this->fixedSize();
        }
      else if ( this->type() == minimumSize )
        {
        delta += this->control()->minimumSize().width();
        }
      else
        {
        delta = (long)ceil( delta / (1.0 - this->relativeSize()) );
        }

      result.sizeTo( IPair( result.width() + this->separatorWidth(),
                            result.height() ));
      break;
    }
  return result;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::baseRectFor                                                 |
|                                                                              |
| Calculates and returns a rectangle from a total rectangle (i.e. a client     |
| rectangle with a frame extension) minus the frame extensions size and        |
| position.                                                                    |
------------------------------------------------------------------------------*/
IRectangle IFrameExtension::baseRectFor( const IRectangle& totalRect ) const
  {
  IMODTRACE_DEVELOP( "IFrameExtension::baseRectFor" );

  ITRACE_ALL( IString( "input=" ) + totalRect.asString() );

  IRectangle result = totalRect;
  long delta = 0;
  Boolean adjust = true;

  switch ( this->location() )
    {
#ifdef  IC_ORIGINUPPERLEFT      // 0,0 upper left
    // If the extension is below the client, decrease the height of the
    // resulting rectangle by the height of the frame extension.
    case IFrameWindow::belowClient:
#else                           // 0,0 lower left
    // If the extension is above the client, decrease the height of the
    // resulting rectangle by the height of the frame extension.
    case IFrameWindow::aboveClient:
#endif
      delta = result.height();
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() > result.height() )
          {
          delta = this->separatorWidth();
          }
        else
          {
          delta -= this->fixedSize();
          }
        }
      else if ( this->type() == minimumSize )
        {
        delta -= this->control()->minimumSize().height();
        }
      else
        {
        delta -= floor( delta * this->relativeSize() );
        }

      result.sizeTo( IPair( result.width(),
                            delta - this->separatorWidth() ));
      break;

#ifdef  IC_ORIGINUPPERLEFT      //  0,0 upper left
    // If the extension is above the client, decrease the height of the
    // resulting rectangle by the height of the frame extension and move it
    // down by the same amount.
    case IFrameWindow::aboveClient:
#else                           //  0,0 lower left
    // If the extension is below the client, decrease the height of the
    // resulting rectangle by the height of the frame extension and move it
    // up by the same amount.
    case IFrameWindow::belowClient:
#endif
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() > result.height() )
          {
          result.sizeTo( IPair( result.width(), 0 ));
          adjust = false;
          }
        else
          {
          delta = this->fixedSize();
          }
        }
      else if ( this->type() == minimumSize )
        {
        delta = this->control()->minimumSize().height();
        }
      else
        {
        delta = (long)floor( this->relativeSize() * result.height() );
        }

      if ( adjust )
      {
         delta += this->separatorWidth();
         result
          .sizeTo( IPair( result.width(), result.height() - delta ))
          .moveTo( IPair( result.left(), result.bottom() + delta ));
      }
      break;

    // If the extension is to the left, decrease the width of the resulting
    // rectangle by the width of the frame extension and move it to the right
    // by the same amount.
    case IFrameWindow::leftOfClient:
    case IFrameWindow::leftOfTitleBar:
    case IFrameWindow::leftOfMenuBar:
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() > result.width() )
          {
          result.sizeTo( IPair( 0, result.height() ));
          adjust = false;
          }
        else
          {
          delta = this->fixedSize();
          }
        }
      else if ( this->type() == minimumSize )
        {
        delta = this->control()->minimumSize().width();
        }
      else
        {
        delta = (long)floor( this->relativeSize() * result.width() );
        }

      if ( adjust )
      {
         delta += this->separatorWidth();
         result
          .sizeTo( IPair( result.width() - delta, result.height() ))
          .moveTo( IPair( result.left() + delta, result.bottom() ));
      }
      break;

    // If the extension is to the right, decrease the width of the resulting
    // rectangle by the width of the frame extension.
    case IFrameWindow::rightOfClient:
    case IFrameWindow::rightOfTitleBar:
    case IFrameWindow::rightOfMenuBar:
      delta = result.width();
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() > result.height() )
          {
          delta = this->separatorWidth();
          }
        else
          {
          delta -= this->fixedSize();
          }
        }
      else if ( this->type() == minimumSize )
        {
        delta -= this->control()->minimumSize().width();
        }
      else
        {
        delta -= floor( delta * this->relativeSize() );
        }

      result.sizeTo( IPair( delta - this->separatorWidth(),
                            result.height() ));
      break;
    }
  return result;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::drawSeparator                                               |
|                                                                              |
| Draws the separator line between an extension and the adjacent control       |
| (or extension) to which it is attached.                                      |
------------------------------------------------------------------------------*/
void IFrameExtension :: drawSeparator ( const IPresSpaceHandle &hps )
  {
  IMODTRACE_DEVELOP( "IFrameExtension::drawSeparator" );

  // If no separator, don't even bother...
  if ( this->extSeparator == IFrameWindow::none )
    return;

  // Set up arrays for line start/end points and colors:
  IPoint
    start[3],
    end  [3];

  long
    color[3] = { SYSCLR_BUTTONDARK, SYSCLR_BUTTONMIDDLE, SYSCLR_BUTTONDARK };

  int
    numLines = (int)this->separatorWidth();

  // Otherwise, set up line start/end points according to location...
  switch ( this->extLocation )
    {
#ifdef  IC_ORIGINUPPERLEFT      // 0,0 upper left
    case IFrameWindow::belowClient:
#else                           // 0,0 lower left
    case IFrameWindow::aboveClient:
#endif
      start[0] = start[1] = start[2] = this->extPosition
                                       -
                                       IPoint( 0, 1 );
      start[1] -= IPoint( 0, 1 );
      start[2] -= IPoint( 0, 2 );
      end[0] = start[0] + IPoint( this->extSize.width(), 0 );
      end[1] = start[1] + IPoint( this->extSize.width(), 0 );
      end[2] = start[2] + IPoint( this->extSize.width(), 0 );
      break;
#ifdef  IC_ORIGINUPPERLEFT      // 0,0 upper left
    case IFrameWindow::aboveClient:
#else                           // 0,0 lower left
    case IFrameWindow::belowClient:
#endif
      start[0] = start[1] = start[2] = this->extPosition
                                       +
                                       IPoint( 0, this->extSize.height() );
      start[1] += IPoint( 0, 1 );
      start[2] += IPoint( 0, 2 );
      end[0] = start[0] + IPoint( this->extSize.width(), 0 );
      end[1] = start[1] + IPoint( this->extSize.width(), 0 );
      end[2] = start[2] + IPoint( this->extSize.width(), 0 );
      break;
    case IFrameWindow::leftOfClient:
    case IFrameWindow::leftOfMenuBar:
    case IFrameWindow::leftOfTitleBar:
      start[0] = start[1] = start[2] = this->extPosition
                                       +
                                       IPoint( this->extSize.width(), 0 );
      start[1] += IPoint( 1, 0 );
      start[2] += IPoint( 2, 0 );
      end[0] = start[0] + IPoint( 0, this->extSize.height() );
      end[1] = start[1] + IPoint( 0, this->extSize.height() );
      end[2] = start[2] + IPoint( 0, this->extSize.height() );
      break;
    case IFrameWindow::rightOfClient:
    case IFrameWindow::rightOfMenuBar:
    case IFrameWindow::rightOfTitleBar:
      start[0] = start[1] = start[2] = this->extPosition
                                       -
                                       IPoint( 1, 0 );
      start[1] -= IPoint( 1, 0 );
      start[2] -= IPoint( 2, 0 );
      end[0] = start[0] + IPoint( 0, this->extSize.height() );
      end[1] = start[1] + IPoint( 0, this->extSize.height() );
      end[2] = start[2] + IPoint( 0, this->extSize.height() );
      break;
    }
  while ( numLines-- )
    {
    ITRACE_ALL( "Line:"
                +
                start[numLines].asString()
                +
                end[numLines].asString()
                +
                IString( color[numLines] ) );
#ifdef IC_WIN
    HPEN hpen = CreatePen( PS_SOLID, 0,
                           IQUERYSYSCOLOR( (int)color[numLines] ) );
    HGDIOBJ hOldObj = SelectObject( hps, hpen );
#else
    GpiSetColor( hps,
                 IQUERYCOLORINDEX( hps,
                                   IQUERYSYSCOLOR( color[numLines] ) ) );
#endif
    IMOVE( hps, ((PPOINTL)(&start[numLines])) );
    ILINE( hps, ((PPOINTL)(&end  [numLines])) );
#ifdef IC_WIN
    SelectObject( hps, hOldObj );
    DeleteObject( hpen );
#endif
    }
  }

/*------------------------------------------------------------------------------
| IFrameExtensions::IFrameExtensions                                           |
------------------------------------------------------------------------------*/
IFrameExtensions :: IFrameExtensions ( )
  {
  }

/*------------------------------------------------------------------------------
| IFrameExtensions::~IFrameExtensions                                          |
------------------------------------------------------------------------------*/
IFrameExtensions :: ~IFrameExtensions ( )
  {
  }

/*------------------------------------------------------------------------------
| IFrameExtension::separatorWidth                                              |
|                                                                              |
| Returns the width of the separator line.                                     |
------------------------------------------------------------------------------*/
unsigned long IFrameExtension :: separatorWidth ( ) const
  {
  unsigned long thickness = 0;
  switch ( this->extSeparator )
    {
    case IFrameWindow::none:
    default:
      break;            // Use default = 0.
    case IFrameWindow::thinLine:
      thickness = 1;
      break;
    case IFrameWindow::thickLine:
      thickness = 3;
      break;
    }
  return thickness;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::setSize                                                     |
|                                                                              |
| Resets the size of the extension.                                            |
------------------------------------------------------------------------------*/
IFrameExtension &IFrameExtension :: setSize ( unsigned long widthOrHeight )
  {
  this->extType = IFrameExtension::fixed;
  this->extFixedSize = widthOrHeight;
  return *this;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::setSize                                                     |
|                                                                              |
| Convert width/height arg to unsigned long and use overloaded function.       |
------------------------------------------------------------------------------*/
IFrameExtension &IFrameExtension :: setSize ( int widthOrHeight )
  {
  return this->setSize( (unsigned long)widthOrHeight );
  }

/*------------------------------------------------------------------------------
| IFrameExtension::setSize                                                     |
|                                                                              |
| Resets the size of the extension.                                            |
------------------------------------------------------------------------------*/
IFrameExtension &IFrameExtension :: setSize ( double widthOrHeight )
  {
  this->extType = IFrameExtension::relative;
  this->extRelativeSize = widthOrHeight;
  return *this;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::useMinimumSize                                              |
|                                                                              |
| Specifies that the height or width of the extension should be sized to the   |
| height or width of its window's minimum size.                                |
------------------------------------------------------------------------------*/
IFrameExtension &IFrameExtension :: useMinimumSize ( )
  {
  this->extType = IFrameExtension::minimumSize;
  this->control()->setLayoutDistorted( IWindow::minimumSizeChanged, 0 );
  return *this;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::relativeSize                                                |
|                                                                              |
| Returns the relative width or height as a percentage                         |
| (0.0 if the extension is of type fixed).                                     |
------------------------------------------------------------------------------*/
double IFrameExtension :: relativeSize ( ) const
  {
  return ( this->extType == IFrameExtension::relative )
    ? this->extRelativeSize
    : 0.0;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::fixedSize                                                   |
|                                                                              |
| Returns the fixed width or height in window coordinates                      |
| (0 if the extension is of type relative).                                    |
------------------------------------------------------------------------------*/
unsigned long IFrameExtension :: fixedSize ( ) const
  {
  return ( this->extType == IFrameExtension::fixed )
    ? this->extFixedSize
    : 0;
  }
