/*******************************************************************************
* FILE NAME: ipainevt.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ipainevt.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #define INCL_GPILOGCOLORTABLE
  #define INCL_GPIPRIMITIVES
  #define INCL_GPIREGIONS
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
  #include <string.h>
}

#include <ipainevt.hpp>
#include <icolor.hpp>
#include <iexcept.hpp>
#include <ihandle.hpp>
#include <irect.hpp>
#include <istring.hpp>
#include <iwindow.hpp>
#include <ireslib.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IPAINEVT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IPaintEvent::IPaintEvent                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPaintEvent::IPaintEvent ( const IEvent& evt )
  : IEvent(evt), hpsCl(), rectCl(), setGC(false)
{
#ifdef IC_PM
   RECTL rcl;
   WinQueryUpdateRect( this->handle(), &rcl );
   if ((rcl.xLeft==rcl.xRight) || (rcl.yBottom==rcl.yTop))    // null rectangle
   {
      rectCl = IRectangle(0,0,0,0);   // Make it really null (both dimensions!)
   }
   else
   {
      rectCl = IRectangle( rcl.xLeft, rcl.yBottom,
                           rcl.xRight, rcl.yTop );
   } /* endif */

   //--------------------------------------------------------------------------
   // On OS/2 BIDI systems, we need to copy the BIDI attributes from the
   // window to the HPS (since OS/2 does not do this for us)
   //--------------------------------------------------------------------------
   IWindow* win = this->window();
   if ( IBidiSettings::isBidiSupported() )
   {
     IBidiSettings settings(*win);
     settings.apply(this->presSpaceHandle());
   }
#else
  RECT rct;
  GetUpdateRect( this->handle(), &rct, false );
  if ( (rct.left == 0) && (rct.right == 0) &&
       (rct.bottom == 0) && (rct.top == 0) )
  {
     rectCl = IRectangle(0,0,0,0);   // Make it really null (both dimensions!)
  }
  else
  {
     rectCl = IRectangle( rct.left, rct.top, rct.right, rct.bottom );
  }
#endif
}

/*------------------------------------------------------------------------------
| IPaintEvent::~IPaintEvent                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPaintEvent::~IPaintEvent ( )
{
#ifdef IC_PM
  HRGN oldrgn;
  if ( hpsCl )
  {
     if ( GpiSetClipRegion( hpsCl, NULLHANDLE, &oldrgn ) == RGN_ERROR )
        ITHROWGUIERROR( "GpiSetClipRegion" );
     if ( oldrgn )
     {
       if ( !( GpiDestroyRegion( hpsCl, oldrgn )))
          ITHROWGUIERROR( "GpiDestroyRegion" );
     }
     if ( !setGC )
     {
        this->window()->releasePresSpace( hpsCl );
     } /* endif */
  } /* endif */
#else
  if ( hpsCl )
  {
     if ( !setGC )
     {
        this->window()->releasePresSpace( hpsCl );
     } /* endif */
  } /* endif */
#endif
}

/*------------------------------------------------------------------------------
| IPaintEvent::presSpaceHandle                                                 |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPresSpaceHandle IPaintEvent::presSpaceHandle ( )
{
   if (!hpsCl)
   {
      hpsCl = window()->presSpace();
      this->initializeGC();
   }
   return hpsCl;
}

/*------------------------------------------------------------------------------
| IPaintEvent::setGraphicContext                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPaintEvent&
  IPaintEvent::setGraphicContext ( const IPresSpaceHandle& presHandle,
                                   Boolean setClipRegion )
{
#ifdef IC_PM
  HRGN oldrgn;
  if ( hpsCl )
  {
     if ( GpiSetClipRegion( hpsCl, NULLHANDLE, &oldrgn ) == RGN_ERROR )
        ITHROWGUIERROR( "GpiSetClipRegion" );
     if ( oldrgn )
     {
       if ( !( GpiDestroyRegion( hpsCl, oldrgn )))
          ITHROWGUIERROR( "GpiDestroyRegion" );
     }
     if ( !setGC )
     {
        this->window()->releasePresSpace( hpsCl );
     } /* endif */
  } /* endif */
#else
  if ( hpsCl )
  {
     if ( !setGC )
     {
        this->window()->releasePresSpace( hpsCl );
     } /* endif */
  } /* endif */
#endif
  hpsCl = presHandle;
  if ( setClipRegion )
  {
     this->initializeGC();
  }
  setGC = true;
  return *this;
}

/*------------------------------------------------------------------------------
| IPaintEvent::setGraphicContext                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPaintEvent& IPaintEvent::setGraphicContext ( const IGraphicContext& gc,
                                              Boolean setClipRegion )
{
#ifdef IC_WIN
  return this->setGraphicContext( (void*)gc.handle(), setClipRegion );
#else
  return this->setGraphicContext( gc.handle(), setClipRegion );
#endif
}

/*------------------------------------------------------------------------------
| IPaintEvent::clearBackground                                                 |
|                                                                              |
| Clear and fill a given rectangle area                                        |
------------------------------------------------------------------------------*/
IPaintEvent& IPaintEvent::clearBackground ( const IRectangle& rect,
                                            const IColor& clr )
{
   Boolean createdHPS = false;
   if (!hpsCl)
   {
     hpsCl = window()->presSpace();
     createdHPS = true;
   }
#ifdef IC_PM
   RECTL rcl = rect.asRECTL();
   if ( GpiQueryLogColorTable( hpsCl, LCOLOPT_INDEX, 0, 0, 0 ) == QLCT_RGB )
     WinFillRect( hpsCl, &rcl, clr.asRGBLong() );
   else
     WinFillRect( hpsCl, &rcl, clr.index() );
#else
   HBRUSH brush;
   RECT   rct;

   rct.top    = rect.minY();
   rct.bottom = rect.maxY();
   rct.left   = rect.minX();
   rct.right  = rect.maxX();

   brush = CreateSolidBrush( PALETTERGB( clr.redMix(),
                                         clr.greenMix(),
                                         clr.blueMix() ));

   FillRect( hpsCl, &rct, brush );

   DeleteObject( brush );

#endif
   if ( createdHPS )
   {
     this->window()->releasePresSpace( hpsCl );
     hpsCl = 0;
   }

   return *this;
}

/*------------------------------------------------------------------------------
| IPaintEvent::drawText                                                        |
|                                                                              |
| Draw a line of text starting at a given position.                            |
------------------------------------------------------------------------------*/
IPaintEvent& IPaintEvent::drawText ( const char* pszText,
                                     const IPoint& atPoint,
                                     const IColor& clr )
{
#ifdef IC_PM
   POINTL ptl = atPoint.asPOINTL();
   GpiSetMix( hpsCl, FM_OVERPAINT );
   if ( GpiQueryLogColorTable( hpsCl, LCOLOPT_INDEX, 0, 0, 0 ) == QLCT_RGB )
     GpiSetColor( hpsCl, clr.asRGBLong() );
   else
     GpiSetColor( hpsCl, clr.index() );
   GpiCharStringAt( hpsCl, &ptl, strlen(pszText), (PSZ)pszText );
#else
   // Not sure if I need to set the mix mode - MJS
   // Also, need to figure out a way to query if using the default color table - MJS

   SetTextColor(hpsCl, PALETTERGB(clr.redMix(), clr.greenMix(), clr.blueMix()));
   TextOut( hpsCl, atPoint.x(), atPoint.y(), pszText, strlen(pszText) );
#endif
   return *this;
}

/*------------------------------------------------------------------------------
| IPaintEvent::initializeGC                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPaintEvent& IPaintEvent::initializeGC ( )
{
  if ( rectCl.area() )
  {
#ifdef IC_PM
    HRGN  oldrgn, hrgn;
    RECTL rcl = rectCl.asRECTL();

    if ( (hrgn = GpiCreateRegion( hpsCl, 0, &rcl )) == RGN_ERROR )
       ITHROWGUIERROR( "GpiCreateRegion" );
    if ( WinQueryUpdateRegion( this->handle(), hrgn ) == RGN_ERROR )
       ITHROWGUIERROR( "WinQueryUpdateRegion" );
    if ( GpiSetClipRegion( hpsCl, hrgn, &oldrgn ) == RGN_ERROR )
       ITHROWGUIERROR( "GpiSetClipRegion" );
    if ( oldrgn == HRGN_ERROR )
       ITHROWGUIERROR( "GpiSetClipRegion" );
    if ( oldrgn != NULLHANDLE  &&  !( GpiDestroyRegion( hpsCl, oldrgn )))
       ITHROWGUIERROR( "GpiDestroyRegion" );
#else
    HRGN hrgnClip(CreateRectRgn ( 0,0,0,0 ));

    GetUpdateRgn( this->handle(), hrgnClip, false );

    SelectClipRgn ( hpsCl, hrgnClip );

    DeleteObject ( hrgnClip );
#endif
  }
  return *this;
}
