/*******************************************************************************
* FILE NAME: islider.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in islider.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C"
{
   #define INCL_WINERRORS
   #define INCL_WINSTDSLIDER
   #define INCL_DOSMODULEMGR
   #define INCL_WINMESSAGEMGR
   #define INCL_WINSYS
   #include <iwindefs.h>
   #ifdef IC_WIN
     #include <commctrl.h>
   #endif
}

#ifdef IC_WIN
  #ifndef _ICLSLIDERW_
  #include <iclsldw.h>
  #endif
  #include <icctlsta.hpp>         // statics to load dll
  #include <isliderw.hpp>         // native slider class
#endif
#include <islider.hpp>
#include <icconst.h>
#include <icmnfun.hpp>
#include <icolor.hpp>
#include <icoordsy.hpp>
#include <ievent.hpp>
#include <iexcept.hpp>
#include <ifont.hpp>
#include <ihandle.hpp>
#include <inotifev.hpp>
#include <iplatfrm.hpp>
#include <ipoint.hpp>
#include <irect.hpp>
#include <ireslib.hpp>
#include <istring.hpp>
#include <itrace.hpp>
#include <iwcname.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISLIDER_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public progress indicator styles.                                            |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)

const IProgressIndicator::Style
  IProgressIndicator::horizontal        ( 0, ISLS_HORIZONTAL ),
  IProgressIndicator::vertical          = SLS_VERTICAL,
  IProgressIndicator::alignCentered     ( 0, ISLS_CENTER ),
  IProgressIndicator::alignBottom       = SLS_BOTTOM,
  IProgressIndicator::alignTop          = SLS_TOP,
  IProgressIndicator::alignLeft         = SLS_LEFT,
  IProgressIndicator::alignRight        = SLS_RIGHT,
  IProgressIndicator::snapToTickMark    = SLS_SNAPTOINCREMENT,
  IProgressIndicator::handleDrawItem    = SLS_OWNERDRAW,
  IProgressIndicator::ribbonStrip       = SLS_RIBBONSTRIP,
  IProgressIndicator::homeBottom        ( 0, ISLS_HOMEBOTTOM ),
  IProgressIndicator::homeTop           = SLS_HOMETOP,
  IProgressIndicator::homeLeft          ( 0, ISLS_HOMELEFT ),
  IProgressIndicator::homeRight         = SLS_HOMERIGHT,
  IProgressIndicator::primaryScale1     ( 0, ISLS_PRIMARYSCALE1 ),
  IProgressIndicator::primaryScale2     = SLS_PRIMARYSCALE2,
  IProgressIndicator::pmCompatible      ( 0, ISLS_PMCOMPATIBILITY),
  IProgressIndicator::border3D          ( 0, IWS_BORDER3D ),
  IProgressIndicator::classDefaultStyle ( (WS_VISIBLE      |
                                           SLS_RIBBONSTRIP |
                                           SLS_READONLY),
                                                       (ISLS_HORIZONTAL |
                                                        ISLS_CENTER     |
                                                        ISLS_HOMELEFT   |
                                                        ISLS_PRIMARYSCALE1 |
                                                        IWS_BORDER3D) );
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new progress indicator objects (initial value).            |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)

IProgressIndicator::Style
  IProgressIndicator::currentDefaultStyle ( (WS_VISIBLE      |
                                             SLS_RIBBONSTRIP |
                                             SLS_READONLY),
                                                        (ISLS_HORIZONTAL |
                                                         ISLS_CENTER     |
                                                         ISLS_HOMELEFT   |
                                                         ISLS_PRIMARYSCALE1 |
                                                         IWS_BORDER3D) );
#pragma data_seg()

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| DefaultHandler                                                               |
|                                                                              |
| This class wrappers an object pointer and destroys the pointed               |
| to object when the wrapper is destructed.                                    |
------------------------------------------------------------------------------*/
static struct DefaultHandler {
  operator INativeSliderHandler* ();
INativeSliderHandler
 *operator->()
    {
    return *this;
    }
  ~DefaultHandler ( )
    {
    if (ptr)
      {
      delete ptr;
      ptr = 0;
      }
    }
INativeSliderHandler
 *ptr;
} defaultHandler;

DefaultHandler::operator INativeSliderHandler* ()
{
  if ( !ptr )
     ptr = new INativeSliderHandler;
  return ptr;
}
#endif  // IC_WIN

/*------------------------------------------------------------------------------
| IProgressIndicator::IProgressIndicator                                       |
------------------------------------------------------------------------------*/
IProgressIndicator :: IProgressIndicator ( unsigned long     windowId,
                                           IWindow*          parent,
                                           IWindow*          owner,
                                           const IRectangle& initial,
                                           unsigned long     numberOfTicks,
                                           unsigned long     tickSpacing,
                                           const Style&      style )
{
   IMODTRACE_DEVELOP("IProgressIndicator::ctor1");
   IASSERTPARM(parent!=0);             //Assertions on input parms

   Style readonlyStyle(style);
   readonlyStyle |= SLS_READONLY;

   // Save the extended style to make sure we have a copy of it stored
   setExtendedStyle( extendedStyle() |
                     style.asExtendedUnsignedLong() |
                     readonlyStyle.asExtendedUnsignedLong() );

#ifdef IC_WIN
   // If this is a native Windows slider, we have to save the snap to tick
   // style as an extended style.  This style is implemented with the native
   // trackbar by setting the max range to SHRT_MAX and interpreting tick
   // positions accordingly.  We also have to save the alignment and home
   // position as extended styles.
   if (!isPMCompatible())
   {
      unsigned long nativeExtended = 0;

      nativeExtended |= (style & snapToTickMark) ? ISLS_SNAPTOTICK : 0;

      nativeExtended |= (style & alignTop || style & alignRight)
                        ? ISLS_TOP
                      : ((style & alignBottom || style & alignLeft)
                        ? ISLS_BOTTOM : 0);

      nativeExtended |= (style & homeTop || style & homeRight)
                        ? ISLS_HOMETOP : 0;

      setExtendedStyle( extendedStyle() | nativeExtended );
   }
#endif

   SLDCDATA sldcData;                  //Slider control data
   sldcData.cbSize = sizeof(SLDCDATA);
   if (style & primaryScale2)
   {
      sldcData.usScale2Increments = (unsigned short)numberOfTicks;
      sldcData.usScale2Spacing = (unsigned short)tickSpacing;
      sldcData.usScale1Increments = (unsigned short)0;
      sldcData.usScale1Spacing = (unsigned short)0;
   }
   else
   {
      sldcData.usScale1Increments = (unsigned short)numberOfTicks;
      sldcData.usScale1Spacing = (unsigned short)tickSpacing;
      sldcData.usScale2Increments = (unsigned short)0;
      sldcData.usScale2Spacing = (unsigned short)0;
   }
   unsigned long hwndOwner = 0;
   if (owner != 0)
#ifdef IC_WIN
      hwndOwner = (unsigned long)(void *)owner->handle();

   initialize(windowId, (unsigned long)(void *)parent->handle(), hwndOwner,
              convertToGUIStyle( readonlyStyle ), initial,
              (void*)&sldcData);

#else
      hwndOwner = owner->handle();

   initialize(windowId, parent->handle(), hwndOwner,
              convertToGUIStyle( readonlyStyle ), initial,
              (void*)&sldcData);

#endif
}


/*------------------------------------------------------------------------------
| IProgressIndicator::IProgressIndicator                                       |
------------------------------------------------------------------------------*/
IProgressIndicator :: IProgressIndicator ( unsigned long   windowId,
                                           IWindow*        parent,
                                           IWindow*        owner,
                                           const IRectangle& initial,
                                           unsigned long   scale1NumberOfTicks,
                                           unsigned long   scale1TickSpacing,
                                           unsigned long   scale2NumberOfTicks,
                                           unsigned long   scale2TickSpacing,
                                           const Style&    style )
{
   IMODTRACE_DEVELOP("IProgressIndicator::ctor2");
   IASSERTPARM(parent!=0);             //Assertions on input parms

   Style readonlyStyle(style);
   readonlyStyle |= SLS_READONLY;

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() |
                    style.asExtendedUnsignedLong() |
                    readonlyStyle.asExtendedUnsignedLong() );

#ifdef IC_WIN
   // If this is a native Windows slider, we have to save the snap to tick
   // style as an extended style.  This style is implemented with the native
   // trackbar by setting the max range to SHRT_MAX and interpreting tick
   // positions accordingly.  We also have to save the alignment and home
   // position as extended styles.
   if (!isPMCompatible())
   {
      unsigned long nativeExtended = 0;

      nativeExtended |= (style & snapToTickMark) ? ISLS_SNAPTOTICK : 0;

      nativeExtended |= (style & alignTop || style & alignRight)
                        ? ISLS_TOP
                      : ((style & alignBottom || style & alignLeft)
                        ? ISLS_BOTTOM : 0);

      nativeExtended |= (style & homeTop || style & homeRight)
                        ? ISLS_HOMETOP : 0;

      setExtendedStyle( extendedStyle() | nativeExtended );
   }
#endif

   SLDCDATA sldcData;                  //Slider control data
   sldcData.cbSize = sizeof(SLDCDATA);
   sldcData.usScale1Increments = (unsigned short)scale1NumberOfTicks;
   sldcData.usScale2Increments = (unsigned short)scale2NumberOfTicks;
   sldcData.usScale1Spacing = (unsigned short)scale1TickSpacing;
   sldcData.usScale2Spacing = (unsigned short)scale2TickSpacing;
   unsigned long hwndOwner = 0;
   if (owner != 0)
#ifdef IC_WIN
      hwndOwner = (unsigned long)(void *)owner->handle();
   initialize(windowId, (unsigned long)(void *)parent->handle(), hwndOwner,
              convertToGUIStyle( readonlyStyle ), initial,
              (void*)&sldcData);
#else
      hwndOwner = owner->handle();
   initialize(windowId, parent->handle(), hwndOwner,
              convertToGUIStyle( readonlyStyle ), initial,
              (void*)&sldcData);
#endif
}


/*------------------------------------------------------------------------------
| IProgressIndicator::IProgressIndicator                                       |
------------------------------------------------------------------------------*/
IProgressIndicator :: IProgressIndicator( unsigned long ulId,
                                          IWindow* pwndParent )
{
  IMODTRACE_DEVELOP("IProgressIndicator::ctor3");
  IASSERTPARM(pwndParent!=0);

#ifdef IC_WIN
  // Make sure this is a CCL control and set the pmCompatible style
  // to indicate so.
  IWindowClassName className( IWINDOWFROMID( pwndParent, ulId ));
  if (className == WC_SLIDER)
     setExtendedStyle( extendedStyle()
                       | pmCompatible.asExtendedUnsignedLong()
                       | SLS_READONLY );
  else
     ITHROWLIBRARYERROR( IC_WRONG_CONTROL_TYPE,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
#endif

  setAutoDestroyWindow(false);
  startHandlingEventsFor(ulId, pwndParent);
#ifndef IC_WIN
  setBackgroundColor(defaultbackgroundcolor(handle()));
#endif

}

/*------------------------------------------------------------------------------
| IProgressIndicator::IProgressIndicator                                       |
|                                                                              |
| Constructor from an existing progress indicator.                             |
------------------------------------------------------------------------------*/
IProgressIndicator :: IProgressIndicator( const IWindowHandle& wh)
{
#ifdef IC_WIN
  // Make sure this is a CCL control and set the pmCompatible style
  // to indicate so.
  IWindowClassName className( wh );
  if (className == WC_SLIDER)
     setExtendedStyle( extendedStyle()
                       | pmCompatible.asExtendedUnsignedLong()
                       | SLS_READONLY );
  else
     ITHROWLIBRARYERROR( IC_WRONG_CONTROL_TYPE,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
#endif

  setAutoDestroyWindow(false);
  startHandlingEventsFor(wh);
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: ProgressIndicator                                      |
------------------------------------------------------------------------------*/
IProgressIndicator :: IProgressIndicator ( )
{}


/*------------------------------------------------------------------------------
| IProgressIndicator::~IProgressIndicator                                      |
------------------------------------------------------------------------------*/
IProgressIndicator :: ~IProgressIndicator()
{
  IMODTRACE_DEVELOP("IProgressIndicator::dtor");

#ifdef IC_WIN
  this->removeHandler( defaultHandler );

  if (!isPMCompatible())
  {
     fProgressIndicatorData->destroyNativeSlider();
     delete fProgressIndicatorData;
  }
#endif
}


/*------------------------------------------------------------------------------
| IProgressIndicator::defaultStyle                                             |
|                                                                              |
| Return the default style for new progress indicator objects.                 |
------------------------------------------------------------------------------*/
IProgressIndicator::Style  IProgressIndicator :: defaultStyle ( )
{
   IMODTRACE_DEVELOP("IProgressIndicator::defaultStyle");

   return currentDefaultStyle;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IProgressIndicator::setDefaultStyle                                          |
|                                                                              |
| Sets the default style for new progress indicator objects.                   |
------------------------------------------------------------------------------*/
void IProgressIndicator::setDefaultStyle( const IProgressIndicator::Style& style )
{
   currentDefaultStyle = style;
}
#endif

/*------------------------------------------------------------------------------
| IProgressIndicator::convertToGUIStyle                                        |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IProgressIndicator :: convertToGUIStyle(const IBitFlag& guiStyle,
                                                      Boolean bExtOnly) const
{
  IMODTRACE_DEVELOP("IProgressIndicator::convertToGUIStyle");

  // Obtain the style from the class (IControl) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= guiStyle.asExtendedUnsignedLong() & ISLS_EXTGUIMASK;
  }
  else
  {
#ifdef IC_WIN
    if (extendedStyle() & ISLS_PMCOMPATIBILITY)
    {
#endif
    // SLS_ styles have a one-to-one correspondence to our style bits, and
    // inhabit the lower word of the base GUI style.  Therefore, obtain
    // that portion asis, masking out the upper word. Five of the styles
    // were denoted as 0 (OS/2 PM), due to the lack of available bits.
    // We are representing those as extended bits so we or the developer
    // can accurately test for them later.  Since their native values are 0,
    // we do not have to do any additional processing here.  Note that
    // since this class only deals with shaft related styles, we only
    // mask those out here.
    ulStyle |= guiStyle.asUnsignedLong() & ISLS_SHAFTMASK;
#ifdef IC_WIN
    }
    else
    {
       // Since we are using the CCL control SLS_ values, we need to translate
       // for the Windows native trackbar.

       unsigned long PMStyle = guiStyle.asUnsignedLong() & ISLS_SHAFTMASK;

       ulStyle |= (PMStyle & SLS_VERTICAL) ?
             ((PMStyle & SLS_PRIMARYSCALE2) ?
                   ( (TBS_VERT | TBS_LEFT) ) :
                   ( (TBS_VERT | TBS_RIGHT) )) :
             ((PMStyle & SLS_PRIMARYSCALE2) ?
                   ( (TBS_HORZ | TBS_BOTTOM) ) :
                   ( (TBS_HORZ | TBS_TOP) ));

       ulStyle |= (PMStyle & SLS_READONLY) ? TBS_NOTHUMB : 0;

       ulStyle |= (PMStyle & SLS_RIBBONSTRIP) ? TBS_ENABLESELRANGE : 0;
    }

    ulStyle |= WS_CHILD;
#endif
  }

  return( ulStyle );
}


/*------------------------------------------------------------------------------
| IProgressIndicator::tickPosition                                             |
|                                                                              |
| Return the pixel position of the tick at the specified index.  Ticks are     |
| numbered starting with zero.                                                 |
------------------------------------------------------------------------------*/
IPoint IProgressIndicator :: tickPosition ( unsigned long ulTickNum ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::tickPosition");

   IEventResult evtTickData;
#ifdef IC_WIN
   if (!isPMCompatible())
   {
      return IPoint();
   }
   else
   {
#endif
   evtTickData = handle().sendEvent(SLM_QUERYTICKPOS,
                                 IEventParameter1(ulTickNum),
                                 IEventParameter2((unsigned long)0));

   if (evtTickData.number1() == 0 || evtTickData.number2() == 0)
      ITHROWGUIERROR2("SLM_QUERYTICKPOS",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);
#ifdef IC_WIN
   }
#endif

   IPoint pt((unsigned short)evtTickData.number1(),
             (unsigned short)evtTickData.number2());

   // Always returns end of tick closest to shaft
   return ICoordinateSystem::convertToApplication( pt, this->size() );
}


/*------------------------------------------------------------------------------
| IProgressIndicator::tickLength                                               |
|                                                                              |
| Return the length of the tick at the specified index, in pixels.             |
------------------------------------------------------------------------------*/
unsigned long IProgressIndicator :: tickLength ( unsigned long ulTickNum ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::tickLength");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      Scale primary = primaryScale();
      IASSERTPARM( ulTickNum <
            fProgressIndicatorData->numberTicks[primary]);

      if (fProgressIndicatorData->visibleTicks[primary].contains( ulTickNum )
          || ulTickNum == 0
          || ulTickNum == fProgressIndicatorData->numberTicks[primary]-1)
         return (1);
      else
         return (0);
   }
   else
   {
#endif
   IEventResult evtTickData =
           handle().sendEvent(SLM_QUERYTICKSIZE,
                              IEventParameter1(ulTickNum),
                              IEventParameter2((unsigned long)0));

   if (evtTickData.number1() == (USHORT)SLDERR_INVALID_PARAMETERS)
      ITHROWGUIERROR2("SLM_QUERYTICKSIZE",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);

   return evtTickData.number1();
#ifdef IC_WIN
   }
#endif

}

/*------------------------------------------------------------------------------
| IProgressIndicator::tickSpacing                                              |
|                                                                              |
| Returns the number of pixels between ticks for the specified scale.          |
------------------------------------------------------------------------------*/
unsigned long IProgressIndicator :: tickSpacing  ( Scale scale ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::tickSpacing");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      if (primaryScale() == scale
          && fProgressIndicatorData->tickSpacing[ scale ] == 0)
         return ((this->armRange()-1) / (this->numberOfTicks( scale )-1));
      else
         return fProgressIndicatorData->tickSpacing[ scale ];
   }
   else
   {
      // Query Slider control for the tick spacing information
      unsigned long ulTickSpacing =
            handle().sendEvent( SLM_QUERYSLIDERINFO,
                                IEventParameter1(SMA_SLIDERSCALEINFO,
                                                 SMA_SLIDERTICKSPACING),
                                IEventParameter2((unsigned long)0));

      // Check if it return an error (or the correct information
      if (ulTickSpacing == (ULONG)SLDERR_INVALID_PARAMETERS)
        ITHROWLIBRARYERROR( IC_SLDR_GET_CTLDATA,
                            IBaseErrorInfo::accessError,
                            IException::recoverable);

      // Return correct information based on scale requested
      if (scale == scale1)
        return ( LOWORD(ulTickSpacing) );
      else
        return ( HIWORD(ulTickSpacing) );
   }
#else
   WNDPARAMS wndParams;
   SLDCDATA sldcData;

   // Set up buffer with query information
   wndParams.fsStatus = WPM_CTLDATA | WPM_CBCTLDATA;
   wndParams.cbCtlData = sizeof( SLDCDATA );
   wndParams.pCtlData = (PVOID)&sldcData;

   unsigned long ulSuccess =
           handle().sendEvent(WM_QUERYWINDOWPARAMS,
                              IEventParameter1(&wndParams),
                              IEventParameter2((unsigned long)0));

   if (!ulSuccess)
      ITHROWLIBRARYERROR(IC_SLDR_GET_CTLDATA,
                         IBaseErrorInfo::accessError,
                         IException::recoverable);

   if (scale == scale1)
      return ((PSLDCDATA)(wndParams.pCtlData))->usScale1Spacing;
   else
      return ((PSLDCDATA)(wndParams.pCtlData))->usScale2Spacing;
#endif
}

/*------------------------------------------------------------------------------
| IProgressIndicator::numberOfTicks                                            |
|                                                                              |
| Returns the number of ticks for the specified scale.                         |
------------------------------------------------------------------------------*/
unsigned long IProgressIndicator :: numberOfTicks ( Scale scale ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::numberOfTicks");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      // Since native Windows trackbar only supports one scale, private data
      // is used to store the number of ticks per scale.
      return fProgressIndicatorData->numberTicks [scale];
   }
   else
   {
      // Query Slider control for the tick spacing information
      unsigned long ulTickCount =
           handle().sendEvent( SLM_QUERYSLIDERINFO,
                               IEventParameter1(SMA_SLIDERSCALEINFO,
                                                SMA_SLIDERTICKCOUNT),
                               IEventParameter2((unsigned long)0));

      // Check if it return an error (or the correct information
      if (ulTickCount == (ULONG)SLDERR_INVALID_PARAMETERS)
        ITHROWLIBRARYERROR( IC_SLDR_GET_CTLDATA,
                            IBaseErrorInfo::accessError,
                            IException::recoverable);

      // Return correct information based on scale requested
      if (scale == scale1)
        return ( LOWORD(ulTickCount) );
      else
        return ( HIWORD(ulTickCount) );
   }
#else
   WNDPARAMS wndParams;
   SLDCDATA sldcData;

   // Set up buffer with query information
   wndParams.fsStatus = WPM_CTLDATA | WPM_CBCTLDATA;
   wndParams.cbCtlData = sizeof( SLDCDATA );
   wndParams.pCtlData = (PVOID)&sldcData;

   unsigned long ulSuccess =
           handle().sendEvent(WM_QUERYWINDOWPARAMS,
                              IEventParameter1(&wndParams),
                              IEventParameter2((unsigned long)0));

   if (!ulSuccess)
      ITHROWLIBRARYERROR(IC_SLDR_GET_CTLDATA,
                         IBaseErrorInfo::accessError,
                         IException::recoverable);
   if (scale == scale1)
      return ((PSLDCDATA)(wndParams.pCtlData))->usScale1Increments;
   else
      return ((PSLDCDATA)(wndParams.pCtlData))->usScale2Increments;
#endif
}

/*------------------------------------------------------------------------------
| IProgressIndicator::setTicks                                                 |
|                                                                              |
| Sets the number of ticks and number of pixels between ticks.                 |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: setTicks (
                                             unsigned long scale1NumberOfTicks,
                                             unsigned long scale2NumberOfTicks,
                                             unsigned long scale1TickSpacing,
                                             unsigned long scale2TickSpacing )
{
   IMODTRACE_DEVELOP("IProgressIndicator::setTicks");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      setTicks( scale1,
                scale1NumberOfTicks,
                scale1TickSpacing );
      setTicks( scale2,
                scale2NumberOfTicks,
                scale2TickSpacing );
   }
   else
   {
#endif
   SLDCDATA sldcData;

   sldcData.cbSize = sizeof(SLDCDATA);
   sldcData.usScale1Increments = (unsigned short)scale1NumberOfTicks;
   sldcData.usScale1Spacing =  (unsigned short)scale1TickSpacing;
   sldcData.usScale2Increments = (unsigned short)scale2NumberOfTicks;
   sldcData.usScale2Spacing = (unsigned short)scale2TickSpacing;
#ifdef IC_WIN
      unsigned long ulResult =
            handle().sendEvent( SLM_SETSLIDERINFO,
                                IEventParameter1( SMA_SLIDERSCALEINFO,
                                                  SMA_SLIDERTICKCOUNT),
                                IEventParameter2( &sldcData ));
      if (ulResult == SLDERR_INVALID_PARAMETERS)
         ITHROWGUIERROR2( "SLM_SETSLIDERINFO",
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);

      ulResult =
            handle().sendEvent( SLM_SETSLIDERINFO,
                                IEventParameter1( SMA_SLIDERSCALEINFO,
                                                  SMA_SLIDERTICKSPACING),
                                IEventParameter2( &sldcData ));
      if (ulResult == SLDERR_INVALID_PARAMETERS)
         ITHROWGUIERROR2( "SLM_SETSLIDERINFO",
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);
   }
#else
   WNDPARAMS wndParams;
   wndParams.fsStatus = WPM_CTLDATA;
   wndParams.cbCtlData = sizeof(SLDCDATA);
   wndParams.pCtlData = (void*)(&sldcData);

   unsigned long ulSuccess =
           handle().sendEvent(WM_SETWINDOWPARAMS,
                              IEventParameter1(&wndParams),
                              IEventParameter2((unsigned long)0));
   if (!ulSuccess)
      ITHROWLIBRARYERROR(IC_SLDR_SET_CTLDATA,
                         IBaseErrorInfo::accessError,
                         IException::recoverable);
   refresh();
#endif

   return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator::setTicks                                                 |
|                                                                              |
| Sets the number of ticks and number of pixels between ticks.                 |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: setTicks (Scale scale,
                                                    unsigned long numberOfTicks,
                                                    unsigned long tickSpacing )
{
   IMODTRACE_DEVELOP("IProgressIndicator::setTicks");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      // Make sure at least two ticks have been specified for the scale if it
      // is the primary scale.
      Scale primary = primaryScale();
      if (scale == primary && numberOfTicks < 2)
         ITHROWGUIERROR2( "SLM_SETSLIDERINFO",
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);

      Boolean primaryChanged (false);

      unsigned long prevNumberTicks = fProgressIndicatorData->numberTicks[scale];
      if (scale == primary &&
          (prevNumberTicks != numberOfTicks
           || fProgressIndicatorData->tickSpacing[scale] != tickSpacing))
         primaryChanged = true;

      fProgressIndicatorData->numberTicks [scale] = numberOfTicks;
      fProgressIndicatorData->tickSpacing [scale] = tickSpacing;

      // Adjust trackbar for the primary scale.
      if (primaryChanged)
      {
         // If the number of ticks has changed, need to adjust the tick text
         // set.  Text for non-existent ticks will be removed.
         if (!fProgressIndicatorData->tickText[scale].isEmpty()
             && ((fProgressIndicatorData->scaleFactor == 1
                    && prevNumberTicks > numberOfTicks)
                || (fProgressIndicatorData->scaleFactor == -1
                    && prevNumberTicks != numberOfTicks)))
         {
            IGList tempGList;
            IGList::Cursor textCursor( fProgressIndicatorData->tickText[scale] );
            forCursor (textCursor)
            {
               tempGList.addAsLast(
                     fProgressIndicatorData->tickText[scale].graphicAt( textCursor ));
            }
            IGList::Cursor cursor (tempGList);
            fProgressIndicatorData->tickText[scale].removeAll();
            if (fProgressIndicatorData->scaleFactor == 1)
            {
               forCursor (cursor)
               {
                  IGString* gString = (IGString*)&(tempGList.graphicAt( cursor ));
                  if (gString->id() < numberOfTicks)
                     fProgressIndicatorData->tickText[scale].addAsLast( *gString );
                  else
                     cursor.setToLast();
               }
            }
            else
            {
               for (cursor.setToLast(); cursor.isValid(); cursor.setToPrevious())
               {
                  IGString* gString = (IGString*)&(tempGList.graphicAt( cursor ));
                  unsigned long tickNum = prevNumberTicks - gString->id() - 1;
                  if (tickNum < numberOfTicks)
                  {
                     gString->setId( numberOfTicks - tickNum - 1 );
                     fProgressIndicatorData->tickText[scale].addAsFirst( *gString );
                  }
                  else
                     cursor.setToFirst();
               }
            }
         }

         fProgressIndicatorData->setupNativeScale( scale );

         // Force a relayout of the trackbar.
         fProgressIndicatorData->layoutNativeSlider();
      }
   }
   else
   {
#endif
   SLDCDATA sldcData;
   sldcData.cbSize = sizeof(SLDCDATA);
   if (scale == scale1)
   {
      sldcData.usScale1Increments = (unsigned short)numberOfTicks;
      sldcData.usScale1Spacing =  (unsigned short)tickSpacing;
   }
   else
   {
      sldcData.usScale2Increments = (unsigned short)numberOfTicks;
      sldcData.usScale2Spacing = (unsigned short)tickSpacing;
   }
#ifdef IC_WIN
      unsigned long ulResult =
            handle().sendEvent( SLM_SETSLIDERINFO,
                                IEventParameter1( SMA_SLIDERSCALEINFO,
                                                  SMA_SLIDERTICKCOUNT),
                                IEventParameter2( &sldcData ));
      if (ulResult == SLDERR_INVALID_PARAMETERS)
         ITHROWGUIERROR2( "SLM_SETSLIDERINFO",
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);

      ulResult =
            handle().sendEvent( SLM_SETSLIDERINFO,
                                IEventParameter1( SMA_SLIDERSCALEINFO,
                                                  SMA_SLIDERTICKSPACING),
                                IEventParameter2( &sldcData ));
      if (ulResult == SLDERR_INVALID_PARAMETERS)
         ITHROWGUIERROR2( "SLM_SETSLIDERINFO",
                          IBaseErrorInfo::invalidParameter,
                          IException::recoverable);
   }
#else
   WNDPARAMS wndParams;
   unsigned long ulSuccess;

  // Set up buffer with query information
   wndParams.fsStatus = WPM_CTLDATA | WPM_CBCTLDATA;
   wndParams.cbCtlData = sizeof( SLDCDATA );
   wndParams.pCtlData = (PVOID)&sldcData;
   ulSuccess = handle().sendEvent(WM_QUERYWINDOWPARAMS,
                                  IEventParameter1(&wndParams),
                                  IEventParameter2((unsigned long)0));
   if (!ulSuccess)
      ITHROWLIBRARYERROR(IC_SLDR_GET_CTLDATA,
                         IBaseErrorInfo::accessError,
                         IException::recoverable);

   sldcData.cbSize = sizeof(SLDCDATA);
   if (scale == scale1)
   {
      sldcData.usScale1Increments = (unsigned short)numberOfTicks;
      sldcData.usScale1Spacing =  (unsigned short)tickSpacing;
   }
   else
   {
      sldcData.usScale2Increments = (unsigned short)numberOfTicks;
      sldcData.usScale2Spacing = (unsigned short)tickSpacing;
   }
   wndParams.fsStatus = WPM_CTLDATA;
   wndParams.cbCtlData = sizeof(SLDCDATA);
   wndParams.pCtlData = (void*)&sldcData;

   ulSuccess = handle().sendEvent(WM_SETWINDOWPARAMS,
                                  IEventParameter1(&wndParams),
                                  IEventParameter2((unsigned long)0));
   if (!ulSuccess)
      ITHROWLIBRARYERROR(IC_SLDR_SET_CTLDATA,
                         IBaseErrorInfo::accessError,
                         IException::recoverable);

#endif
   refresh();
   return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator::setTickLength                                            |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: setTickLength (
                                                    unsigned long ulTickNum,
                                                    unsigned long ulTickLength )
{
   IMODTRACE_DEVELOP("IProgressIndicator::setTickLength");

   unsigned long ulResult;

#ifdef IC_WIN
   if (!isPMCompatible())
      // Individual tick lengths not supported with native control.  We
      // can either turn them on or off (with the exception of the first
      // and last tick marks which will always be present).
   {
      Scale primary = primaryScale();

      if ((fProgressIndicatorData->visibleTicks[primaryScale()]
                .contains( ulTickNum ) && ulTickLength)
          || (!fProgressIndicatorData->visibleTicks[primaryScale()]
                .contains( ulTickNum ) && !ulTickLength)
          || ulTickNum == 0
          || ulTickNum == fProgressIndicatorData->numberTicks[primary] - 1)
         return *this;

      unsigned long ulPageSize, ulPosition;

      ulPageSize = fProgressIndicatorData->trackbar.sendEvent(
                                       TBM_GETPAGESIZE,
                                       IEventParameter1( 0 ),
                                       IEventParameter2( 0 ));

      if (ulTickLength)
      {
         fProgressIndicatorData->visibleTicks[primary].add( ulTickNum );
         ulPosition = ulTickNum * ulPageSize * fProgressIndicatorData->scaleFactor;
         ulResult = fProgressIndicatorData->trackbar.sendEvent(
                                       TBM_SETTIC,
                                       IEventParameter1( 0 ),
                                       IEventParameter2( ulPosition ));
      }
      else
      {
         fProgressIndicatorData->visibleTicks[primary]
               .remove( ulTickNum );
         fProgressIndicatorData->trackbar.sendEvent(
                                       TBM_CLEARTICS,
                                       IEventParameter1( this->isVisible() ),
                                       IEventParameter2( 0 ));
         TickSet::Cursor cursor(
               fProgressIndicatorData->visibleTicks[primaryScale()] );
         forCursor (cursor)
         {
            ulPosition =
                  fProgressIndicatorData->visibleTicks[primaryScale()].elementAt( cursor )
                  * ulPageSize * fProgressIndicatorData->scaleFactor;
            ulResult = fProgressIndicatorData->trackbar.sendEvent(
                                    TBM_SETTIC,
                                    IEventParameter1( 0 ),
                                    IEventParameter2( ulPosition ));
         }
      }
   }
   else
#endif
   ulResult =
           handle().sendEvent(SLM_SETTICKSIZE,
                              IEventParameter1(ulTickNum, ulTickLength),
                              IEventParameter2((unsigned long)0 ));

   if (!ulResult)
   {
      ITHROWGUIERROR2("SLM_SETTICKSIZE",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);
   }

   return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator::setTickLength                                            |
|                                                                              |
| Set the size of ALL ticks (in pixels).                                       |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: setTickLength (
                                                  unsigned long ulTickLength )
{
   IMODTRACE_DEVELOP("IProgressIndicator::setTickLength (all)");

   unsigned long ulResult;

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      // Tick lengths are minimally supported with native control.
      // The first and last ticks are always visible.  The remainder of the
      // tick marks are optionally visible.  The best we can do here is
      // remove the ticks when a 0 tick length is specified for all of the
      // ticks, and show them when a non-zero tick length is specified.

      if (fProgressIndicatorData->visibleTicks[primaryScale()].isEmpty()
                 && !ulTickLength)
         return *this;

      Scale primary = primaryScale();
      if (ulTickLength)
      {
         unsigned long ulPageSize, ulPosition, ulTickNum, numSizable;

         ulPageSize = fProgressIndicatorData->trackbar.sendEvent(
                                       TBM_GETPAGESIZE,
                                       IEventParameter1( 0 ),
                                       IEventParameter2( 0 ));

         numSizable = fProgressIndicatorData->numberTicks[primary] - 2;
         for (ulTickNum = 1; ulTickNum <= numSizable; ulTickNum++)
         {
            if (!fProgressIndicatorData->visibleTicks[primary].contains( ulTickNum))
            {
               fProgressIndicatorData->visibleTicks[primary].add( ulTickNum );
               ulPosition = ulTickNum * ulPageSize
                            * fProgressIndicatorData->scaleFactor;

               ulResult = fProgressIndicatorData->trackbar.sendEvent(
                                       TBM_SETTIC,
                                       IEventParameter1( 0 ),
                                       IEventParameter2( ulPosition ));
            }
         }
      }
      else
      {
         fProgressIndicatorData->visibleTicks[primary].removeAll();
         fProgressIndicatorData->trackbar.sendEvent(
                                       TBM_CLEARTICS,
                                       IEventParameter1( this->isVisible() ),
                                       IEventParameter2( 0 ));
         ulResult = 1;
      }
   }
   else
#endif
   ulResult =
           handle().sendEvent(SLM_SETTICKSIZE,
                              IEventParameter1((unsigned long)SMA_SETALLTICKS,
                                                ulTickLength),
                              IEventParameter2((unsigned long)0 ));

   if (!ulResult)
   {
      ITHROWGUIERROR2("SLM_SETTICKSIZE",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);
   }
   return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator::setTickText                                              |
|                                                                              |
| Set the text above or text to a tick.                                        |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: setTickText ( unsigned long ulTickNum,
                                                        const char* pszTickText)
{
   IMODTRACE_DEVELOP("IProgressIndicator::setTickText");

   unsigned long ulResult;
#ifdef IC_WIN
   if (!isPMCompatible())
   {
      Scale primary = primaryScale();
      if (ulTickNum > fProgressIndicatorData->numberTicks[primaryScale()])
         ulResult = 0;
      else
      {
         unsigned long stringId;
         IString       newText (pszTickText);
         IGString*     gText = 0;

         if (fProgressIndicatorData->scaleFactor == 1)
            stringId = ulTickNum;
         else
            stringId = numberOfTicks( primary ) - ulTickNum - 1;

         if (!fProgressIndicatorData->tickText[primary].isEmpty())
            gText = (IGString*)(fProgressIndicatorData->
                  tickText[primary].lastGraphic());
         if (!gText || gText->id() < stringId)
         {
            if (newText != "")
            {
               gText = new IGString( newText );
               gText->setFont( this->font() );
               gText->setId( stringId );
               gText->setHitSelectable( false );
               fProgressIndicatorData->tickText[primary].addAsLast( *gText );
            }
         }
         else
         {
            IGList::Cursor cursor( fProgressIndicatorData->tickText[primary] );
            cursor.setToFirst();
            while (cursor.isValid())
            {
               gText = (IGString*)&(fProgressIndicatorData->tickText[primary].
                     graphicAt( cursor ));
               if (gText->id() == stringId)
               {
                  if (newText == "")
                  {
                     // Remove null strings.
                     fProgressIndicatorData->tickText[primary].removeAt( cursor );
                     delete gText;
                     gText = 0;
                  }
                  else
                  {
                     // Set new text value.
                     gText->setText( newText );
                     cursor.invalidate();
                  }
               }
               else if (gText->id() > stringId)
               {
                  if (newText != "")
                  {
                     gText = new IGString( newText );
                     gText->setFont( this->font() );
                     gText->setId( stringId );
                     gText->setHitSelectable( false );
                     fProgressIndicatorData->tickText[primary].addAsPrevious( cursor,
                                                                     *gText );
                     cursor.invalidate();
                  }
               }
               if (cursor.isValid())
                  cursor.setToNext();
            }
         }

         // Force a re-layout.
         Boolean wasVisible = this->isVisible();
         if (wasVisible)
            disableUpdate();
         fProgressIndicatorData->layoutNativeSlider();
         if (wasVisible)
         {
            enableUpdate();
            refresh();
         }
         ulResult = 1;
      }
   }
   else
   {
#endif
   ulResult = handle().sendEvent(SLM_SETSCALETEXT,
                              IEventParameter1(ulTickNum),
                              IEventParameter2((void*)pszTickText));
#ifdef IC_WIN
   }
#endif

   if (!ulResult)
   {
      ITHROWGUIERROR2("SLM_SETSCALETEXT",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);
   }

   return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator::setTickText                                              |
|                                                                              |
| Set the text above or below the tick.                                        |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: setTickText ( unsigned long ulTickNum,
                                                        const IResourceId& text)
{
   IMODTRACE_DEVELOP("IProgressIndicator::setTickText");

   IString str = text.resourceLibrary().loadString(text);
   setTickText(ulTickNum,(char*)str);

   return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator::shaftPosition                                            |
|                                                                              |
| Return the lower left point of the shaft relative to the lower left          |
| of the control window.                                                       |
------------------------------------------------------------------------------*/
IPoint IProgressIndicator :: shaftPosition ( ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::shaftPosition");

   IRectangle shaftRect;
#ifdef IC_WIN
   if (!isPMCompatible())
   {
      return IPoint();
   }
   else
   {
#endif
   IEventResult evtShaftData =
          handle().sendEvent(SLM_QUERYSLIDERINFO,
                             IEventParameter1((unsigned long)SMA_SHAFTPOSITION),
                             IEventParameter2((unsigned long)0));

   shaftRect = IRectangle(
                 IPoint((unsigned short)evtShaftData.number1(),
                        (unsigned short)evtShaftData.number2() ),
                 this->shaftSize() );
#ifdef IC_WIN
   }
#endif

   return ICoordinateSystem::convertToApplication(
             shaftRect,
             this->size() ).minXMinY() ;
}


/*------------------------------------------------------------------------------
| IProgressIndicator::shaftSize                                                |
------------------------------------------------------------------------------*/
ISize IProgressIndicator :: shaftSize ( ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::shaftSize");

   ISize sz;
#ifdef IC_WIN
   if (!isPMCompatible())
   {
      return ISize();
   }
   else
   {
#endif
   IEventResult evtShaftData =
        handle().sendEvent(SLM_QUERYSLIDERINFO,
                           IEventParameter1((unsigned long)SMA_SHAFTDIMENSIONS),
                           IEventParameter2((unsigned long)0));

   if (style() & vertical.asUnsignedLong())
   {
      sz.setWidth((unsigned short)evtShaftData.number2());
      sz.setHeight((unsigned short)evtShaftData.number1());
   }
   else
   {
      sz.setWidth((unsigned short)evtShaftData.number1());
      sz.setHeight((unsigned short)evtShaftData.number2());
   }
#ifdef IC_WIN
   }
#endif
   return sz;
}

/*------------------------------------------------------------------------------
| IProgressIndicator::setShaftPosition                                         |
|                                                                              |
| Set the position of the progress indicator / slider shaft.                   |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: setShaftPosition ( const IPoint& position )
{
   IMODTRACE_DEVELOP("IProgressIndicator::setShaftPosition");

#ifdef IC_WIN
   if (!isPMCompatible())
      return *this;
#endif
   IPoint nativePt = ICoordinateSystem::convertToNative(
                        IRectangle( position, this->shaftSize() ),
                        this->size() ).minXMinY() ;

   unsigned long ulResult =
        handle().sendEvent(SLM_SETSLIDERINFO,
                           IEventParameter1((unsigned long)SMA_SHAFTPOSITION),
                           IEventParameter2(nativePt.x(), nativePt.y()));
   if (!ulResult)
      ITHROWGUIERROR2("SLM_SETSLIDERINFO",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);

   return *this;

}

/*------------------------------------------------------------------------------
| IProgressIndicator::setShaftBreadth                                          |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: setShaftBreadth (
                                                  unsigned long ulShaftBreadth )
{
   IMODTRACE_DEVELOP("IProgressIndicator::setShaftBreadth");

#ifdef IC_WIN
   if (!isPMCompatible())
      return *this;
#endif

   unsigned long ulResult =
        handle().sendEvent(SLM_SETSLIDERINFO,
                           IEventParameter1((unsigned long)SMA_SHAFTDIMENSIONS),
                           IEventParameter2(ulShaftBreadth));
   if (!ulResult)
      ITHROWGUIERROR2("SLM_SETSLIDERINFO",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);

   return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator::armRange                                                 |
|                                                                              |
| Query the range over which the arm can be moved.                             |
------------------------------------------------------------------------------*/
unsigned long IProgressIndicator :: armRange ( ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::armRange");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      // Range here means the pixel extent of the slider.
      return (fProgressIndicatorData->pixelRange + 1);
   }
#endif

   IEventResult evtArmData =
      handle().sendEvent(SLM_QUERYSLIDERINFO,
                         IEventParameter1((unsigned long)SMA_SLIDERARMPOSITION,
                                          (unsigned long)SMA_RANGEVALUE ),
                         IEventParameter2((unsigned long)0 ));
   return evtArmData.number2();
}

/*------------------------------------------------------------------------------
| IProgressIndicator::armPixelOffset                                           |
|                                                                              |
| Get position of arm as an offset from home in pixels.                        |
------------------------------------------------------------------------------*/
unsigned long IProgressIndicator :: armPixelOffset ( ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::armPixelOffset");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      unsigned long ulArmPosition = fProgressIndicatorData->armRangeOffset();
      if (isSnapToTickEnabled())
         // Need to calculate the arm offset in pixels from home position.
         ulArmPosition *= tickSpacing (primaryScale());
      else
      {
         unsigned long armPixelRange = this->armRange() - 1;
         unsigned long trackbarRange = fProgressIndicatorData->trackbarRange();
         if (trackbarRange)
           ulArmPosition = (ulArmPosition * armPixelRange + trackbarRange/2)
                           / trackbarRange;
      }
      return ulArmPosition;
   }
   else
   {
#endif

   IEventResult evtArmData =
      handle().sendEvent(SLM_QUERYSLIDERINFO,
                         IEventParameter1((unsigned long)SMA_SLIDERARMPOSITION,
                                          (unsigned long)SMA_RANGEVALUE),
                         IEventParameter2((unsigned long)0));
   return evtArmData.number1();
#ifdef IC_WIN
   }
#endif
}

/*------------------------------------------------------------------------------
| IProgressIndicator::armTickOffset                                            |
|                                                                              |
| Return the position of the arm as a tick number.  Ticks are numbered         |
| starting at 0.                                                               |
------------------------------------------------------------------------------*/
unsigned long IProgressIndicator :: armTickOffset ( ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::armTickOffset");

#ifdef IC_WIN
   // Native trackbar.
   if (!isPMCompatible())
   {
      unsigned long ulArmPosition
           = fProgressIndicatorData->armRangeOffset();

      if (!isSnapToTickEnabled())
      {
         unsigned long ulPageSize = fProgressIndicatorData->trackbar.sendEvent(
                                          TBM_GETPAGESIZE,
                                          IEventParameter1( 0 ),
                                          IEventParameter2( 0 ));
         if (ulPageSize)
           ulArmPosition = (ulArmPosition + ulPageSize / 2) / ulPageSize;
      }

      return ulArmPosition;
   }
   else
   {
#endif

      IEventResult evtArmData =
         handle().sendEvent(SLM_QUERYSLIDERINFO,
                            IEventParameter1((unsigned long)SMA_SLIDERARMPOSITION,
                                             (unsigned long)SMA_INCREMENTVALUE),
                            IEventParameter2((unsigned long)0));

      return evtArmData.number1();
#ifdef IC_WIN
   }
#endif
}

/*------------------------------------------------------------------------------
| IProgressIndicator::moveArmToTick                                            |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: moveArmToTick (
                                                  unsigned long ulTickNum )
{
  IMODTRACE_DEVELOP("IProgressIndicator::moveArmToTick");

  /*******************************************************************/
  /* Only attempt to set it if we are attempting to set a new value. */
  /*******************************************************************/
  if (armTickOffset() != ulTickNum)
  {
    unsigned long ulResult;
#ifdef IC_WIN
    if (!isPMCompatible())
    {
       if (ulTickNum >= fProgressIndicatorData->numberTicks[primaryScale()])
          ulResult = 0;
       else
       {
          unsigned long ulNewPosition = ulTickNum;
          if (!isSnapToTickEnabled())
          {
             unsigned long ulPageSize = fProgressIndicatorData->trackbar.sendEvent(
                                          TBM_GETPAGESIZE,
                                          IEventParameter1( 0 ),
                                          IEventParameter2( 0 ));
             ulNewPosition *= ulPageSize;
          }
          fProgressIndicatorData->setArmRangeOffset( ulNewPosition );
          ulResult = 1;  // No return code.
       }
    }
    else
    {
#endif

    ulResult =
       handle().sendEvent(SLM_SETSLIDERINFO,
                IEventParameter1((unsigned long)SMA_SLIDERARMPOSITION,
                                 (unsigned long)SMA_INCREMENTVALUE),
                                 IEventParameter2(ulTickNum));
#ifdef IC_WIN
    }
#endif

#if (IC_OBSOLETE <= IC_OBSOLETE_2)
    if (ulResult)
    {
      notifyObservers(INotificationEvent(IProgressIndicator::armTickOffsetId,
                                         *this, true, (void*)ulTickNum));
    }
#endif

    if (!ulResult)
    {
      ITHROWGUIERROR2("SLM_SETSLIDERINFO",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);
    }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| Function Name: IProgressIndicator :: moveArmToPixel                          |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: moveArmToPixel (
                                                  unsigned long ulArmOffset )
{
  IMODTRACE_DEVELOP("IProgressIndicator::moveArmToPixel");

  /*******************************************************************/
  /* Only attempt to set it if we are attempting to set a new value. */
  /*******************************************************************/
  if (armPixelOffset() != ulArmOffset)
  {
    unsigned long ulResult;
#ifdef IC_WIN
    if (!isPMCompatible())
    {
       if (ulArmOffset > this->armRange()-1)
          ulResult = 0;
       else
       {
          unsigned long ulNewPosition = 0;
          if (isSnapToTickEnabled())
          {
             // Adjust position to nearest tick.
             unsigned long spacing = this->tickSpacing( primaryScale() );
             if (spacing)
               ulNewPosition = (ulArmOffset + spacing / 2) / spacing;
          }
          else
          {
             // Calculate the range offset for the specified pixel.
             unsigned long armPixelRange = this->armRange()-1,
                           trackbarRange = fProgressIndicatorData->trackbarRange();
             if (armPixelRange)
               ulNewPosition = (ulArmOffset * trackbarRange + armPixelRange/2)
                               / armPixelRange;
          }

          fProgressIndicatorData->setArmRangeOffset( ulNewPosition );
          ulResult = 1;  // No return code.
       }
    }
    else
    {
#endif
    ulResult = handle().sendEvent(SLM_SETSLIDERINFO,
                          IEventParameter1((unsigned long)SMA_SLIDERARMPOSITION,
                                           (unsigned long)SMA_RANGEVALUE),
                          IEventParameter2(ulArmOffset));
#ifdef IC_WIN
    }
#endif

#if (IC_OBSOLETE <= IC_OBSOLETE_2)
    if (ulResult)
    {
      notifyObservers(INotificationEvent(IProgressIndicator::armPixelOffsetId,
                                         *this, true, (void*)ulArmOffset));
    }
#endif

    if (!ulResult)
    {
      ITHROWGUIERROR2("SLM_SETSLIDERINFO",
                       IBaseErrorInfo::invalidParameter,
                       IException::recoverable);
    }
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator::initialize                                               |
------------------------------------------------------------------------------*/
void IProgressIndicator :: initialize ( unsigned long     windowId,
                                        unsigned long     parent,
                                        unsigned long     owner,
                                        unsigned long     ulStyle,
                                        const IRectangle& initial,
                                        void *            sliderData )
{
  IMODTRACE_DEVELOP("IProgressIndicator::initialize");

  IWindowHandle wndhSlider;

#ifdef IC_WIN
  if (isPMCompatible())
  {
     // CCL control being used.  If needed load the DLL containing the control code.
     IControlStatics::loadControlDLL();
  }

  // Native trackbar.
  if (!isPMCompatible())
  {
     IProgressIndicatorData::registerSlider();

     fProgressIndicatorData = new IProgressIndicatorData();

     wndhSlider =
       this -> create( windowId,
                       0,
                       ulStyle | WS_CLIPCHILDREN,
                       WC_NATIVESLIDER,
                       (void *)parent,
                       (void *)owner,
                       initial,
                       0,
                       0,
                       defaultOrdering(),
                       convertToGUIStyle(
                         IProgressIndicator::Style( 0, extendedStyle() ),
                         true) );

     fProgressIndicatorData->createNativeSlider( wndhSlider,
                                                 wndhSlider,
                                                 ulStyle,
                                                 extendedStyle() );

     this->addHandler( defaultHandler );

     // Want to do startHandlingEventsFor here so we can use setTicks to
     // initialize the primary scale.
     startHandlingEventsFor(wndhSlider);

     SLDCDATA *psldcData = (SLDCDATA*)sliderData;
     Scale primary = primaryScale();
     this->setTicks ( (unsigned long)psldcData->usScale1Increments,
                      (unsigned long)psldcData->usScale2Increments,
                      (unsigned long)psldcData->usScale1Spacing,
                      (unsigned long)psldcData->usScale2Spacing );

     // Set the initial state of the buttons for sliders.
     fProgressIndicatorData->setButtonState();

     // Show now if visible style was specified.  (Trackbar is not created
     // with WS_VISIBLE style in order to prevent flashing during construction).
     if (ulStyle & visible.asUnsignedLong())
        this->show();
  }
  else

  // CCL Progress Indicator.
  {
#endif

  wndhSlider =
      this -> create( windowId,
                      0,
                      ulStyle,
                      WC_SLIDER,
#ifdef IC_WIN
                      (void *)parent,
                      (void *)owner,
#else
                      parent,
                      owner,
#endif
                      initial,
                      sliderData,
                      0,
                      defaultOrdering(),
                      convertToGUIStyle(        // convert extended styles
                         IProgressIndicator::Style(0, extendedStyle() ),
                         true)     );

#ifdef IC_WIN
  this->addHandler( defaultHandler );
#endif

  startHandlingEventsFor(wndhSlider);
#ifdef IC_WIN
  }
#endif


#ifndef IC_WIN
  setBackgroundColor(defaultbackgroundcolor(wndhSlider));
#endif
}


/*------------------------------------------------------------------------------
| IProgressIndicator::calcMinimumSize                                          |
|                                                                              |
| Calculate the minimum screen size needed by the control.                     |
------------------------------------------------------------------------------*/
ISize  IProgressIndicator :: calcMinimumSize ( ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::calcMinimumSize");

   ISize sizMin;
#ifdef IC_WIN
   if (!isPMCompatible())
   {
      if (ISTYLEOF( fProgressIndicatorData->trackbar ) & TBS_VERT)
      {
         sizMin.setHeight( 100 );
         sizMin.setWidth( 34 );
      }
      else
      {
         sizMin.setHeight( 34 );
         sizMin.setWidth( 100 );
      }
   }
   else
   {
#endif

   if (style() & vertical.asUnsignedLong())
   {
      sizMin.setHeight( 100 );
      sizMin.setWidth( 30 );
   }
   else
   {
      sizMin.setHeight( 30 );
      sizMin.setWidth( 100 );
   }

#ifdef IC_WIN
   }
#endif

  if ( ( extendedStyle() & border3D.asExtendedUnsignedLong() ) &&
       ( IPlatform::isWin95() || IPlatform::isNTNewShell() ))
  {
     sizMin += ISize(4,4);
  }

   return sizMin;
}


/*------------------------------------------------------------------------------
| IProgressIndicator::backgroundColor                                          |
|                                                                              |
| Returns the background color of the progress indicator.                      |
------------------------------------------------------------------------------*/
IColor IProgressIndicator::backgroundColor () const
{
  IMODTRACE_DEVELOP("IProgressIndicator::backgroundColor");

#ifdef IC_WIN
  if (!isPMCompatible())
  {
    return (IWindow::color(PP_BACKGROUNDCOLOR,
                           IGUIColor(IGUIColor::dialogBgnd)));
  }
  else
  {
     return Inherited::backgroundColor();
  }
#else

  return (IWindow::color(PP_BACKGROUNDCOLOR,
                         IGUIColor(IGUIColor::defaultControl)));
#endif // IC_WIN
}

/*------------------------------------------------------------------------------
| IProgressIndicator::getTickText                                              |
|                                                                              |
| Return the text associated with a tick.                                      |
------------------------------------------------------------------------------*/
char* IProgressIndicator :: getTickText ( unsigned long ulTickNum,
                                          char* pszBuffer,
                                          unsigned long ulBufLength ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::getTickText");

   char* pszTemp;

   if(pszBuffer==0 || ulBufLength==0)
   {
      ulBufLength = tickTextLength(ulTickNum)+1;
      pszTemp = new char[ulBufLength];
      pszTemp[0] = 0;
      pszBuffer = pszTemp;
   }

   IEventResult evtTickData =
           handle().sendEvent(SLM_QUERYSCALETEXT,
                              IEventParameter1(ulTickNum, ulBufLength),
                              IEventParameter2((void*)pszBuffer));

   if (evtTickData.number1() == (USHORT)SLDERR_INVALID_PARAMETERS)
      ITHROWGUIERROR2("SLM_QUERYSCALETEXT",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);

   return pszBuffer;
}

/*------------------------------------------------------------------------------
| IProgressIndiactor::tickTextLength                                           |
|                                                                              |
| Return the length of the text associated with a tick.                        |
------------------------------------------------------------------------------*/
unsigned long IProgressIndicator :: tickTextLength (
                                                unsigned long ulTickNum ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::tickTextLength");

   IEventResult evtTickData =
           handle().sendEvent(SLM_QUERYSCALETEXT,
                              IEventParameter1(ulTickNum, (unsigned long)0),
                              IEventParameter2((unsigned long)0));

   if (evtTickData.number1() == (USHORT)SLDERR_INVALID_PARAMETERS)
      ITHROWGUIERROR2("SLM_QUERYSCALETEXT",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);

   return evtTickData.number1();
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: alignment                                              |
------------------------------------------------------------------------------*/
IProgressIndicator::Alignment IProgressIndicator :: alignment ( ) const
{
  IMODTRACE_DEVELOP("IProgressIndicator::alignment");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      unsigned long extendedStyle = this->extendedStyle();
      if (extendedStyle & ISLS_BOTTOM)
         return bottomLeft;
      if (extendedStyle & ISLS_TOP)
         return topRight;
   }
   else
   {
#endif

   unsigned long ulStyle = style();
   if (ulStyle & alignLeft.asUnsignedLong())
      return bottomLeft;
   if (ulStyle & alignRight.asUnsignedLong())
      return topRight;
#ifdef IC_WIN
   }
#endif

   return centered;
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: homePosition                                           |
------------------------------------------------------------------------------*/
IProgressIndicator::HomePosition IProgressIndicator :: homePosition ( ) const
{
  IMODTRACE_DEVELOP("IProgressIndicator::homePosition");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      return( (extendedStyle() & ISLS_HOMETOP) ? homeTopRight
                                               : homeBottomLeft);
   }
   else
#endif
   return( (style() & homeTop.asUnsignedLong() ? homeTopRight
                                               : homeBottomLeft) );
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: setHomePosition                                        |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: setHomePosition ( HomePosition home )
{
  IMODTRACE_DEVELOP("IProgressIndicator::setHomePosition");

  unsigned long ulStyle;
#ifdef IC_WIN
  if (!isPMCompatible())
  {
     ulStyle = extendedStyle();
     if (homePosition() != home)
     {
        Boolean wasVisible = this->isVisible();
        if (wasVisible)
           disableUpdate();

        fProgressIndicatorData->trackbar.sendEvent( TBM_CLEARTICS,
                                       IEventParameter1( true ),
                                       IEventParameter2( 0 ));

        unsigned long armOffset = fProgressIndicatorData->armRangeOffset();
        Boolean isRibbon = this->isRibbonStripEnabled();
        if (isRibbon)
           fProgressIndicatorData->setArmRangeOffset( 0 );

        unsigned long ulRangeMin = fProgressIndicatorData->trackbar.sendEvent(
                                       TBM_GETRANGEMIN,
                                       IEventParameter1 ((unsigned long)0),
                                       IEventParameter2 ((unsigned long)0) );

        unsigned long ulRangeMax = fProgressIndicatorData->trackbar.sendEvent(
                                       TBM_GETRANGEMAX,
                                       IEventParameter1 ((unsigned long)0),
                                       IEventParameter2 ((unsigned long)0) );

        // Swap the current min and max ranges if the home position has changed.
        fProgressIndicatorData->trackbar.sendEvent( TBM_SETRANGE,
                                   IEventParameter1 (true),
                                   IEventParameter2 (-ulRangeMax, -ulRangeMin) );
        fProgressIndicatorData->scaleFactor *= -1;

        // Reset the position.
        if (home == homeBottomLeft)
           setExtendedStyle( ulStyle & ~ISLS_HOMETOP );
        else
           setExtendedStyle( ulStyle | ISLS_HOMETOP );

        // Tick marks need to be reset relative to new home position.
        unsigned long ulPageSize, ulPosition, ulTickNum;
        ulPageSize = fProgressIndicatorData->trackbar.sendEvent(
                                       TBM_GETPAGESIZE,
                                       IEventParameter1( 0 ),
                                       IEventParameter2( 0 ));
        Scale primary = this->primaryScale();
        TickSet::Cursor cursor( fProgressIndicatorData->visibleTicks[primary]);

        forCursor( cursor )
        {
           ulTickNum = fProgressIndicatorData->visibleTicks[primary].elementAt( cursor );
           ulPosition = ulTickNum * ulPageSize * fProgressIndicatorData->scaleFactor;
           fProgressIndicatorData->trackbar.sendEvent(
                                       TBM_SETTIC,
                                       IEventParameter1( 0 ),
                                       IEventParameter2( ulPosition ));
        }

        // Tick text needs to be redrawn relative to new home position.
        IGList::Cursor textCursor( fProgressIndicatorData->tickText[primary] );
        IGString *gText;
        unsigned long numTicks = numberOfTicks( primary );
        forCursor( textCursor )
        {
           gText = (IGString*)&(fProgressIndicatorData->tickText[primary]
                         .graphicAt( textCursor ));
           gText->setId( numTicks - gText->id() - 1 );
        }

        // Force a re-layout.
        fProgressIndicatorData->layoutNativeSlider();
        fProgressIndicatorData->setArmRangeOffset( armOffset );

        if (wasVisible)
        {
           enableUpdate();
           refresh();
        }
     }
  }
  else
  {
#endif
  ulStyle = style();
  unsigned long ulOldStyle = ulStyle;

  if (home == homeBottomLeft)
  {
    ulStyle &= ~(homeTop.asUnsignedLong());
  }
  else
  {
    ulStyle |= homeTop.asUnsignedLong();
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();
  }
#ifdef IC_WIN
  }
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: primaryScale                                           |
------------------------------------------------------------------------------*/
IProgressIndicator::Scale IProgressIndicator :: primaryScale ( ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::primaryScale");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      if (!isVertical())
         return( (ISTYLEOF( fProgressIndicatorData->trackbar ) & TBS_TOP ?
                    scale1 : scale2) );
      else
         return( (ISTYLEOF( fProgressIndicatorData->trackbar ) & TBS_LEFT ?
                    scale2 : scale1) );
   }
   else
#endif
   return( (style() & primaryScale2.asUnsignedLong()) ? scale2 : scale1 );
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: setPrimaryScale                                        |
|                                                                              |
| Replaces the primary scale style of the progress indicator.                  |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator :: setPrimaryScale ( Scale primaryScale )
{
  IMODTRACE_DEVELOP("IProgressIndicator::setPrimaryScale");

  unsigned long ulStyle;
  unsigned long ulOldStyle;

#ifdef IC_WIN
  if (!isPMCompatible())
  {
     // Since the native trackbar only supports one range and set of tick
     // positions, we need to setup for the new primary scale using cached
     // info.
     ulStyle = ISTYLEOF( fProgressIndicatorData->trackbar );
     ulOldStyle = ulStyle;
     if (!isVertical())
        if (primaryScale == scale1)
           ulStyle |= TBS_TOP;
        else
           ulStyle &= ~TBS_TOP;
     else
        if (primaryScale == scale1)
           ulStyle &= ~TBS_LEFT;
        else
           ulStyle |= TBS_LEFT;

     if (ulStyle != ulOldStyle)
     {
        Boolean wasVisible = this->isVisible();
        if (wasVisible)
           disableUpdate();
        ISETWINDOWSTYLE( fProgressIndicatorData->trackbar, ulStyle );
        fProgressIndicatorData->setupNativeScale( primaryScale );
        fProgressIndicatorData->layoutNativeSlider();
        if (wasVisible)
           enableUpdate();
     }
  }
else
  {
#endif

  ulStyle = style();
  ulOldStyle = ulStyle;
  if (primaryScale == scale1)
  {
    ulStyle &= ~(primaryScale2.asUnsignedLong());
  }
  else
  {
    ulStyle |= primaryScale2.asUnsignedLong();
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
  }
#ifdef IC_WIN
  }
#endif

  refresh();

  if (ulStyle != ulOldStyle)
    notifyObservers(INotificationEvent(IProgressIndicator::scaleId,
                                       *this, true, (void*)primaryScale));

  return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: isVertical                                             |
------------------------------------------------------------------------------*/
IBase::Boolean IProgressIndicator :: isVertical ( ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::isVertical");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      return( (ISTYLEOF( fProgressIndicatorData->trackbar ) & TBS_VERT) ?
                     true : false );
   }
   else
#endif
   return( (style() & vertical.asUnsignedLong()) ? true : false );
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: isRibbonStripEnabled                                   |
------------------------------------------------------------------------------*/
IBase::Boolean IProgressIndicator :: isRibbonStripEnabled ( ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::isRibbonStripEnabled");

#ifdef IC_WIN
   if (!isPMCompatible())
      return( (ISTYLEOF( fProgressIndicatorData->trackbar ) &
                     TBS_ENABLESELRANGE) ? true : false );
   else
#endif
   return( (style() & ribbonStrip.asUnsignedLong()) ? true : false );
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: isPMCompatible                                         |
------------------------------------------------------------------------------*/
IBase::Boolean IProgressIndicator :: isPMCompatible ( ) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::isPMCompatible");

#ifdef IC_WIN
   return ((extendedStyle() & pmCompatible.asExtendedUnsignedLong()) ?
             true : false );
#else
   return true;
#endif
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: enableRibbonStrip                                      |
|                                                                              |
| Sets the style to be ribbon-strip or not depending on the                    |
| input flag.                                                                  |
------------------------------------------------------------------------------*/
IProgressIndicator&  IProgressIndicator :: enableRibbonStrip ( Boolean ribbon)
{
  IMODTRACE_DEVELOP("IProgressIndicator::enableRibbonStrip");

  unsigned long ulStyle;      // get the current style
  unsigned long ulOldStyle;   // save it

#ifdef IC_WIN
  if (!isPMCompatible())
  {
     ulStyle = ISTYLEOF( fProgressIndicatorData->trackbar );
     ulOldStyle = ulStyle;

     if (ribbon)
        ulStyle |= TBS_ENABLESELRANGE;
     else
        ulStyle &= ~TBS_ENABLESELRANGE;

     if (ulStyle != ulOldStyle)
     {
        if (ulStyle & WS_VISIBLE)
           disableUpdate();

        unsigned long armOffset =
              fProgressIndicatorData->armRangeOffset();

        if (!ribbon)
           fProgressIndicatorData->trackbar.sendEvent(
                                        TBM_CLEARSEL,
                                        IEventParameter1( false ),
                                        IEventParameter2( 0 ));


        ISETWINDOWSTYLE( fProgressIndicatorData->trackbar, ulStyle );

        // If this is a native trackbar, we also need to set the selection if
        // the ribbon strip was enabled.  If the ribbon strip was disabled, we
        // need to set the slider position.
        fProgressIndicatorData->setArmRangeOffset( armOffset );
        if (ulStyle & WS_VISIBLE)
        {
           enableUpdate();
           refresh();
        }
     }
  }
  else
  {
#endif
  ulStyle = style();      // get the current style
  ulOldStyle = ulStyle;   // save it

  if (ribbon)
  {
    ulStyle |= ribbonStrip.asUnsignedLong();     //add ribbon-strip style
  }
  else
  {
    ulStyle &= ~(ribbonStrip.asUnsignedLong());  //remove ribbon-strip style
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();
  }
#ifdef IC_WIN
  }
#endif

  return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: disableRibbonStrip                                     |
------------------------------------------------------------------------------*/
IProgressIndicator&  IProgressIndicator :: disableRibbonStrip ( )
{
   enableRibbonStrip( false );
   return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: isDrawItemEnabled                                      |
------------------------------------------------------------------------------*/
IBase::Boolean IProgressIndicator :: isDrawItemEnabled    ( ) const
{
  IMODTRACE_DEVELOP("IProgressIndicator::isDrawItemEnabled");

#ifdef IC_WIN
  if (!isPMCompatible())
     // Native Windows slider doesn't support owner drawing.
     return false;
#endif

   return( (style() & handleDrawItem.asUnsignedLong() ? true : false) );
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: enableDrawItem                                         |
------------------------------------------------------------------------------*/
IProgressIndicator&  IProgressIndicator :: enableDrawItem( Boolean drawItem )
{
  IMODTRACE_DEVELOP("IProgressIndicator::enableDrawItem");

#ifdef IC_WIN
  if (!isPMCompatible())
     // Native Windows slider doesn't support owner drawing.
     return *this;
#endif

  unsigned long ulStyle = style();    // get current style
  unsigned long ulOldStyle = ulStyle;   // save it

  if (drawItem)
  {
     ulStyle |= handleDrawItem.asUnsignedLong();
     // add owner draw style
  }
  else
  {
     ulStyle &= ~(handleDrawItem.asUnsignedLong());
     // remove previous draw style
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: disableDrawItem                                        |
------------------------------------------------------------------------------*/
IProgressIndicator&  IProgressIndicator :: disableDrawItem ( )
{
  IMODTRACE_DEVELOP("IProgressIndicator::disableDrawItem");

   enableDrawItem( false );
   return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: isSnapToTickEnabled                                    |
------------------------------------------------------------------------------*/
IBase::Boolean IProgressIndicator :: isSnapToTickEnabled  ( ) const
{
  IMODTRACE_DEVELOP("IProgressIndicator::isSnapToTickEnabled");

#ifdef IC_WIN
  if (!isPMCompatible())
     return( extendedStyle() & ISLS_SNAPTOTICK );
#endif

   return( (style() & snapToTickMark.asUnsignedLong() ? true : false) );
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: enableSnapToTick                                       |
|                                                                              |
| Sets the style to be snap-to-tick or not depending on the                    |
| input flag.                                                                  |
------------------------------------------------------------------------------*/
IProgressIndicator& IProgressIndicator::enableSnapToTick ( Boolean snapToTick )
{
  IMODTRACE_DEVELOP("IProgressIndicator::enableSnapToTick");

  unsigned long ulStyle;
  unsigned long ulOldStyle;

#ifdef IC_WIN
  if (!isPMCompatible())
  {
     ulStyle = extendedStyle();
     ulOldStyle = ulStyle;
     if (snapToTick)
        ulStyle |= ISLS_SNAPTOTICK;
     else
        ulStyle &= ~ISLS_SNAPTOTICK;
     if (ulStyle != ulOldStyle)
     {
        unsigned long position = armTickOffset();
        setExtendedStyle( ulStyle );
        if (ulStyle & WS_VISIBLE)
           disableUpdate();
        fProgressIndicatorData->setupNativeScale( primaryScale() );
        moveArmToTick( position );
        if (ulStyle & WS_VISIBLE)
        {
           enableUpdate();
           refresh();
        }
     }
  }
else
  {
#endif

  ulStyle = style();        // get the current style
  ulOldStyle = ulStyle;     // save it

  if (snapToTick)
  {
    ulStyle |= snapToTickMark.asUnsignedLong();
                                              //add snap-to-increment style
  }
  else
  {
    ulStyle &= ~(snapToTickMark.asUnsignedLong());
                                              //remove snap-to-increment style
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();
  }
#ifdef IC_WIN
  }
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: disableSnapToTick                                      |
------------------------------------------------------------------------------*/
IProgressIndicator&  IProgressIndicator :: disableSnapToTick ( )
{
   IMODTRACE_DEVELOP("IProgressIndicator::disableSnapToTick");

   enableSnapToTick( false );
   return *this;
}

/*------------------------------------------------------------------------------
| IProgressIndicator :: tickText                                               |
------------------------------------------------------------------------------*/
IString IProgressIndicator :: tickText(unsigned long ulTickNum) const
{
   IMODTRACE_DEVELOP("IProgressIndicator::tickText");

   IString strTickText;
#ifdef IC_WIN
   if (!isPMCompatible())
   {
      if (ulTickNum >= this->numberOfTicks(this->primaryScale()))
         ITHROWLIBRARYERROR( IC_SLDR_GET_CTLDATA,
                             IBaseErrorInfo::accessError,
                             IException::recoverable);
      strTickText = IString();
      Scale primary = primaryScale();
      if (!fProgressIndicatorData->tickText[primary].isEmpty())
      {
         unsigned long stringId;
         if (fProgressIndicatorData->scaleFactor == 1)
            stringId = ulTickNum;
         else
            stringId = numberOfTicks( primary ) - ulTickNum - 1;
         IGList::Cursor cursor( fProgressIndicatorData->tickText[primary] );
         cursor.setToFirst();
         while (cursor.isValid())
         {
            IGString *gText = (IGString*)&(fProgressIndicatorData->
                  tickText[primary].graphicAt( cursor ));
            if (gText->id() == stringId)
            {
               strTickText = gText->text();
               cursor.invalidate();
            }
            else if (gText->id() > stringId)
               cursor.invalidate();
            else
               cursor.setToNext();
         }
      }
   }
   else
   {
#endif
   char* pszTickText = getTickText(ulTickNum);

   strTickText = IString(pszTickText);
   delete(pszTickText);
#ifdef IC_WIN
   }
#endif

   return strTickText;
}


/*------------------------------------------------------------------------------
| Public slider styles                                                         |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const ISlider::Style
  ISlider::buttonsLeft       = SLS_BUTTONSLEFT,
  ISlider::buttonsRight      = SLS_BUTTONSRIGHT,
  ISlider::buttonsBottom     = SLS_BUTTONSBOTTOM,
  ISlider::buttonsTop        = SLS_BUTTONSTOP,
  ISlider::classDefaultStyle ( (WS_VISIBLE |
                                SLS_BUTTONSLEFT),
                               (ISLS_HORIZONTAL |
                                ISLS_CENTER     |
                                ISLS_HOMELEFT   |
                                ISLS_PRIMARYSCALE1 |
                                IWS_BORDER3D ) );
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new slider objects (initial value)                         |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
ISlider::Style
  ISlider::currentDefaultStyle ( (WS_VISIBLE |
                                  SLS_BUTTONSLEFT),
                                 (ISLS_HORIZONTAL |
                                  ISLS_CENTER     |
                                  ISLS_HOMELEFT   |
                                  ISLS_PRIMARYSCALE1 |
                                  IWS_BORDER3D ) );
#pragma data_seg()


/*------------------------------------------------------------------------------
| ISlider::ISlider                                                             |
|                                                                              |
| Constructor to create a scale control on a standard window.                  |
------------------------------------------------------------------------------*/
ISlider :: ISlider( unsigned long     windowId,
                    IWindow*          parent,
                    IWindow*          owner,
                    const IRectangle& initial,
                    unsigned long     numberOfTicks,
                    unsigned long     tickSpacing,
                    const Style&      style )
                 : IProgressIndicator()
{
   IMODTRACE_DEVELOP("ISlider::ctor1");
   IASSERTPARM(parent!=0);             //Assertions on input parms

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

#ifdef IC_WIN
   // If this is a native Windows slider, we have to save the snap to tick
   // style as an extended style.  This style is implemented with the native
   // trackbar by setting the max range to SHRT_MAX and interpreting tick
   // positions accordingly.  We also have to save the alignment, home
   // position, and button position as extended styles.
   if (!isPMCompatible())
   {
      unsigned long nativeExtended = 0;

      nativeExtended |= (style & snapToTickMark) ? ISLS_SNAPTOTICK : 0;

      nativeExtended |= (style & alignTop || style & alignRight)
                        ? ISLS_TOP
                      : ((style & alignBottom || style & alignLeft)
                        ? ISLS_BOTTOM : 0);

      nativeExtended |= (style & homeTop || style & homeRight)
                        ? ISLS_HOMETOP : 0;

      // If both button styles are on, then invalid so don't set native flags
      if (!((style & buttonsRight) && (style & buttonsLeft)))
      {
        nativeExtended |= (style & buttonsBottom || style & buttonsLeft)
                          ? ISLS_BUTTONSBOTTOM
                        : ((style & buttonsTop || style & buttonsRight)
                          ? ISLS_BUTTONSTOP : 0);
      }
      setExtendedStyle( extendedStyle() | nativeExtended );
   }
#endif

   SLDCDATA sldcData;                  //Slider control data
   sldcData.cbSize = sizeof(SLDCDATA);
   if (style & primaryScale2)
   {
      sldcData.usScale2Increments = (unsigned short)numberOfTicks;
      sldcData.usScale2Spacing = (unsigned short)tickSpacing;
      sldcData.usScale1Increments = (unsigned short)0;
      sldcData.usScale1Spacing = (unsigned short)0;
   }
   else
   {
      sldcData.usScale1Increments = (unsigned short)numberOfTicks;
      sldcData.usScale1Spacing = (unsigned short)tickSpacing;
      sldcData.usScale2Increments = (unsigned short)0;
      sldcData.usScale2Spacing = (unsigned short)0;
   }
   unsigned long hwndOwner = 0;
   if (owner != 0)
#ifdef IC_WIN
      hwndOwner = (unsigned long)(void *)owner->handle();

   initialize(windowId, (unsigned long)(void *)parent->handle(), hwndOwner,
              convertToGUIStyle( style ), initial, (void*)&sldcData);
#else
      hwndOwner = owner->handle();

   initialize(windowId, parent->handle(), hwndOwner,
              convertToGUIStyle( style ), initial, (void*)&sldcData);
#endif
}

/*------------------------------------------------------------------------------
| ISlider::ISlider                                                             |
|                                                                              |
| Constructor to create a scale control on a standard window.                  |
------------------------------------------------------------------------------*/
ISlider :: ISlider ( unsigned long         windowId,
                     IWindow*              parent,
                     IWindow*              owner,
                     const IRectangle&     initial,
                     unsigned long         scale1NumberOfTicks,
                     unsigned long         scale1TickSpacing,
                     unsigned long         scale2NumberOfTicks,
                     unsigned long         scale2TickSpacing,
                     const Style&          style )
                 : IProgressIndicator()
{
   IMODTRACE_DEVELOP("ISlider::ctor");
   IASSERTPARM(parent!=0);             //Assertions on input parms

   // Save the extended style to make sure we have a copy of it stored
   setExtendedStyle( extendedStyle() | style.asExtendedUnsignedLong() );

#ifdef IC_WIN
   // If this is a native Windows slider, we have to save the snap to tick
   // style as an extended style.  This style is implemented with the native
   // trackbar by setting the max range to SHRT_MAX and interpreting tick
   // positions accordingly.  We also have to save the alignment, home
   // position, and buttons position as extended styles.
   if (!isPMCompatible())
   {
      unsigned long nativeExtended = 0;

      nativeExtended |= (style & snapToTickMark) ? ISLS_SNAPTOTICK : 0;

      nativeExtended |= (style & alignTop || style & alignRight)
                        ? ISLS_TOP
                      : ((style & alignBottom || style & alignLeft)
                        ? ISLS_BOTTOM : 0);

      nativeExtended |= (style & homeTop || style & homeRight)
                        ? ISLS_HOMETOP : 0;

      // If both button styles are on, then invalid so don't set native flags
      if (!((style & buttonsRight) && (style & buttonsLeft)))
      {
        nativeExtended |= (style & buttonsBottom || style & buttonsLeft)
                          ? ISLS_BUTTONSBOTTOM
                        : ((style & buttonsTop || style & buttonsRight)
                          ? ISLS_BUTTONSTOP : 0);
      }
      setExtendedStyle( extendedStyle() | nativeExtended );
   }
#endif

   SLDCDATA sldcData;                  //Slider control data
   sldcData.cbSize = sizeof(SLDCDATA);
   sldcData.usScale1Increments = (unsigned short)scale1NumberOfTicks;
   sldcData.usScale2Increments = (unsigned short)scale2NumberOfTicks;
   sldcData.usScale1Spacing = (unsigned short)scale1TickSpacing;
   sldcData.usScale2Spacing = (unsigned short)scale2TickSpacing;

   unsigned long hwndOwner = 0;
   if (owner != 0)
#ifdef IC_WIN
      hwndOwner = (unsigned long)(void *)owner->handle();

   initialize(windowId, (unsigned long)(void *)parent->handle(), hwndOwner,
              convertToGUIStyle( style ), initial, (void*)&sldcData);
#else
      hwndOwner = owner->handle();

   initialize(windowId, parent->handle(), hwndOwner,
              convertToGUIStyle( style ), initial, (void*)&sldcData);
#endif
}

/*------------------------------------------------------------------------------
| ISlider :: ISlider                                                           |
------------------------------------------------------------------------------*/
ISlider ::  ISlider( unsigned long ulId,
                      IWindow* pwndParent)
               : IProgressIndicator(ulId, pwndParent)
{ }

/*------------------------------------------------------------------------------
| ISlider :: ISlider                                                           |
------------------------------------------------------------------------------*/
ISlider :: ISlider( const IWindowHandle& wh )
               : IProgressIndicator(wh)
{ }

/*------------------------------------------------------------------------------
| ISlider::~ISlider                                                            |
------------------------------------------------------------------------------*/
ISlider :: ~ISlider ( )
{}

/*------------------------------------------------------------------------------
| ISlider::defaultStyle                                                        |
|                                                                              |
| Return the default style for new slider objects.                             |
------------------------------------------------------------------------------*/
ISlider::Style  ISlider :: defaultStyle ( )
{
   return currentDefaultStyle;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| ISlider::setDefaultStyle                                                     |
|                                                                              |
| Sets the default style for new slider objects.                               |
------------------------------------------------------------------------------*/
void ISlider::setDefaultStyle( const ISlider::Style& style )
{
   currentDefaultStyle = style;
}
#endif

/*------------------------------------------------------------------------------
| ISlider::convertToGUIStyle                                                   |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long ISlider::convertToGUIStyle(const IBitFlag& guiStyle,
                                         Boolean bExtOnly) const
{
  IMODTRACE_DEVELOP("ISlider::convertToGUIStyle");

  // Obtain the style from the class (IProgressIndicator) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= guiStyle.asExtendedUnsignedLong() & IS_EXTMASK;
  }
  else
  {
#ifdef IC_WIN
    // Native trackbars don't support buttons, so no styles for them here.
    if (isPMCompatible())
    {
#endif
    // Note that since this class only deals with button related styles,
    // we only mask those out here.
    ulStyle |= guiStyle.asUnsignedLong() & ISLS_BUTTONMASK;
#ifdef IC_WIN
    }
#endif
  }

  return( ulStyle );
}

/*------------------------------------------------------------------------------
| ISlider::buttonsPosition                                                     |
|                                                                              |
| Returns the position of the slider's buttons.                                |
------------------------------------------------------------------------------*/
ISlider::ButtonsPosition ISlider :: buttonsPosition ( ) const
{
   IMODTRACE_DEVELOP("ISlider::buttonsPosition");

   unsigned long ulStyle;

#ifdef IC_WIN
   if (!isPMCompatible())
   {
     ulStyle = extendedStyle();
     if (ulStyle & ISLS_BUTTONSTOP)
        if (isVertical())
           return top;
        else
           return right;
     else if (ulStyle & ISLS_BUTTONSBOTTOM)
        if (isVertical())
           return bottom;
        else
           return left;
   }
   else
   {
#endif

   ulStyle = style();
   if (ulStyle & buttonsTop.asUnsignedLong())
   {
      if (ulStyle & vertical.asUnsignedLong())
         return top;
      else
         return right;
   }
   else if (ulStyle & buttonsBottom.asUnsignedLong())
   {
      if (ulStyle & vertical.asUnsignedLong())
         return bottom;
      else
         return left;
   }
#ifdef IC_WIN
   }
#endif

   return none;
}

/*------------------------------------------------------------------------------
| ISlider::detentPosition                                                      |
|                                                                              |
| Returns the offset of a detent from the home position, in pixels.            |
------------------------------------------------------------------------------*/
unsigned long ISlider :: detentPosition ( unsigned long ulDetentId ) const
{
   IMODTRACE_DEVELOP("ISlider::detentPosition");

#ifdef IC_WIN
   if (!isPMCompatible())
      return 0;
#endif

   IEventResult evtDetentData =
           handle().sendEvent(SLM_QUERYDETENTPOS,
                              IEventParameter1(ulDetentId),
                              IEventParameter2((unsigned long)0));

   if ((short)(evtDetentData.number1()) == SLDERR_INVALID_PARAMETERS)
      ITHROWGUIERROR2("SLM_QUERYDETENTPOS",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);

   return evtDetentData.number1();
}

/*------------------------------------------------------------------------------
| ISlider::addDetent                                                           |
------------------------------------------------------------------------------*/
unsigned long ISlider :: addDetent (unsigned long ulOffset )
{
   IMODTRACE_DEVELOP("ISlider::addDetent");

#ifdef IC_WIN
   if (!isPMCompatible())
      return 0;
#endif

   unsigned long ulDetentID =
           handle().sendEvent(SLM_ADDDETENT,
                              IEventParameter1(ulOffset),
                              IEventParameter2((unsigned long)0));
   if (!ulDetentID)
   {
      ITHROWGUIERROR2("SLM_ADDDETENT",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);
   }

   return ulDetentID;
}

/*------------------------------------------------------------------------------
| ISlider::removeDetent                                                        |
------------------------------------------------------------------------------*/
ISlider& ISlider :: removeDetent ( unsigned long ulDetentId )
{
   IMODTRACE_DEVELOP("ISlider::removeDetent");

#ifdef IC_WIN
   if (!isPMCompatible())
      return *this;
#endif

   unsigned long ulResult =
           handle().sendEvent(SLM_REMOVEDETENT,
                              IEventParameter1(ulDetentId),
                              IEventParameter2((unsigned long)0));
   if (!ulResult)
      ITHROWGUIERROR2("SLM_REMOVEDETENT",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);

   return *this;
}

/*------------------------------------------------------------------------------
| ISlider::armSize                                                             |
------------------------------------------------------------------------------*/
ISize ISlider :: armSize ( ) const
{
   IMODTRACE_DEVELOP("ISlider::armSize");

   ISize sz;
#ifdef IC_WIN
   if (!isPMCompatible())
   {
      RECT armRect;
      handle().sendEvent( TBM_GETTHUMBRECT,
                          IEventParameter1((unsigned long)0),
                          IEventParameter2(&armRect) );
      sz = IRectangle( armRect ).size();
   }
   else
   {
#endif

   IEventResult evtShaftData =
      handle().sendEvent(SLM_QUERYSLIDERINFO,
                      IEventParameter1((unsigned long)SMA_SLIDERARMDIMENSIONS),
                      IEventParameter2((unsigned long)0));

   if (style() & vertical.asUnsignedLong())
   {
      sz.setWidth((unsigned short)evtShaftData.number2());
      sz.setHeight((unsigned short)evtShaftData.number1());
   }
   else
   {
      sz.setWidth((unsigned short)evtShaftData.number1());
      sz.setHeight((unsigned short)evtShaftData.number2());
   }
#ifdef IC_WIN
   }
#endif

   return sz;
}

/*------------------------------------------------------------------------------
| ISlider::setArmSize                                                          |
------------------------------------------------------------------------------*/
ISlider& ISlider :: setArmSize(const ISize& sizSliderArm)
{
   IMODTRACE_DEVELOP("ISlider::setArmSize");

#ifdef IC_WIN
   if (!isPMCompatible())
   {
      // We can only set the thumb length for the native trackbar.
   }
   else
   {
#endif

   unsigned short us1;
   unsigned short us2;
   if (style() & vertical.asUnsignedLong())
   {
      us1 = sizSliderArm.height();
      us2 = sizSliderArm.width();
   }
   else
   {
      us1 = sizSliderArm.width();
      us2 = sizSliderArm.height();
   }

   unsigned long ulResult =
      handle().sendEvent(SLM_SETSLIDERINFO,
                       IEventParameter1((unsigned long)SMA_SLIDERARMDIMENSIONS),
                       IEventParameter2(us1, us2));
   if (!ulResult)
      ITHROWGUIERROR2("SLM_SETSLIDERINFO",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);
#ifdef IC_WIN
   }
#endif

   return *this;
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| ISlider::hasFocus                                                            |
|                                                                              |
| Is the window the focus window?                                              |
------------------------------------------------------------------------------*/
Boolean ISlider::hasFocus() const
{
   if (isPMCompatible())
      return Inherited::hasFocus();
   else
   {
      HWND hwndFocus = IQUERYFOCUS(HWND_DESKTOP);
      if (handle() == IPARENTOF( hwndFocus ))
      {
         IWindowClassName wcname( hwndFocus );
         if (wcname == TRACKBAR_CLASS)
            return true;
      }
      return false;
   }
}


/*------------------------------------------------------------------------------
| IProgressIndicator::setFont                                                  |
|                                                                              |
| Override to allow a native slider to traverse its internal list of IGStrings |
| and update them for the newly set font.                                      |
------------------------------------------------------------------------------*/
IWindow& IProgressIndicator::setFont( const IFont& fm )
{
  Inherited::setFont( fm );
  if (!isPMCompatible())
  {
    for (unsigned long scale = 0; scale < 2; scale++)
    {
      IGList::Cursor cursor( fProgressIndicatorData->tickText[scale] );
      IGString *gText;
      for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
      {
        gText = (IGString*)&((fProgressIndicatorData->tickText[scale].
                                                  graphicAt( cursor )));
        gText->setFont( fm );
      }
    }
  }
  return *this;
}


/*------------------------------------------------------------------------------
| IProgressIndicator::resetFont                                                |
|                                                                              |
| Override to allow a native slider to traverse its internal list of IGStrings |
| and update them for the newly reset font.                                    |
------------------------------------------------------------------------------*/
IWindow& IProgressIndicator::resetFont( )
{
  Inherited::resetFont();
  if (!isPMCompatible())
  {
    for (unsigned long scale = 0; scale < 2; scale++)
    {
      IGList::Cursor cursor( fProgressIndicatorData->tickText[scale] );
      IGString *gText;
      for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
      {
        gText = (IGString*)&((fProgressIndicatorData->tickText[scale].
                                                  graphicAt( cursor )));
        gText->removeFont();
      }
    }
  }
  return *this;
}

#endif // IC_WIN

