/*******************************************************************************
* FILE NAME: iwindow6.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions used for        *
*   notification in IWindow.                                                   *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #include <iwindefs.h>
}


#include <iwinnhdr.hpp>
#include <iwindow.hpp>
#include <iexcept.hpp>
#include <ievent.hpp>
#include <inotifev.hpp>
#include <iwinpriv.hpp>
#include <iobslist.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IWINDOW6_CPP_
  #include <ipagetun.h>
#endif

// Class to clean up static objects
class IWindowNotifyStatics
{
public:

~IWindowNotifyStatics ( )
 {
   delete windowNotifyHandler;
   windowNotifyHandler = 0;
 }
static IWindowNotifyHandler
 &notificationHandler ( );

static IWindowNotifyHandler
  *windowNotifyHandler;
};

// Initialize static objects
#pragma data_seg(ICLStaticConst)
static IWindowNotifyStatics windowStatics;
INotificationId const IWindow::sizeId                    = "IWindow::size";
INotificationId const IWindow::positionId                = "IWindow::position";
INotificationId const IWindow::visibleId                 = "IWindow::visible";
INotificationId const IWindow::enableId                  = "IWindow::enable";
INotificationId const IWindow::deleteId                  = "IWindow::delete";
INotificationId const IWindow::focusId                   = "IWindow::focus";
INotificationId const IWindow::foregroundColorId         = "IWindow::foregroundColor";
INotificationId const IWindow::backgroundColorId         = "IWindow::backgroundColor";
INotificationId const IWindow::borderColorId             = "IWindow::borderColor";
INotificationId const IWindow::shadowColorId             = "IWindow::shadowColor";
INotificationId const IWindow::disabledForegroundColorId = "IWindow::disabledForegroundColor";
INotificationId const IWindow::disabledBackgroundColorId = "IWindow::disabledBackgroundColor";
INotificationId const IWindow::activeColorId             = "IWindow::activeColor";
INotificationId const IWindow::inactiveColorId           = "IWindow::inactiveColor";
INotificationId const IWindow::hiliteForegroundColorId   = "IWindow::hiliteForegroundColor";
INotificationId const IWindow::hiliteBackgroundColorId   = "IWindow::hiliteBackgroundColor";
INotificationId const IWindow::fontId                    = "IWindow::font";
INotificationId const IWindow::commandId                 = "IWindow::command";
INotificationId const IWindow::systemCommandId           = "IWindow::systemCommand";
#pragma data_seg()

#pragma data_seg(ICLNonConst)
IWindowNotifyHandler* IWindowNotifyStatics::windowNotifyHandler = 0;
#pragma data_seg()

IWindowNotifyHandler& IWindowNotifyStatics::notificationHandler ()
{
  if(IWindowNotifyStatics::windowNotifyHandler == 0)
     IWindowNotifyStatics::windowNotifyHandler =
         new IWindowNotifyHandler();
  return *IWindowNotifyStatics::windowNotifyHandler;
}

/*------------------------------------------------------------------------------
| IWindow::enableNotification                                                  |
|                                                                              |
| Flags this part as being ready to accept notifications.  A Notifier, like    |
| IWindow, does not send notifications to observer objects until it is enabled.|
|                                                                              |
| In the window hierarchy we override this function to allocate the  correct   |
| notification handler.                                                        |
------------------------------------------------------------------------------*/
IWindow& IWindow::enableNotification( Boolean enabled )
{
  this->pWindowData->observerData().enabled = enabled;
  if(notificationHandler() == 0)
     setNotificationHandler(&(IWindowNotifyStatics::notificationHandler()));
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::disableNotification                                                 |
|                                                                              |
| Flags this part as being not ready to accept notifications.  A Notifier,     |
| like IWindow, does not send notifications to observer objects unless it is   |
| enabled                                                                      |
------------------------------------------------------------------------------*/
IWindow& IWindow::disableNotification( )
{
  this->pWindowData->observerData().enabled = false;
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::isEnabledForNotification                                            |
|                                                                              |
| Returns true if this part is enabled for notification.                       |
------------------------------------------------------------------------------*/
IBase::Boolean IWindow::isEnabledForNotification( ) const
{
  if(this->pWindowData->_observerData &&
     this->pWindowData->observerData().enabled == true)
    return true;
  else
    return false;
}


/*------------------------------------------------------------------------------
| IWindow::addObserver                                                         |
|                                                                              |
| This is the implementation of the INotifier::addObserver protocol.  We just  |
| pass requests on to the observer list.                                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::addObserver( IObserver&       anObserver,
                               const IEventData& userData)
{
  observerList().add(anObserver, userData);
  return *this;
}



/*------------------------------------------------------------------------------
| IWindow::removeObserver                                                      |
|                                                                              |
| This is the implementation of the INotifier::removeObserver protocol. We just|
| pass requests on to the observer list.                                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::removeObserver( IObserver& anObserver )
{
  observerList().remove(anObserver);
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::removeAllObservers                                                  |
|                                                                              |
| This is the implementation of the INotifier::removeAllObservers protocol. We |
| pass requests on to the observer list.                                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::removeAllObservers ( )
{
  observerList().removeAll();
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::notifyObservers                                                     |
|                                                                              |
| This is the implementation of the INotifier::notifyObservers protocol. We    |
| pass requests on to the observer list.                                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::notifyObservers( const INotificationEvent& anEvent)
{
  if(isEnabledForNotification())
    observerList().notifyObservers(anEvent);
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::notifyObservers                                                     |
|                                                                              |
| This is the implementation of the INotifier::notifyObservers protocol. We    |
| pass requests on to the observer list.                                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::notifyObservers(const INotificationId& nId)
{
  if (isEnabledForNotification())
  {
    INotificationEvent anEvent(nId, *this);
    observerList().notifyObservers(anEvent);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IWindow::observerList                                                        |
|                                                                              |
| Returns a reference to the observer list. The creation of the observer       |
| list is delayed until this function is called for the first time -- all      |
| access to the observer list MUST be obtained through this function.          |
------------------------------------------------------------------------------*/
IObserverList& IWindow::observerList( ) const
{
  if(this->pWindowData->observerData().observers==0)
    ((IWindow*)this)->pWindowData->observerData().observers =
         new IObserverList();
  return *(this->pWindowData->observerData().observers);
}

/*------------------------------------------------------------------------------
| IWindow::notificationHandler                                                 |
|                                                                              |
| Returns a pointer to the notification handler or zero if it has not been     |
| constructed.                                                                 |
------------------------------------------------------------------------------*/
IWindowNotifyHandler* IWindow::notificationHandler( ) const
{
  return this->pWindowData->observerData().notifyHandler;
}

/*------------------------------------------------------------------------------
| IWindow::setNotifcationHandler                                               |
|                                                                              |
| Sets the notification handler to the passed value.                           |
------------------------------------------------------------------------------*/
IWindow& IWindow::setNotificationHandler( IWindowNotifyHandler* handler )
{
  IASSERT(handler!=0);
  this->pWindowData->observerData().notifyHandler = handler;
  return *this;
}

/*------------------------------------------------------------------------------
| IObserverData::IObserverData                                                 |
|                                                                              |
| Constructor.                                                                 |
------------------------------------------------------------------------------*/
IObserverData::IObserverData()
     : observers(0),
       notifyHandler(0),
       enabled(false) {}

/*------------------------------------------------------------------------------
| IObserverData::~IObserverData                                                |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IObserverData::~IObserverData ()
{
  delete observers;
}

/*------------------------------------------------------------------------------
| IWindowData::observerData                                                    |
|                                                                              |
| Returns a reference to the observerData object.                              |
------------------------------------------------------------------------------*/
IObserverData& IWindowData::observerData( )
{ if (_observerData == 0)
     _observerData = new IObserverData;
  return *_observerData;
}

