/*******************************************************************************
* FILE NAME: idmtbbit.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IDMTBarButtonItem                                                        *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #include <iwindefs.h>
}

#ifndef _IDMTBBIT_
  #include <idmtbbit.hpp>
#endif

#ifndef _IDMEVENT_
  #include <idmevent.hpp>
#endif

#ifndef _IDMSRCOP_
  #include <idmsrcop.hpp>
#endif

#ifndef _IDMIMAGE_
  #include <idmimage.hpp>
#endif

#ifndef _ITBAR_
  #include <itbar.hpp>
#endif

#ifndef _ITBARBUT_
  #include <itbarbut.hpp>
#endif

#ifndef _IGBITMAP_
  #include <igbitmap.hpp>
#endif

#ifndef _ISTRING_
  #include <istring.hpp>
#endif

// tracing functions
#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

// Exception handling
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif

#include <igrafctx.hpp>
#include <igrect.hpp>
#include <igbitmap.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDMEFIT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IDMTBarButtonItem::IDMTBarButtonItem                                         |
|                                                                              |
| Constructor for toolbar button source items.                                 |
------------------------------------------------------------------------------*/
IDMTBarButtonItem::IDMTBarButtonItem( IDMSourceOperation* pDMSrcOp )  :
                          IDMItem(pDMSrcOp,
                                  IDM::toolBarButton,
                                  IDMItem::moveable,
                                  none),
                          pTargetBtn(0)
{
  IMODTRACE_DEVELOP( "IDMTBarButtonItem::ctor (source)" );

  /********************************************************************/
  /* Obtain pointer to the toolbar button from the source window      */
  /********************************************************************/
  this->pSourceBtn = (IToolBarButton *)pDMSrcOp->sourceWindow();

  /********************************************************************/
  /* Set the native rendering mechanism and format for the item.      */
  /* Add the WPS shredder RMF, unless disableDragDelete style is      */
  /* specified.                                                       */
  /********************************************************************/
  setRMFs( rmfFrom( IDM::rmLibrary, rfForThisProcess() ) );
  if (this->pSourceBtn->allowsDragDelete())
    addRMF( rmfFrom( IDM::rmDiscard, IDM::rfUnknown ) );

  /********************************************************************/
  /* Add a drag image for this item.  Use the bitmap handle from the  */
  /* toolbar button for the drag image.                               */
  /********************************************************************/
  IBitmapHandle hbitMap = this->pSourceBtn->bitmap();

  if (hbitMap)
  {
    IGBitmap        bitmap(hbitMap);
    ISize           bmpSize = bitmap.size();
    IGraphicContext gc(ISize(bmpSize.width()+5,bmpSize.height()+5));

    gc.setPenColor(IGUIColor(IGUIColor::defaultButton));
    gc.setFillColor(this->pSourceBtn->backgroundColor());
    gc.draw(IGRectangle(IRectangle(0,0,bmpSize.width()+4,bmpSize.height()+4)));

    if (this->pSourceBtn->hasTransparentColor())
      bitmap.setTransparentColor(this->pSourceBtn->transparentColor());
    bitmap.moveTo(IPoint(4,4)).drawOn(gc);

    IGBitmap* newBitmap =
         new IGBitmap(gc, IRectangle(0,0,bmpSize.width()+5,bmpSize.height()+5));

    IDMImage image = IDMImage( newBitmap->handle() );
    setImage( image );

    delete newBitmap;
  }
}

/*------------------------------------------------------------------------------
| IDMTBarButtonItem::IDMTBarButtonItem                                         |
|                                                                              |
------------------------------------------------------------------------------*/
IDMTBarButtonItem::IDMTBarButtonItem( const IDMItem::Handle& pTBarItem )  :
                          IDMItem(pTBarItem)
{
  IMODTRACE_DEVELOP( "IDMTBarBtnItem::ctor (target)" );
  this->pSourceBtn = (IToolBarButton *)sourceWindow();
  this->pTargetBtn = (IToolBarButton *)targetOperation()->targetWindow();
}

/*------------------------------------------------------------------------------
| IDMTBarButtonItem::~IDMTBarButtonItem                                        |
|                                                                              |
| Destructor                                                                   |
------------------------------------------------------------------------------*/
IDMTBarButtonItem :: ~IDMTBarButtonItem( )
{
  IMODTRACE_DEVELOP( "IDMToolBarItem::dtor" );
}

/*------------------------------------------------------------------------------
| IDMTBarButtonItem::generateSourceItems                                       |
|                                                                              |
| Gives IDMTBarButtonItem an opportunity to attach new IDMTBarButtonItem's to  |
| the argument IDMSourceOperation object.  This function is called when a drag |
| operation commences over an IDMTBarButtonItem object.  Returns true if items |
| were provided, false otherwise.                                              |
------------------------------------------------------------------------------*/
Boolean IDMTBarButtonItem ::
               generateSourceItems( IDMSourceOperation* pDMSrcOp )
{
  IMODTRACE_DEVELOP( "IDMTBarButtonItem::generateSourceItems" );

  /********************************************************************/
  /* Create the toolbar button item                                   */
  /********************************************************************/
  IDMTBarButtonItem *pTBarBtnItem = new IDMTBarButtonItem( pDMSrcOp );

  /********************************************************************/
  /* Add it to the source operation.                                  */
  /********************************************************************/
  pDMSrcOp->addItem( pTBarBtnItem );

  /********************************************************************/
  /* Set the image style.                                             */
  /********************************************************************/
  pDMSrcOp->setImageStyle( IDM::allStacked );

  /********************************************************************/
  /* Set the default operation.                                       */
  /********************************************************************/
  pDMSrcOp->setOperation( IDMOperation::move );

  /********************************************************************/
  /* Indicate that the toolbar button drag item was successfully      */
  /* created.                                                         */
  /********************************************************************/
  return( true );
}

/*------------------------------------------------------------------------------
| IDMTBarButtonItem::supportedOperationsFor                                    |
|                                                                              |
| Returns operations this item supports using the specified RMFs.              |
------------------------------------------------------------------------------*/
unsigned long IDMTBarButtonItem ::
              supportedOperationsFor( const IString& rmf ) const
{
  IMODTRACE_DEVELOP( "IDMTBarButtonItem::supportedOperationsFor" );

  /********************************************************************/
  /* If using file rendering mechanism, then only copy is supported   */
  /* Else, whatever the base class supports is used.                  */
  /********************************************************************/
  if (rmf == IDMItem::rmfFrom( IDM::rmFile, IDM::rfText ))
    return( IDMItem::copyable & supportedOperations() );
  else
    return( Inherited::supportedOperationsFor( rmf ) );
}

/*------------------------------------------------------------------------------
| IDMTBarButtonItem::targetDrop                                                |
|                                                                              |
| Sets button bitmap and text into the target toolbar button,  or moves the    |
| toolbar button to a new position within the toolbar.                         |
------------------------------------------------------------------------------*/
Boolean IDMTBarButtonItem :: targetDrop( IDMTargetDropEvent& evt )
{
  IMODTRACE_DEVELOP( "IDMTBarButtonItem::targetDrop" );

  /********************************************************************/
  /* Get the target button and the toolbar it resides in              */
  /********************************************************************/
  IToolBarButton *pTgtBtn = targetToolBarButton();
  IToolBar *pTgtTBar = targetToolBar();

  /********************************************************************/
  /* If type includes "toolBarButton" then process the move request   */
  /********************************************************************/
  if (types().includes( IDM::toolBarButton ))
  {
    /******************************************************************/
    /* Get the source button and the toolbar it resides in            */
    /******************************************************************/
    IToolBarButton *pSrcBtn = sourceToolBarButton();
    IToolBar *pSrcTBar = sourceToolBar();
    IPoint pt= IWindow::mapPoint(evt.dropPosition(),
                                 IWindow::desktopWindow()->handle(),
                                 pTgtBtn->handle());

    /******************************************************************/
    /* If we're in the same toolbar ...                               */
    /******************************************************************/
    if (pSrcTBar == pTgtTBar)
    {
      /****************************************************************/
      /* Move it after or before the button that was dropped on ...   */
      /****************************************************************/
      if ( pTgtTBar->deckOrientation() == ISetCanvas::vertical )
      {
        if ( pt.y() > (pTgtBtn->size().height()/2) )
          pTgtTBar->moveBefore( pSrcBtn, pTgtBtn );
        else
          pTgtTBar->moveAfter( pSrcBtn, pTgtBtn );
      }
      else
      {
        if ( pt.x() < (pTgtBtn->size().width()/2) )
          pTgtTBar->moveBefore( pSrcBtn, pTgtBtn );
        else
          pTgtTBar->moveAfter( pSrcBtn, pTgtBtn );
      }
    }
    else
    {
      /****************************************************************/
      /* Add it to the new toolbar after the button it was dropped on */
      /* and remove it from the old one ...                           */
      /****************************************************************/
      pSrcTBar->remove( pSrcBtn );
      pSrcBtn->setParent( pTgtTBar );
      pSrcBtn->setOwner( pTgtTBar );

      if ( pTgtTBar->deckOrientation() == ISetCanvas::vertical )
      {
        if ( pt.y() > (pTgtBtn->size().height()/2) )
          pTgtTBar->addAsPrevious( pSrcBtn, pTgtBtn );
        else
          pTgtTBar->addAsNext( pSrcBtn, pTgtBtn );
      }
      else
      {
        if ( pt.x() < (pTgtBtn->size().width()/2) )
          pTgtTBar->addAsPrevious( pSrcBtn, pTgtBtn );
        else
          pTgtTBar->addAsNext( pSrcBtn, pTgtBtn );
      }
    }

    return( true );
  } //end toolBarButton processing

  /********************************************************************/
  /* If type includes "menuItem" then process the menuitem link       */
  /* request.                                                         */
  /********************************************************************/
  if (types().includes( IDM::menuItem ))
  {
    /******************************************************************/
    /* Create the toolbar button and add it after the button that was */
    /* dropped on ...                                                 */
    /******************************************************************/
    IToolBarButton *pButton = new IToolBarButton( (unsigned long)object(),
                                                  pTgtTBar,
                                                  pTgtTBar );
    pButton->setAutoDeleteObject();
    IPoint pt= IWindow::mapPoint(evt.dropPosition(),
                                 IWindow::desktopWindow()->handle(),
                                 pTgtBtn->handle());

    if ( pTgtTBar->deckOrientation() == ISetCanvas::vertical )
    {
      if ( pt.y() > (pTgtBtn->size().height()/2) )
        pTgtTBar->addAsPrevious( pButton, pTgtBtn );
      else
        pTgtTBar->addAsNext( pButton, pTgtBtn );
    }
    else
    {
      if ( pt.x() < (pTgtBtn->size().width()/2) )
        pTgtTBar->addAsPrevious( pButton, pTgtBtn );
      else
        pTgtTBar->addAsNext( pButton, pTgtBtn );
    }

    return( true );
  } //end menuItem processing

  /********************************************************************/
  /* If type includes "bitmap" or "plainText" then process the bitmap */
  /* file copy request ...                                            */
  /********************************************************************/
  if ((types().includes( IDM::bitmap ))  ||
      (types().includes( IDM::plainText )))
  {
    /******************************************************************/
    /* Attempt to load the bitmap from a system file                  */
    /******************************************************************/
    IString fname = containerName() + sourceName();
    IGBitmap bitMap( fname );

    /******************************************************************/
    /* Set the new bitmap into the button that was dropped on ...     */
    /******************************************************************/
    pTgtBtn->setBitmap( bitMap.handle() );

    return( true );
  } //end bitmap file processing

  return( false );
}


/*------------------------------------------------------------------------------
| IDMTBarButtonItem::sourceDiscard                                             |
|                                                                              |
| Called by the source handler to allow deletion of the toolbar button after   |
| it has been dropped on the shredder.                                         |
------------------------------------------------------------------------------*/
Boolean IDMTBarButtonItem :: sourceDiscard( IDMSourceDiscardEvent& evt )
{
  IMODTRACE_DEVELOP( "IDMTBarButtonItem::sourceDiscard" );

  IToolBarButton *pTBarBtn = sourceToolBarButton();
  sourceToolBar()->remove( pTBarBtn );
  IDESTROYWINDOW( pTBarBtn->handle() );

  return( true );
}

/*------------------------------------------------------------------------------
| IDMTBarButtonItem::sourceToolBar                                                |
|                                                                              |
------------------------------------------------------------------------------*/
IToolBar* IDMTBarButtonItem :: sourceToolBar( ) const
{
  IMODTRACE_DEVELOP( "IDMTBarButtonItem::sourceToolBar" );

  return( (IToolBar *)sourceToolBarButton()->parent() );
}

/*------------------------------------------------------------------------------
| IDMTBarButtonItem::targetToolBar                                                |
|                                                                              |
------------------------------------------------------------------------------*/
IToolBar* IDMTBarButtonItem :: targetToolBar( ) const
{
  IMODTRACE_DEVELOP( "IDMTBarButtonItem::targetToolBar" );

  return( (IToolBar *)targetToolBarButton()->parent() );
}

/*------------------------------------------------------------------------------
| IDMTBarButtonItem::sourceToolBarButton                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IToolBarButton* IDMTBarButtonItem :: sourceToolBarButton( ) const
{
  IMODTRACE_DEVELOP( "IDMTBarButtonItem::sourceToolBarButton" );

  return( this->pSourceBtn );
}

/*------------------------------------------------------------------------------
| IDMTBarButtonItem::targetToolBarButton                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IToolBarButton* IDMTBarButtonItem :: targetToolBarButton( ) const
{
  IMODTRACE_DEVELOP( "IDMTBarButtonItem::targetToolBarButton" );

  return( this->pTargetBtn );
}

