/*******************************************************************************
* FILE NAME: ihelptbl.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ihelptbl.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include "ihelptbl.hpp"
#include <iapp.hpp>
#include <ireslib.hpp>
extern "C" {
  #include "iwindefs.h"
}
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif

#ifndef _ISTRING_
  #include <istring.hpp>
#endif
#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

#pragma data_seg(ICLStaticConst)
// used in parsing help table resources
const IString helpBegin("BEGIN");
const IString helpItem("HELPITEM");
const IString helpTable("HELPTABLE");
const IString helpSubTable("HELPSUBTABLE");
const IString helpSubItem("HELPSUBITEM");
const IString helpEnd("END");
const IString helpDelimiter(" ");
#pragma data_seg()

/*------------------------------------------------------------------------------
| IHelpTable::IHelpTable                                                       |
------------------------------------------------------------------------------*/
IHelpTable::IHelpTable ( IResourceId tableId ) :
   tableID (tableId.id() ),
   itemList( 10 )
{
  IFUNCTRACE_DEVELOP();
  HANDLE hHlpTblResource = 0;
   // load resources.  Exception thrown if not found
  this->hInstance = tableId.resourceLibrary().handle();

  HRSRC tempRsrc = FindResource( (HANDLE)this->hInstance,
                                 MAKEINTRESOURCE(tableId.id()),
                                 IHELPTABLERESOURCE) ;

  if (tempRsrc)
  {
    hHlpTblResource = LoadResource( (HANDLE)this->hInstance, tempRsrc );
  }

  if (!hHlpTblResource)
  {
     ITHROWGUIERROR(IString("IHelpTable LoadResource(")+
                            "IResourceId:"+IString(tableId.id())+")");
  }
  ITRACE_DEVELOP("hHlpTblResource = " + IString((unsigned long)hHlpTblResource) );

  struct helpTableResBuf
  {
    unsigned short usTopic;
    unsigned short usHelpSubTable;
    unsigned short usExtPanel;
  } *pHTResBuf;
  pHTResBuf = (struct helpTableResBuf *) LockResource( hHlpTblResource );

  int idxHT = 0;
  while ( (pHTResBuf[idxHT].usTopic != 0) &&
          (pHTResBuf[idxHT].usHelpSubTable != 0) &&
          (pHTResBuf[idxHT].usExtPanel != 0) )
  {
    addItem(pHTResBuf[idxHT].usTopic,
            pHTResBuf[idxHT].usHelpSubTable,
            pHTResBuf[idxHT].usExtPanel);
    idxHT++;
  }
  UnlockResource( hHlpTblResource );

  // this trace dumps the entire help table as read in
  ITRACE_DEVELOP( this->asString() );
}

/*------------------------------------------------------------------------------
| IHelpTable::~IHelpTable                                                      |
------------------------------------------------------------------------------*/
IHelpTable::~IHelpTable          ( )
{
}

/*------------------------------------------------------------------------------
| IHelpTable::extendedHelpForWindowID                                          |
------------------------------------------------------------------------------*/
unsigned short IHelpTable::extendedHelpForWindowID( unsigned short  frameID)
{
  IFUNCTRACE_DEVELOP();
   // default result is the same as the controlID
   unsigned short result = IHelpTable::noHelpFound;

   // first, locate the HELPITEM  in the helptable with the frameID
   IHelpTableItemList::Cursor cursor( itemList );
   Boolean  found = false;
   for (cursor.setToFirst(); (!found && cursor.isValid()); cursor.setToNext() )
   {
     // first, locate the HELPITEM  in the helptable with the frameID
     found = ( frameID == cursor.element().windowID );
     if (found)
     {
       result = cursor.element().extendedHelpID;
       break;
     }
   }

   /*****
   IHelpTableItemList::Cursor cursor( itemList );
   if ( itemList.locateElementWithKey( frameID, cursor ) )
      {
      // the current answer is the extended help panel for the frame...
      result = cursor.element().extendedHelpID;
      }
   ****/

   return result;
}

/*------------------------------------------------------------------------------
| IHelpTable::helpForWindowID                                                  |
------------------------------------------------------------------------------*/
unsigned short IHelpTable::helpForWindowID( unsigned short controlID,
                                            unsigned short frameID)
{
  IFUNCTRACE_DEVELOP();
   // default result is the same as the controlID
   unsigned short result = IHelpTable::noHelpFound;

  ITRACE_DEVELOP("Help requested on controlID " + IString(controlID) +
                 " frameID " + IString(frameID) );

   IHelpTableItemList::Cursor cursor( itemList );
   Boolean  found = false;
   for (cursor.setToFirst(); (!found && cursor.isValid()); cursor.setToNext() )
   {
     ITRACE_DEVELOP(" checking .... " + cursor.element().asString() );

     // first, locate the HELPITEM  in the helptable with the frameID
     found = ( frameID == cursor.element().windowID );
     if (found)
     {
       IHelpSubTable &subtable = cursor.element().subTable();
       IHelpSubItemList::Cursor subCursor( subtable.subItemList );
       // now, search the HELPSUBTABLE for this HELPITEM to see
       // if we have an exact match
       for (subCursor.setToFirst(); subCursor.isValid(); subCursor.setToNext() )
       {
         if ( controlID == subCursor.element().controlID )
         {
           result = subCursor.element().helpID;
           break;
         }
       }
       break;
     }
   }


   /***********
   // first, locate the HELPITEM  in the helptable with the frameID
   IHelpTableItemList::Cursor cursor( itemList );
   if ( itemList.locateElementWithKey( frameID, cursor ) )
      {

      // now, search the HELPSUBTABLE for this HELPITEM to see
      // if we have an exact match
      IHelpSubTable &subtable = cursor.element().subTable();
      IHelpSubItemList::Cursor subCursor( subtable.subItemList );
      if (subtable.subItemList.locateElementWithKey( controlID, subCursor) )
         {
         // got exact hit...
         result = subCursor.element().helpID;
         }
      }
    *****/
   return result;
}

/*------------------------------------------------------------------------------
| IHelpTable::addItem                                                          |
------------------------------------------------------------------------------*/
IHelpTable & IHelpTable::addItem   ( unsigned short id,
                                     unsigned short subtableid,
                                     unsigned short helpid)
{
  IFUNCTRACE_DEVELOP();

  // search itemList to see if subtable has already been referenced
  IHelpSubTable *subTable=0;
  IHelpTableItemList::Cursor stcursor( itemList );
  Boolean  stfound = false;
  for (stcursor.setToFirst(); (!stfound && stcursor.isValid()); stcursor.setToNext() )
  {
    stfound = ( subtableid == stcursor.element().subTable().subTableID );
    if (stfound)
    {
      subTable=&(stcursor.element().subTable());
      subTable->addRef();
    }
  }

  // if this is a new subtable then create new subtable object, and later load it.
  if (!stfound)
    subTable = new IHelpSubTable(subtableid);

  // add the helpitem to the helptable
  IHelpTableItem helpItem( id, subTable, helpid );

  ITRACE_DEVELOP(helpItem.asString());
  try
  {
    int rc = itemList.add( helpItem );
    ITRACE_DEVELOP(" add rc = " + IString(rc));
  }
  catch ( IException &excp )
  {
    ITRACE_DEVELOP( " add throw exception " +
                    IString(excp.name()) + IString(excp.text()) );
  }

  // if the subtable was loaded for a prior item, then we're done.
  if (stfound)
    return *this;

  HANDLE hHlpSubTblResource = 0;
  // construct the rest of the subtable.  Load the resources
  // needed ... exception thrown if not found

  HRSRC tempRsrc = FindResource( (HANDLE)this->hInstance,
                                 MAKEINTRESOURCE(subtableid),
                                 IHELPSUBTABLERESOURCE) ;

  if (tempRsrc)
  {
    hHlpSubTblResource = LoadResource( (HANDLE)this->hInstance, tempRsrc );
  }

  if (!hHlpSubTblResource)
  {
     ITHROWGUIERROR(IString("IHelpTable LoadResource(")+
                            "IResourceId:"+IString(subtableid)+")");
  }
  ITRACE_DEVELOP("hHlpSubTblResource = " +
                IString((unsigned long)hHlpSubTblResource) );

  struct helpSubTableResBuf
  {
    unsigned short usControlID;
    unsigned short usHelpID;
  } *pHSTResBuf;
  pHSTResBuf =
              (struct helpSubTableResBuf *) LockResource( hHlpSubTblResource );
  int idxHST = 0;
  while ( (pHSTResBuf[idxHST].usControlID != 0) &&
          (pHSTResBuf[idxHST].usHelpID != 0) )
  {
    addSubItem( subtableid,
                pHSTResBuf[idxHST].usControlID,
                pHSTResBuf[idxHST].usHelpID);
    idxHST++;
  }
  UnlockResource( hHlpSubTblResource );

  return *this;
}

/*------------------------------------------------------------------------------
| IHelpTable::addSubItem                                                       |
------------------------------------------------------------------------------*/
IHelpTable & IHelpTable::addSubItem( unsigned short subtableid,
                                     unsigned short controlid,
                                     unsigned short helpid)
{
  IFUNCTRACE_DEVELOP();
   IHelpTableItemList::Cursor cursor( itemList );
   Boolean  found = false;
   for (cursor.setToFirst(); (!found && cursor.isValid()); cursor.setToNext() )
   {
     found = ( subtableid == cursor.element().subTable().subTableID );
     if (found)
     {
       IHelpSubItem subitem( controlid, helpid );
       ITRACE_DEVELOP(subitem.asString());
       try
       {
         int rc = cursor.element().subTable().subItemList.add( subitem );
         ITRACE_DEVELOP(" add rc = " + IString(rc));
       }
       catch ( IException &excp )
       {
         ITRACE_DEVELOP( " add throw exception " +
                         IString(excp.name()) + IString(excp.text()) );
       }
     }
   }

   return *this;
}


/*------------------------------------------------------------------------------
| IHelpSubItem::asString                                                       |
------------------------------------------------------------------------------*/
IString  IHelpSubItem::asString() const
{
  IFUNCTRACE_DEVELOP();
   return helpSubItem + helpDelimiter +
          IString(controlID) + helpDelimiter + IString(helpID);
}

/*------------------------------------------------------------------------------
| IHelpSubTable::asString                                                      |
------------------------------------------------------------------------------*/
IString  IHelpSubTable::asString() const
{
  IFUNCTRACE_DEVELOP();
   IString result = helpSubTable + helpDelimiter + IString(subTableID) + IString('\n');
   result = result + helpBegin + IString('\n');
   IHelpSubItemList::Cursor cursor( subItemList );
   for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
      {
      result = result + cursor.element().asString() + IString('\n');
      }
   result = result + helpEnd;
   return result;
}

/*------------------------------------------------------------------------------
| IHelpTableItem::asString                                                     |
------------------------------------------------------------------------------*/
IString  IHelpTableItem::asString() const
{
  IFUNCTRACE_DEVELOP();
   return helpItem + helpDelimiter + IString(windowID) + helpDelimiter +
          IString( subTable().subTableID ) + helpDelimiter +
          IString( extendedHelpID );
}

/*------------------------------------------------------------------------------
| IHelpTable::asString                                                         |
------------------------------------------------------------------------------*/
IString  IHelpTable::asString() const
{
  IFUNCTRACE_DEVELOP();
   IString result = helpTable + helpDelimiter + IString(tableID) + IString('\n');
   result = result + helpBegin + IString('\n');
   IHelpTableItemList::Cursor cursor( itemList );
   for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
      {
      result = result + cursor.element().asString() + IString('\n');
      }
   result = result + helpEnd  + IString('\n');
   // extract the referenced subtables
   for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
      {
      result = result + cursor.element().subTable().asString() + IString('\n');
      }
   return result;
}
