/*******************************************************************************
* FILE NAME: imenuprv.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in imenuprv.hpp                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1996             *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
{
  #include <iwindefs.h>
}

#include <imenuprv.hpp>
#include <imenu.hpp>
#include <imnitem.hpp>
#include <iwindow.hpp>
#include <irect.hpp>
#include <ireslib.hpp>
#include <iplatfrm.hpp>
#include <iframe.hpp>
#include <iframprv.hpp>


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| findIdForPopup                                                               |
|  Local function to locate a menuhandle within a specific menu                |
------------------------------------------------------------------------------*/
unsigned long IMenuPrivate::findIdForPopup( const IMenuHandle& hBaseMenu,
                                            const IMenuHandle& hMenuToFind )
{
  unsigned long idFound = 0;
  if ((hBaseMenu) && (hMenuToFind))
  {
    //*******************************************************************
    // Begin with first item (0 based) and get count of items
    //*******************************************************************
    int itemPos = 0;
    int itemCount = GetMenuItemCount( hBaseMenu );

    //*******************************************************************
    // Look at each item in the menu to see if it has the specified id
    //*******************************************************************
    while ( itemPos < itemCount )
    {
      //*****************************************************************
      // Get id for this menu item
      //*****************************************************************
      MENUITEMINFO miItem;
      miItem.cbSize = sizeof(MENUITEMINFO);
      miItem.fMask = MIIM_STATE | MIIM_ID | MIIM_SUBMENU;
      GetMenuItemInfo( hBaseMenu, itemPos, TRUE, &miItem );

      idFound = miItem.wID;

      //*****************************************************************
      // Check for match and if found save position and exit
      //*****************************************************************
      if ( miItem.hSubMenu == hMenuToFind )
        return idFound;

      //*****************************************************************
      // Otherwise if submenu found, navigate it before moving on
      //*****************************************************************
      if ( miItem.hSubMenu )
        if (idFound = findIdForPopup( miItem.hSubMenu, hMenuToFind ))
          return idFound;

      //*****************************************************************
      // Otherwise no match so go to next menu position
      //*****************************************************************
      itemPos++;
    }
  }
  return 0;
}

/*------------------------------------------------------------------------------
| locateMenuItem                                                               |
|  Local function to locate a menu item within a specific menu                 |
------------------------------------------------------------------------------*/
Boolean IMenuPrivate::locateMenuItem( IFrameWindow*      owner,
                                      const IMenuHandle& hMenu,
                                      unsigned long      itemId,
                                      long*              pItemPos,
                                      IMenuHandle*       pMenu )
{
  // If id is 0, invalid so exit
  if (itemId)
  {
    //*******************************************************************
    // Begin with first item (0 based) and get count of items
    //*******************************************************************
    int itemPos = 0;
    int itemCount = GetMenuItemCount( hMenu );
    int idFound = 0;

    //*******************************************************************
    // Look at each item in the menu to see if it has the specified id
    //*******************************************************************
    while ( itemPos < itemCount )
    {
      //*****************************************************************
      // Depending on platform, get id differently
      //*****************************************************************
      if (!(IPlatform::isWin95() || IPlatform::isNTNewShell()))
      {
        //***************************************************************
        // Get the id of the item at the current position
        //***************************************************************
        idFound = GetMenuItemID( hMenu, itemPos );

        //***************************************************************
        // If error, item might be a submenu so query for a menu handle
        //***************************************************************
        if ( idFound == 0xFFFFFFFFul )
        {
          HMENU hPop = GetSubMenu( hMenu, itemPos );

          //*************************************************************
          // If a popup, navigate its menu before checking this item
          //*************************************************************
          if (locateMenuItem( owner, hPop, itemId, pItemPos, pMenu ))
            return true;

          //*************************************************************
          // If a menu handle is returned, then get id from lookup table
          //*************************************************************
          if (hPop)
            idFound = (int)getItemIdFromLookUpTable( owner, hPop );
        }
      }
      //*****************************************************************
      // Otherwise, use extended menu api to get id
      //*****************************************************************
      else
      {
        //***************************************************************
        // Get the info for the item at the current position
        //***************************************************************
        MENUITEMINFO miItem;
        miItem.cbSize = sizeof(MENUITEMINFO);
        miItem.fMask = MIIM_STATE | MIIM_ID | MIIM_SUBMENU;
        GetMenuItemInfo( hMenu, itemPos, TRUE, &miItem );

        //***************************************************************
        // Check if this item is a pulldown menu
        //***************************************************************
        if ( miItem.hSubMenu )
        {
          //*************************************************************
          // If a pulldown, navigate its menu before checking this item
          //*************************************************************
          if (locateMenuItem( owner, miItem.hSubMenu, itemId, pItemPos, pMenu ))
            return true;
        }
        //***************************************************************
        // If not a pulldown, or no match in pulldown, use this id
        //***************************************************************
        idFound = miItem.wID;
      }

      //*****************************************************************
      // Check for match and if found save position and exit
      //*****************************************************************
      if ( idFound == itemId )
      {
        *pItemPos = (long)itemPos;
        if (pMenu)
          *pMenu = hMenu;
        return true;
      }
      //*****************************************************************
      // Otherwise no match so go to next menu position
      //*****************************************************************
      else itemPos++;
    }
  }
  return false;
}

/*------------------------------------------------------------------------------
| locateFrame                                                                  |
|  Local function to find the IFrameWindow for an IWindow passed               |
------------------------------------------------------------------------------*/
IFrameWindow* IMenuPrivate::locateFrame( IWindow* window )
{
  IWindow* tempWindow = window;
  IWindow* newWindow;
  Boolean exit = false;
  Boolean needToFree = false;
  while (!exit)
  {
    // Query if window to check is a frame window, if so return this window
    if (tempWindow->isFrameWindow())
      exit = true;
    else
    {
      // Get the parent handle (both through IOC and the operating system
      HWND hwndParent = IPARENTOF(tempWindow->handle());
      newWindow = tempWindow->parent();

      // If the previous loop through we wrappered a window, free it
      if (needToFree)
      {
        delete tempWindow;
        needToFree = false;
      }

      // If no IWindow object for this, then potentially create one
      if ( !newWindow )
      {
        // If no handle, then no frame tables so don't create one
        if (!hwndParent)
          return 0;

        // Otherwise we have a handle, so wrapper it for now
        newWindow = new IWindow( hwndParent );
        needToFree = true;
      }
      // Copy it over for the next iteration
      tempWindow = newWindow;
    }
  }
  // If the IWindow to return was created by us, that means there is no
  // lookup table information to use, so return 0 (after freeing object)
  if (needToFree)
  {
    delete tempWindow;
    return 0;
  }

  return (IFrameWindow*)tempWindow;
}


/*------------------------------------------------------------------------------
| addToLookUpTable                                                             |
|  Local function used to build mapping table for menu handles to item ids     |
|  since this is NOT supported in Windows 3.1 or NT old shell                  |
------------------------------------------------------------------------------*/
void IMenuPrivate::addToLookUpTable( IFrameWindow*      owner,
                                     const IMenuHandle& menuHandle,
                                     unsigned long      itemId )
{
  /********************************************************************/
  /* If invalid owner, just exit (don't touch table)                  */
  /********************************************************************/
  if (!owner)
    return;

  /********************************************************************/
  /* Create a new menu lookup table entry and fill with info passed   */
  /********************************************************************/
  FMTABLE* pFmTable = new FMTABLE;
  pFmTable->fId = itemId;
  pFmTable->fHMenu = (HMENU)menuHandle;

  /********************************************************************/
  /* Default the next entry to null                                   */
  /********************************************************************/
  pFmTable->pfmtNext = 0;

  /********************************************************************/
  /* Check if there are current entries, and if so chain to list      */
  /********************************************************************/
  if (owner->fFrameWindowData->fpSubmenuList)
    pFmTable->pfmtNext = (FMTABLE*)(owner->fFrameWindowData->fpSubmenuList);

  /********************************************************************/
  /* Finally, since this is now the start of the list, save it        */
  /********************************************************************/
  owner->fFrameWindowData->fpSubmenuList = (void*)pFmTable;
  return;
}


/*------------------------------------------------------------------------------
| removeFromLookUpTable                                                        |
|  Local function used to remove item id from mapping table                    |
------------------------------------------------------------------------------*/
void IMenuPrivate::removeFromLookUpTable( IFrameWindow*      owner,
                                          const IMenuHandle& menuHandle,
                                          unsigned long      itemId )
{
  /********************************************************************/
  /* If invalid owner, just exit (don't touch table)                  */
  /********************************************************************/
  if (!owner)
    return;

  /********************************************************************/
  /* Start at top and keep searching as long as entries exist         */
  /********************************************************************/
  Boolean  bFound = false;
  FMTABLE* pFmPrev = 0;
  FMTABLE* pFmTable = (FMTABLE*)(owner->fFrameWindowData->fpSubmenuList);
  while (pFmTable)
  {
    /******************************************************************/
    /* If match is found, set found flag and stop searching           */
    /******************************************************************/
    if ((pFmTable->fId == itemId) && (pFmTable->fHMenu == (HMENU)menuHandle))
    {
      bFound = true;
      break;
    }
    /******************************************************************/
    /* Otherwise, go to next entry, but save this as previous         */
    /******************************************************************/
    pFmPrev = pFmTable;
    pFmTable = pFmTable->pfmtNext;
  }

  /********************************************************************/
  /* If the entry was found, then update list and delete item         */
  /********************************************************************/
  if (bFound)
  {
    /******************************************************************/
    /* Check for special case of first entry                          */
    /******************************************************************/
    if (pFmPrev == 0)
      owner->fFrameWindowData->fpSubmenuList = (void*)pFmTable->pfmtNext;

    /******************************************************************/
    /* otherwise remove middle entry and point previous to next       */
    /******************************************************************/
    else
      pFmPrev->pfmtNext = pFmTable->pfmtNext;

    /******************************************************************/
    /* Delete the actual item found to be removed and exit            */
    /******************************************************************/
    delete pFmTable;
  }
  return;
}


/*------------------------------------------------------------------------------
| getHMenuFromLookUpTable                                                      |
|  Local function used to retrieve the menu handle associated with a specific  |
|  item id                                                                     |
------------------------------------------------------------------------------*/
void* IMenuPrivate::getHMenuFromLookUpTable( IFrameWindow*  owner,
                                             unsigned long  itemId )
{
  /*******************************************************************/
  /* If a valid owner passed, check table for id passed              */
  /*******************************************************************/
  if (owner)
  {
    /*****************************************************************/
    /* Start at top and keep searching as long as entries exist      */
    /*****************************************************************/
    FMTABLE* pFmTable = (FMTABLE*)(owner->fFrameWindowData->fpSubmenuList);
    while (pFmTable)
    {
      /***************************************************************/
      /* If match is found, return match, otherwise go to next entry */
      /***************************************************************/
      if ( pFmTable->fId == itemId )
        return pFmTable->fHMenu;
      pFmTable = pFmTable->pfmtNext;
    }
  }
  /*******************************************************************/
  /* If no match, return null handle                                 */
  /*******************************************************************/
  return NULL;
}

/*------------------------------------------------------------------------------
| getItemIdFromLookUpTable                                                     |
|  Local function used to retrieve the item id associated with a specific      |
|  menu handle                                                                 |
------------------------------------------------------------------------------*/
long IMenuPrivate::getItemIdFromLookUpTable( IFrameWindow*      owner,
                                             const IMenuHandle& hMenu )
{
  /*******************************************************************/
  /* If a valid owner passed, check table for menuhandle passed      */
  /*******************************************************************/
  if (owner)
  {
    /*****************************************************************/
    /* Start at top and keep searching as long as entries exist      */
    /*****************************************************************/
    FMTABLE* pFmTable = (FMTABLE*)(owner->fFrameWindowData->fpSubmenuList);
    while (pFmTable)
    {
      /***************************************************************/
      /* If match is found, return match, otherwise go to next entry */
      /***************************************************************/
      if ( pFmTable->fHMenu == hMenu )
        return (long)pFmTable->fId;
      pFmTable = pFmTable->pfmtNext;
    }
  }
  /*******************************************************************/
  /* If no match, return invalid value (-1) to indicate this         */
  /*******************************************************************/
  return (long)0xFFFFFFFF;
}


/*------------------------------------------------------------------------------
| buildLookUpTableFromResourceLoad                                             |
|  Local function to prime a lookup table from the loaded menu                 |
------------------------------------------------------------------------------*/
Boolean IMenuPrivate::buildLookUpTableFromResourceLoad( IFrameWindow*      owner,
                                                 const IMenuHandle& hMenu,
                                                 const IResourceId& menuResId )
{
  HMENU hMenuParent[9];     // Array to save parent menu handle
  int   iMenuPosition[9];   // Array to save nested menu position

  // Get the module (instance) handle to load resource file from
  HMODULE hModule = menuResId.resourceLibrary().handle();

  // Find the menu resource, load it and get a pointer to the data
  HRSRC hrc = FindResource( hModule, MAKEINTRESOURCE(menuResId), RT_MENU );
  HGLOBAL hgb = LoadResource( hModule, hrc );
  LPVOID pRes = LockResource( hgb );

  // Convert pointer to dword since we need to remain on boundary
  LPDWORD pb = (LPDWORD)pRes;

  // Initialize global variables and flags needed for processing
  unsigned long menuItemId = 0;
  BOOL bPopUp, bFoundEnd;
  DWORD dwString;
  hMenuParent[1] = (HMENU)hMenu;
  iMenuPosition[1] = -1;

  // Check for correct version of menu resource (Extended Menu)
  // Version is 1 with offset 4 to start of menu template
  DWORD dwVersion = *pb;
  if (dwVersion == 0x00040001)
    pb += 2;         // If correct, move pointer to start of menu info
  else
    return false;    // otherwise exit with error

  // Start at menubar level (1) until end resets this to 0
  UINT iCascadeLevel = 1;
  UINT iNestedLevel = 1;
  while (iNestedLevel)
  {
    // Reset popup flag for each item, and bump position count by one
    bPopUp = FALSE;
    (iMenuPosition[iNestedLevel])++;

    // First move pointer past dwType & dwState, to menu id and get it
    pb += 2;
    menuItemId = (unsigned long)*pb;
    pb++;

    // Then check the next byte for the template item type
    dwString = *pb;
    if ( LOBYTE(LOWORD(dwString)) & MFR_POPUP )
    {
      // If this is a popup, get the submenu handle for this popup
      HMENU hPop = GetSubMenu( hMenuParent[iNestedLevel],
                               iMenuPosition[iNestedLevel] );

      // and add it to the lookup table for this menu
      addToLookUpTable( owner, hPop, menuItemId );

      // Now bump the nesting level up one (for new popup), set flag to
      // indicate this is a popup, save the menu handle and reset position
      iNestedLevel++;
      bPopUp = TRUE;
      hMenuParent[iNestedLevel] = hPop;
      iMenuPosition[iNestedLevel] = -1;
    }

    // If this is the end of popup (or menubar), decrement level by 1
    if ( LOBYTE(LOWORD(dwString)) & MFR_END )
    {
      if (bPopUp)
        iCascadeLevel++;
      else
      {
        iNestedLevel -= iCascadeLevel;
        iCascadeLevel = 1;
      }
    }

    // Initialize flag for looking for string end with first character compare
    if ( bFoundEnd = (HIWORD(dwString) == 0))
      pb++;

    // Loop through string until end of string
    while ( !bFoundEnd )
    {
      // If character, then move to next block to check
      pb++;

      // Get next block and check both characters for NULL
      dwString = *pb;
      if ( (LOWORD(dwString) == 0) || (HIWORD(dwString) == 0))
      {
        // If null termination, then set flag to show string end found
        // and move pointer after this block
        bFoundEnd = TRUE;
        pb++;
      }
    }
    // If popup, then skip over extra dword for help id as well.
    if (bPopUp)
      pb++;
  }
  return true;
}


/*------------------------------------------------------------------------------
| loadMenuForWin32s                                                            |
|  Local function to prime a lookup table from the loaded menu                 |
------------------------------------------------------------------------------*/
void* IMenuPrivate::loadMenuForWin32s( void*          hinstance,
                                       unsigned long  menuResId )
{
  char    menuText[IC_MENUTEXTLENGTH];
  HMENU   hMenuParent[9];     // Array to save parent menu handle
  int     iMenuPosition[9];   // Array to save nested menu position
  Boolean bClosedLevel[9];    // Array to save closed levels

  //*******************************************************************
  // Get the module (instance) handle to load resource file from
  //*******************************************************************
  HMODULE hModule = (HMODULE)hinstance;

  //*******************************************************************
  // Find the menu resource, load it and get a pointer to the data
  //*******************************************************************
  HRSRC hrc = FindResource( hModule, MAKEINTRESOURCE(menuResId), RT_MENU );
  HGLOBAL hgb = LoadResource( hModule, hrc );
  LPVOID pRes = LockResource( hgb );

  //*******************************************************************
  // Convert pointer to dword since we need to remain on boundary
  //*******************************************************************
  LPDWORD pb = (LPDWORD)pRes;

  //*******************************************************************
  // Initialize global variables and flags needed for processing
  //*******************************************************************
  unsigned long menuItemId = 0;
  BOOL bPopUp, bFoundEnd;
  DWORD dwType, dwState, dwString;

  //*******************************************************************
  // Create an empty menu for the menu bar, and set items to invalid
  // and set lowest level false to prevent trap when traversing down
  //*******************************************************************
  hMenuParent[1] = CreateMenu();
  iMenuPosition[1] = -1;
  bClosedLevel[0] = false;

  //*******************************************************************
  // Check for correct version of menu resource (Extended Menu)
  // Version is 1 with offset 4 to start of menu template
  //*******************************************************************
  DWORD dwVersion = *pb;
  if (dwVersion == 0x00040001)
    pb += 2;         // If correct, move pointer to start of menu info
  else
    return NULL;     // otherwise exit with error

  // Start at menubar level (1) until end resets this to 0
  UINT iStartLevel, fuFlags;
  UINT iNestedLevel = 1;
  while (iNestedLevel)
  {
    //*****************************************************************
    // Reset popup flag for each item, and bump position count by one
    //*****************************************************************
    HMENU hPop = NULL;
    bPopUp = FALSE;
    fuFlags = MF_BYPOSITION;
    (iMenuPosition[iNestedLevel])++;
    iStartLevel = iNestedLevel;

    //*****************************************************************
    // First, get the type and state information for the menu item
    //*****************************************************************
    dwType = *pb;
    pb++;
    dwState = *pb;
    pb++;

    //*****************************************************************
    // Next get the menu item id
    //*****************************************************************
    menuItemId = (unsigned long)*pb;
    if (!menuItemId)
      fuFlags |= MF_SEPARATOR;
    pb++;

    //*****************************************************************
    // Then check the next byte for the template item type
    //*****************************************************************
    dwString = *pb;

    //*****************************************************************
    // If the item is a popup item, then we need to build a popup
    //*****************************************************************
    if ( LOBYTE(LOWORD(dwString)) & MFR_POPUP )
    {
      //***************************************************************
      // If this is a popup, the create a popup menu for this item
      //***************************************************************
      hPop = CreatePopupMenu();

      //***************************************************************
      // Now bump the nesting level up one (for new popup), set flag to
      // indicate this is a popup, save the menu handle and reset position
      //***************************************************************
      bClosedLevel[iNestedLevel] = false;
      iNestedLevel++;
      bPopUp = TRUE;
      fuFlags |= MF_POPUP;
      hMenuParent[iNestedLevel] = hPop;
      iMenuPosition[iNestedLevel] = -1;
    }

    //*****************************************************************
    // If this is the end of popup (or menubar), decrement level by 1
    //*****************************************************************
    if ( LOBYTE(LOWORD(dwString)) & MFR_END )
    {
      //***************************************************************
      // If this item is both a popup AND an ending item, close this
      // level, deferring going down until the cascaded end occurs
      //***************************************************************
      if (bPopUp)
        bClosedLevel[iNestedLevel-1] = true;

      //***************************************************************
      // If not a popup, then check closed levels to see how far to
      // cascade back to
      //***************************************************************
      else
      {
        while ( bClosedLevel[--iNestedLevel] )
          bClosedLevel[iNestedLevel] = false;
      }
    }

    //*****************************************************************
    // Initialize flag for looking for string end with first character compare
    //*****************************************************************
    if ( bFoundEnd = (HIWORD(dwString) == 0))
    {
      pb++;
      menuText[0] = 0;
    }
    else
    {
      LPBYTE pbs = (LPBYTE)pb;
      WideCharToMultiByte( CP_ACP, 0, (LPCWSTR)(pbs+2), -1, menuText,
                           IC_MENUTEXTLENGTH, NULL, NULL );
    }

    //*****************************************************************
    // Loop through string until end of string
    //*****************************************************************
    while ( !bFoundEnd )
    {
      //***************************************************************
      // If character, then move to next block to check
      //***************************************************************
      pb++;

      //***************************************************************
      // Get next block and check both characters for NULL
      //***************************************************************
      dwString = *pb;
      if ( (LOWORD(dwString) == 0) || (HIWORD(dwString) == 0))
      {
        //*************************************************************
        // If null termination, then set flag to show string end found
        // and move pointer after this block
        //*************************************************************
        bFoundEnd = TRUE;
        pb++;
      }
    }
    //*****************************************************************
    // If popup, then skip over extra dword for help id as well.
    //*****************************************************************
    if (bPopUp)
      pb++;

    //*****************************************************************
    // Insert this item into the parent menu now that all information
    //  has been retrieved for this item
    //*****************************************************************
    InsertMenu( hMenuParent[iStartLevel], 0xFFFFFFFF,
                (UINT)(dwType | dwState | fuFlags),
                (UINT)(bPopUp ? (unsigned long)hPop : menuItemId),
                menuText );
  }
  return hMenuParent[1];
}


/*------------------------------------------------------------------------------
| deleteLookUpTable                                                            |
|  Local function to delete a previously built lookup table                    |
------------------------------------------------------------------------------*/
void IMenuPrivate::deleteLookUpTable( IFrameWindow* owner )
{
  if (owner->fFrameWindowData->fpSubmenuList)
  {
    FMTABLE* pFmTable = (FMTABLE*)(owner->fFrameWindowData->fpSubmenuList);
    FMTABLE* pNext;
    do {
      pNext = pFmTable->pfmtNext;
      delete pFmTable;
      pFmTable = pNext;
    } while (pFmTable);
  }
  owner->fFrameWindowData->fpSubmenuList = 0;
}


/**********************************************************************/
/* Procedure: EnumAllFrames - Callback function for EnumWindows.      */
/*            Receives the handles of top-level windows and checks    */
/*            each one looking for a lookup table function            */
/**********************************************************************/
int __stdcall IMenuPrivate::EnumAllFrames( void* hwnd, long  lpData )
{
  PFRAMEINFO pfrminfo = (PFRAMEINFO)lpData;

  /*------------------------------------------------------------------*/
  /* Check if there is an IWindow object for this window              */
  /*------------------------------------------------------------------*/
  IWindow* pwndTop = IWindow::windowWithHandle( hwnd );
  if ( pwndTop )
  {
    long id = 0;
    if (!(IPlatform::isWin95() || IPlatform::isNTNewShell()))
    {
      /*--------------------------------------------------------------*/
      /* For the frame passed, check if it contains a lookup list     */
      /*--------------------------------------------------------------*/
      IFrameWindow* tempFrame = (IFrameWindow*)GetProp( hwnd,
                                                  PROP_MENUBAR_OBJECT );
      if (tempFrame)
      {
        /*------------------------------------------------------------*/
        /* If it does, check if the lookup table contains the HMENU   */
        /*------------------------------------------------------------*/
        id = getItemIdFromLookUpTable( tempFrame, pfrminfo->hmenu );
        if ( id != 0xFFFFFFFFul )
        {
          /*----------------------------------------------------------*/
          /* If it contains the menu, then return the IMenu* of the   */
          /*  table and stop the enumeration                          */
          /*----------------------------------------------------------*/
          pfrminfo->owner = hwnd;
          pfrminfo->frame = tempFrame;
          return FALSE;
        }
      }
    }
    else
    {
      /*--------------------------------------------------------------*/
      /* See if this menu is contained somewhere in this window       */
      /*--------------------------------------------------------------*/
      HMENU hmTemp = GetMenu(hwnd);
      id = findIdForPopup( hmTemp, pfrminfo->hmenu );
      if ( id || (hmTemp == pfrminfo->hmenu))
      {
        /*------------------------------------------------------------*/
        /* If found, then save the window as the owner                */
        /*------------------------------------------------------------*/
        pfrminfo->owner = hwnd;

        /*------------------------------------------------------------*/
        /* For the window passed, get the frame window object (if any)*/
        /*------------------------------------------------------------*/
        pfrminfo->frame = (IFrameWindow*)GetProp( hwnd, PROP_MENUBAR_OBJECT );
        return FALSE;
      }
    }
  }
  /*------------------------------------------------------------------*/
  /* If it fails above, return true to continue searching for frame   */
  /*------------------------------------------------------------------*/
  return (TRUE);
}
#endif // IC_WIN


/*------------------------------------------------------------------------------
| setMenuItemHelpId                                                            |
|  Local function used to build help id table for menu items                   |
------------------------------------------------------------------------------*/
void IMenuPrivate::setMenuItemHelpId( IFrameWindow*  owner,
                                      unsigned long  menuItemId,
                                      unsigned long  helpTopicId )
{
  /********************************************************************/
  /* If a valid owner passed, check table for id passed               */
  /********************************************************************/
  if (owner)
  {
    /******************************************************************/
    /* Start at top and keep searching as long as entries exist       */
    /******************************************************************/
    MIHELPTBL* pMiHTable = (MIHELPTBL*)(owner->fFrameWindowData->fpHelpList);
    while (pMiHTable)
    {
      /****************************************************************/
      /* If match is found, return match, otherwise go to next entry  */
      /****************************************************************/
      if ( pMiHTable->fId == menuItemId )
      {
        pMiHTable->fHelpId = helpTopicId;
        return;
      }
      pMiHTable = pMiHTable->pmitNext;
    }

    /******************************************************************/
    /* Create a new menu lookup table entry and fill with info passed */
    /******************************************************************/
    MIHELPTBL* pMi2Table = new MIHELPTBL;
    pMi2Table->fId = menuItemId;
    pMi2Table->fHelpId = helpTopicId;

    /******************************************************************/
    /* Chain list (if any) to list                                    */
    /******************************************************************/
    pMi2Table->pmitNext = (MIHELPTBL*)(owner->fFrameWindowData->fpHelpList);

    /******************************************************************/
    /* Finally, since this is now the start of the list, save it      */
    /******************************************************************/
    owner->fFrameWindowData->fpHelpList = (void*)pMi2Table;
  }
  return;
}


/*------------------------------------------------------------------------------
| queryMenuItemHelpId                                                          |
|  Local function used to build help id table for menu items                   |
------------------------------------------------------------------------------*/
unsigned long IMenuPrivate::queryMenuItemHelpId( IFrameWindow*  owner,
                                                 unsigned long  menuItemId )
{
  /********************************************************************/
  /* If a valid owner passed, check table for id passed               */
  /********************************************************************/
  if (owner)
  {
    /******************************************************************/
    /* Start at top and keep searching as long as entries exist       */
    /******************************************************************/
    MIHELPTBL* pMiHTable = (MIHELPTBL*)(owner->fFrameWindowData->fpHelpList);
    while (pMiHTable)
    {
      /****************************************************************/
      /* If match is found, return match, otherwise go to next entry  */
      /****************************************************************/
      if ( pMiHTable->fId == menuItemId )
        return pMiHTable->fHelpId;

      pMiHTable = pMiHTable->pmitNext;
    }
  }
  return 0;
}


/*------------------------------------------------------------------------------
| deleteHelpIdTable                                                            |
|  Local function to delete a previously built help id table                   |
------------------------------------------------------------------------------*/
void IMenuPrivate::deleteHelpIdTable( IFrameWindow* owner )
{
  if (owner->fFrameWindowData->fpHelpList)
  {
    MIHELPTBL* pMiHTable = (MIHELPTBL*)(owner->fFrameWindowData->fpHelpList);
    MIHELPTBL* pNext;
    do {
      pNext = pMiHTable->pmitNext;
      delete pMiHTable;
      pMiHTable = pNext;
    } while (pMiHTable);
  }
  owner->fFrameWindowData->fpHelpList = 0;
}

