/*******************************************************************************
* FILE NAME: ireslib3.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of Pointer functions and classes.    *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #define INCL_WINPOINTERS
  #include <iwindefs.h>
}

#include <ireslib.hpp>
#include <iresstat.hpp>
#include <istring.hpp>
#include <irect.hpp>
#include <itrace.hpp>
#include <iset.h>
#include <ithread.hpp>
#include <ireslock.hpp>
#include <iexcept.hpp>
#include <iplatfrm.hpp>


// Segment definitions
#ifdef IC_PAGETUNE
  #define _IRESLIB3_CPP_
  #include <ipagetun.h>
#endif



/* Class to represent the real Icon  */
class  IPointer : public IBase
{
public:
 IPointer(unsigned long ulId, unsigned long handle,
            const IString& strFileName);
 ~IPointer();

 unsigned long addRef();

 unsigned long removeRef()        { return(--ulClUseCount);  }
 unsigned long useCount() const   { return ulClUseCount;  }
 unsigned long id() const         { IMODTRACE_ALL("Ptr::id");
                                    return ulClId;        }
 unsigned long handle()  const   { IMODTRACE_ALL("Ptr::handle");
                                    return ulClHandle;    }
 IString resourceFileName() const { IMODTRACE_ALL("Ptr::fileName");
                                    return strClFileName; }
      operator ==(const IPointer& ptr) const
      {
        if( (ptr.id() == this->id()) &&
             (ptr.resourceFileName() == this->resourceFileName())
          )
           return(1);
        else
           return(0);
      }

 private:
 unsigned long ulClId;
 unsigned long ulClHandle;
      IString strClFileName;
unsigned long ulClUseCount;
};

/*------------------------------------------------------------------------------
| IPointer::IPointer                                                           |
------------------------------------------------------------------------------*/
IPointer :: IPointer(unsigned long ulId, unsigned long handle,
            const IString& strFileName)
          : ulClId(ulId),  ulClHandle(handle),
            strClFileName(strFileName),
            ulClUseCount(0)     {}

/*------------------------------------------------------------------------------
| IPointer::~IPointer                                                          |
------------------------------------------------------------------------------*/
IPointer :: ~IPointer() {}

// Internal set for reference counting
class IPointerSet : public ISet<IPointer*>
{
public:
  IPointerSet( );
 ~IPointerSet( );
};

/*------------------------------------------------------------------------------
| IPointer::~IPointer                                                          |
------------------------------------------------------------------------------*/
unsigned long IPointer :: addRef()
{
  return( ++ulClUseCount );
}


/*------------------------------------------------------------------------------
| IPointerSet::IPointerSet                                                     |
------------------------------------------------------------------------------*/
IPointerSet ::  IPointerSet ( )
      : ISet<IPointer*>() {}

/*------------------------------------------------------------------------------
| IPointerSet::~IPointerSet                                                    |
------------------------------------------------------------------------------*/
IPointerSet :: ~IPointerSet ( ) {}


/*------------------------------------------------------------------------------
| IResLibStaticPtr::deletePointerSet                                           |
------------------------------------------------------------------------------*/
void IResLibStaticPtr :: deletePointerSet()
{
  if(pPtrSet)
    delete pPtrSet;
}


/*------------------------------------------------------------------------------
| IResLibStaticPtr::pointerSet                                                 |
------------------------------------------------------------------------------*/
IPointerSet& IResLibStaticPtr ::  pointerSet()
{
  if(pPtrSet==0)
    pPtrSet = new IPointerSet;
  return *pPtrSet;
}


/*------------------------------------------------------------------------------
| IResourceLibrary::loadIcon                                                   |
------------------------------------------------------------------------------*/
IPointerHandle IResourceLibrary :: loadIcon(unsigned long ulIconId,
                                        Boolean fManaged) const
{

  IMODTRACE_ALL("ResLib::loadIcon");
  IPointerHandle hptr = tryToLoadIcon(ulIconId, fManaged);
  if(hptr == 0)
      ITHROWGUIERROR(IString("ILOADPOINTER(")+
                     "IResourceId:"+IString(ulIconId)+")");

  return hptr;

}

/*------------------------------------------------------------------------------
| IResourceLibrary::loadPointer                                                |
------------------------------------------------------------------------------*/
IPointerHandle IResourceLibrary :: loadPointer(unsigned long iconId,
                                               Boolean cached) const
{ return loadIcon(iconId, cached); }


/*------------------------------------------------------------------------------
| IResourceLibrary::tryToLoadIcon                                              |
------------------------------------------------------------------------------*/
IPointerHandle IResourceLibrary :: tryToLoadIcon(unsigned long ulIconId,
                                        Boolean fManaged) const
{

  IMODTRACE_ALL("ResLib::tryToLoadIcon");
#ifdef IC_WIN
  // Windows manages icons loaded from resource files automatically.
  // We don't need to do it also.
  fManaged = false;
#endif
  /* Lock Resource Table */
  IResourceLock reslibLock(IResLibStaticPtr::libraryKey());

  unsigned long hptrReturn = 0;
  IPointer* pptr = 0;

  /* If resources are managed, build a cursor to see if we have already */
  /* loaded this Icon.  If so just update the use count                 */
  Boolean fFound = false;
  if(fManaged)
  {
      IPointerSet::Cursor cursor(IResLibStaticPtr::pointerSet());
      for(cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
      {
        pptr = IResLibStaticPtr::pointerSet().elementAt(cursor);
        if( (pptr->id() == ulIconId)               &&
          (pptr->resourceFileName() == fileName()) )
        {
           ITRACE_ALL("PTR Match found");
           fFound = true;
           break;
        }
      }
  }
  if(fFound==true)
  {
    hptrReturn = pptr->handle();
  }
  else
  /* Either we are not managing the resource, or it doesn't exist yet   */
  {

#ifdef IC_WIN
     // In windows, need to try to load both an icon and pointer
     if (IPlatform::isWin95())
        {
        hptrReturn = (unsigned long)
                        LoadImage( handle(),
                                   MAKEINTRESOURCE(ulIconId),
                                   IMAGE_ICON, 0, 0,
                                   LR_DEFAULTCOLOR );
        if (hptrReturn == 0)
          {
          hptrReturn = (unsigned long)
                          LoadImage( handle(),
                                     MAKEINTRESOURCE(ulIconId),
                                     IMAGE_CURSOR, 0, 0,
                                     LR_DEFAULTCOLOR );
          } //
        }
     else
        {
        hptrReturn = (unsigned long)
                        LoadIcon( handle(), MAKEINTRESOURCE(ulIconId) );
        if (hptrReturn == 0)
          {
          hptrReturn = ILOADPOINTER( handle(), ulIconId );
          } //
        }
#endif
#ifndef IC_WIN
     /* Load the Icon or Pointer */
     hptrReturn = ILOADPOINTER( handle(), ulIconId );
#endif
     ITRACE_ALL(IString("Pointer loaded is: ")+IString(hptrReturn));

     if ( hptrReturn && fManaged)
     {
        IPointer* pptrNew = new IPointer(ulIconId, hptrReturn,
                                           fileName());
        IResLibStaticPtr::pointerSet().add(pptrNew);
     }
  }

#ifdef IC_WIN
  return IPointerHandle((void*)hptrReturn);
#else
  return IPointerHandle(hptrReturn);
#endif
}

/*------------------------------------------------------------------------------
| IResourceLibrary::systemPointer                                              |
------------------------------------------------------------------------------*/
unsigned long IResourceLibrary :: systemPointer ( unsigned long systemPointerId,
                                        Boolean fMakeCopy, Boolean fIconApi )
{
  unsigned long hptr = IQUERYSYSPOINTER( systemPointerId, fMakeCopy,
                                         fIconApi );
  if (!hptr)
  {
     ITHROWGUIERROR("IQUERYSYSPOINTER");
  }
  /* If we return a copy we will manage reference count this resource */
  /* otherwise not (don't want to destroy a system pointer).          */
  if(fMakeCopy)
  {
    IPointer* pptrNew = new IPointer( systemPointerId,
                                      hptr,
                                      IString());
    IResLibStaticPtr::pointerSet().add(pptrNew);
  }
  return hptr;
}


/*------------------------------------------------------------------------------
| IResourceLibrary::addRef                                                     |
------------------------------------------------------------------------------*/
unsigned long IResourceLibrary :: addRef(const IPointerHandle& ptrh)
{
  IMODTRACE_ALL("ResLib::addRef");
  ITRACE_ALL(IString("PTRHandle is: ") + IString(ptrh.asUnsigned()));

  unsigned long ulUseCount = 0;
  IPointer* pptr;
  if(ptrh!=0)
  {
     /* Lock Resource Table */
     IResourceLock reslibLock(IResLibStaticPtr::libraryKey());
     IPointerSet::Cursor cursor(IResLibStaticPtr::pointerSet());
     for(cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
     {
        pptr = IResLibStaticPtr::pointerSet().elementAt(cursor);
#ifdef IC_WIN
        if((void*)pptr->handle() == ptrh)
#else
        if(pptr->handle() == ptrh)
#endif
        {
          ulUseCount = pptr->addRef();
          ITRACE_ALL(IString("Handle is: ")+IString(pptr->handle()) +
                     " UseCount now: " + IString(ulUseCount));
          break;
        }

     }
  } /* endif */
  return(ulUseCount);
}

/*------------------------------------------------------------------------------
| IResourceLibrary::removeRef                                                  |
------------------------------------------------------------------------------*/
unsigned long IResourceLibrary :: removeRef(const IPointerHandle& ptrh)
{
  IMODTRACE_ALL("ResLib::removeRef");
  ITRACE_ALL(IString("PTRHandle is: ") + IString(ptrh.asUnsigned()));

  unsigned long ulUseCount = 0;
  IPointer* pptr;
  if(ptrh!=0)
  {
     /* Lock Resource Table */
     IResourceLock reslibLock(IResLibStaticPtr::libraryKey());

     IPointerSet::Cursor cursor(IResLibStaticPtr::pointerSet());
     for(cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
     {
        pptr = IResLibStaticPtr::pointerSet().elementAt(cursor);
#ifdef IC_WIN
        if((void*)pptr->handle() == ptrh)
#else
        if(pptr->handle() == ptrh)
#endif
        {
          ulUseCount = pptr->removeRef();
          ITRACE_ALL(IString("Handle is: ")+IString(pptr->handle()) +
                     " UseCount now: " + IString(ulUseCount));
          if(ulUseCount==0)
          {
            Boolean fSuccess = IDESTROYPOINTER(ptrh);
            ITRACE_ALL("Pointer destroyed");
            if(!fSuccess)
            {
                ITHROWGUIERROR("IDESTROYPOINTER");
            }
            IResLibStaticPtr::pointerSet().removeAt(cursor);
            delete(pptr);
          }
          break;
        }
     }
  } /* endif */
  return ulUseCount;
}

