/*******************************************************************************
* FILE NAME: isliderw.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the native Windows slider         *
*   functions declared in islider.hpp.                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
{
  #include <iwindefs.h>
  #include <commctrl.h>
}

#include <iclsldw.h>
#include <islider.hpp>
#include <isliderw.hpp>
#include <icoordsy.hpp>
#include <iexcept.hpp>
#include <ifont.hpp>
#include <igbitmap.hpp>
#include <igrafctx.hpp>
#include <igrect.hpp>
#include <igregion.hpp>
#include <ihandle.hpp>
#include <irect.hpp>
#include <itrace.hpp>
#include <iwcname.hpp>

IBase::Boolean IProgressIndicatorData::fRegistered( false );

#define IC_SLIDER_WIDTH  34
#define IC_BUTTON_SIZE   20

#ifndef MIN
  #define MIN(a, b) ( ( (a) <= (b) ) ? (a) : (b) )
#endif
#ifndef MAX
  #define MAX(a, b) ( ( (a) >= (b) ) ? (a) : (b) )
#endif

DefaultHandler :: operator INativeSliderHandler* ()
{
  if ( !ptr )
     ptr = new INativeSliderHandler;
  return ptr;
}

/*------------------------------------------------------------------------------
| IProgressIndicatorData::IProgressIndicatorData                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IProgressIndicatorData :: IProgressIndicatorData () :
  trackbar( 0 ),
  scaleFactor( 1 ),
  leftDown( 0 ),
  leftDownPressed( 0 ),
  leftDownDisabled( 0 ),
  rightUp( 0 ),
  rightUpPressed( 0 ),
  rightUpDisabled( 0 ),
  tickPos0( 0 ),
  textVisible( true ),
  timer( 0 )
  {}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::~IProgressIndicatorData                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IProgressIndicatorData :: ~IProgressIndicatorData ()
{
  for (unsigned long scale = 0; scale < 2; scale++)
  {
     IGList::Cursor cursor( tickText[scale] );
     cursor.setToFirst();
     IGString *gText;
     while (cursor.isValid())
     {
        gText = (IGString*)&((tickText[scale].graphicAt( cursor )));
        tickText[scale].removeAt( cursor );
        delete gText;
        cursor.setToFirst();
     }
  }

  delete leftDown;
  delete leftDownPressed;
  delete leftDownDisabled;
  delete rightUp;
  delete rightUpPressed;
  delete rightUpDisabled;
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::registerSlider                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::registerSlider( )
{
  static Boolean fRegistered(false);

  // Register the wrapper window class.
  if (!fRegistered)
  {
    InitCommonControls();

    fRegistered = true;
    WNDCLASS wndClass;
    HBRUSH   defBkgrndBrush = CreateSolidBrush( IGUIColor( IGUIColor::buttonLight )
                                .asRGBLong() );

    wndClass.style         = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
    wndClass.cbClsExtra    = 0;
    wndClass.cbWndExtra    = sizeof(LONG);
    wndClass.hCursor       = LoadCursor(0,IDC_ARROW);
    wndClass.hInstance     = GetModuleHandle(0);
    wndClass.lpfnWndProc   = (WNDPROC)IDEFWINDOWPROC;
    wndClass.hIcon         = 0;
    wndClass.hbrBackground = defBkgrndBrush;
    wndClass.lpszMenuName  = 0;
    wndClass.lpszClassName = WC_NATIVESLIDER;

    if (!RegisterClass(&wndClass))
      ITHROWGUIERROR( "RegisterClass" );
  }
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::createNativeSlider                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::createNativeSlider(
                                   const IWindowHandle& parent,
                                   const IWindowHandle& owner,
                                   unsigned long        ulStyle,
                                   unsigned long        extStyle )
{
  IMODTRACE_ALL("IProgressIndicatorData::createNativeSlider");

  HINSTANCE hInst = GetModuleHandle(0);

  // Create the native trackbar.  Use an initial size of 100 x 100
  // in order to obtain a valid left/topmost tick offset.
  trackbar = ICREATEWINDOW( parent,
                            TRACKBAR_CLASS,
                            0,
                            ulStyle | WS_VISIBLE,
                            0,
                            0,
                            100,
                            100,
                            owner,
                            HWND_BOTTOM,
                            IIDOF( parent ),
                            0,
                            0,
                            hInst );

  if (!trackbar)
    ITHROWGUIERROR(IString("WinCreateWindow: Id=")+IString(IIDOF( parent ))+
           IString(" Class=")+IString((unsigned short)(unsigned long)TRACKBAR_CLASS));

  // If this is a slider with buttons, create the bitmaps for the buttons.
  if (extStyle & ISLS_BUTTONSLEFT || extStyle & ISLS_BUTTONSRIGHT)
  {
     if (ulStyle & TBS_VERT)
     {
        leftDown = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarDownArrow )));
        leftDownPressed = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarDownArrowPressed )));
        leftDownDisabled = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarDownArrowDisabled )));
        rightUp = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarUpArrow )));
        rightUpPressed = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarUpArrowPressed )));
        rightUpDisabled = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarUpArrowDisabled )));
     }
     else
     {
        leftDown = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarLeftArrow )));
        leftDownPressed = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarLeftArrowPressed )));
        leftDownDisabled = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarLeftArrowDisabled )));
        rightUp = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarRightArrow )));
        rightUpPressed = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarRightArrowPressed )));
        rightUpDisabled = new IGBitmap( IGBitmap::copy(
              ISystemBitmapHandle( ISystemBitmapHandle::scrollBarRightArrowDisabled )));
     }

     // Only the normal and pressed buttons are hit selectable.
     leftDownDisabled->setHitSelectable( false );
     rightUpDisabled->setHitSelectable( false );

     // Size the button bitmaps.
     leftDown->sizeTo( ISize( IC_BUTTON_SIZE, IC_BUTTON_SIZE ));
     leftDownPressed->sizeTo( ISize( IC_BUTTON_SIZE, IC_BUTTON_SIZE ));
     leftDownDisabled->sizeTo( ISize( IC_BUTTON_SIZE, IC_BUTTON_SIZE ));
     rightUp->sizeTo( ISize( IC_BUTTON_SIZE, IC_BUTTON_SIZE ));
     rightUpPressed->sizeTo( ISize( IC_BUTTON_SIZE, IC_BUTTON_SIZE ));
     rightUpDisabled->sizeTo( ISize( IC_BUTTON_SIZE, IC_BUTTON_SIZE ));

     // Add bitmaps to the buttons IGList.  The correct state will be determined in
     // setButtonState.
     buttons.addAsLast( *leftDown );
     buttons.addAsLast( *rightUp );
  }

  // Get the offset of the left/topmost tick and store it to be
  // used later in layout.  We want to get it before we set the
  // scale to avoid problems when using the homeRight or
  // homeBottom styles.  The offset of this tick within the
  // trackbar does not change.
  unsigned long ulResult = trackbar.sendEvent(
                                       TBM_SETTIC,
                                       IEventParameter1( 0 ),
                                       IEventParameter2( 0 ));
  if (!ulResult)
  {
     ITHROWGUIERROR2("TBM_SETTIC",
                     IBaseErrorInfo::invalidParameter,
                     IException::recoverable);
  }
  tickPos0 = trackbar.sendEvent( TBM_GETTICPOS,
                                 IEventParameter1( 0 ),
                                 IEventParameter2( 0 ));

  // Subclass the trackbar's window procedure so we can send focus notifications
  // to our parent.  This is necessary to enable an IFocusHandler attached to the
  // slider.  This subclasses window procedure is needed for progress indicators
  // to prevent them from receiving input focus.
  fdefaultTrackbarProc = (WNDPROC)ISUBCLASSWINDOW( trackbar,
                                                   isliderTrackbarWinProc );
  if (!fdefaultTrackbarProc)
  {
     ITHROWGUIERROR( "ISUBCLASSWINDOW" );
  }
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::destroyNativeSlider                                  |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::destroyNativeSlider( )
{
   // Remove the subclass of the trackbar window proc
   if ( trackbar && IISWINDOW( 0, trackbar ) )
   {
      ISUBCLASSWINDOW( trackbar,
                       fdefaultTrackbarProc );
   }
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::setupNativeScale                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::setupNativeScale( IProgressIndicator::Scale scale )
{
  IMODTRACE_ALL("IProgressIndicatorData::setupNativeScale");

  unsigned long range,
                pageSize;

  IProgressIndicator *progressInd = (IProgressIndicator*)IWindow::windowWithHandle(
                                          IPARENTOF (trackbar) );
  if (numberTicks [scale] == 0)
  {
     range = 0;
     pageSize = 0;
  }
  else
  {
     if (progressInd->isSnapToTickEnabled())
     {
        // SnapToTick is implemented via the page and line size for
        // the trackbar.  These values represent the movement
        // increment for mouse and keyboard.
        range = numberTicks [scale] - 1;
        pageSize = 1;
     }
     else
     {
        // If this slider is not snapToTick, the range is set to the closest
        // multiple of the page size less that the maximum short value.
        // Set the page size to the proper proportion.
        pageSize = SHRT_MAX / (numberTicks [scale] - 1);
        range = pageSize * (numberTicks[scale] - 1);
     }
  }
  if (!progressInd->isVertical() &&
         progressInd->homePosition() == IProgressIndicator::homeBottomLeft ||
       progressInd->isVertical() &&
         progressInd->homePosition() == IProgressIndicator::homeTopRight)
  {
     trackbar.sendEvent( TBM_SETRANGEMAX,
                         IEventParameter1( true ),
                         IEventParameter2( range ));
     scaleFactor = 1;
  }
  else
  {
     trackbar.sendEvent( TBM_SETRANGEMIN,
                         IEventParameter1( true ),
                         IEventParameter1( -range ));
     scaleFactor = -1;
  }

  trackbar.sendEvent( TBM_SETPAGESIZE,
                      IEventParameter1( 0 ),
                      IEventParameter2( pageSize ));
  trackbar.sendEvent( TBM_SETLINESIZE,
                      IEventParameter1( 0 ),
                      IEventParameter2( pageSize ));

  // Now clear and reset ticks for this scale.
  trackbar.sendEvent( TBM_CLEARTICS,
                      IEventParameter1( true ),
                      IEventParameter2( 0 ));

  if (!visibleTicks[scale].isEmpty())
  {
     TickSet tempTicks = visibleTicks[scale];
     TickSet::Cursor cursor (tempTicks);
     unsigned long ulTickNum;
     visibleTicks[scale].removeAll();
     forCursor (cursor)
     {
        ulTickNum = tempTicks.elementAt( cursor );
        if (ulTickNum < numberTicks[scale])
           visibleTicks[scale].add( ulTickNum );
     }

     if (!visibleTicks[scale].isEmpty())
     {
        unsigned long ulPosition, ulResult;
        TickSet::Cursor tickCursor (visibleTicks[scale]);
        forCursor (tickCursor)
        {
           ulTickNum = visibleTicks[scale].elementAt( tickCursor );

           ulPosition = ulTickNum * pageSize * scaleFactor;

           ulResult = trackbar.sendEvent( TBM_SETTIC,
                                    IEventParameter1( 0 ),
                                    IEventParameter2( ulPosition ));
           if (!ulResult)
              ITHROWGUIERROR2("TBM_SETTIC",
                              IBaseErrorInfo::invalidParameter,
                              IException::recoverable);
        }
     }
  }

  // Determine whether text will be visible given this alignment and
  // primary scale.  If not, we won't calculate text positions nor draw
  // text.
  IProgressIndicator::Alignment align = progressInd->alignment();
  if ((align == IProgressIndicator::topRight
            && scale == IProgressIndicator::scale1)
      || (align == IProgressIndicator::bottomLeft
            && scale == IProgressIndicator::scale2))
     textVisible = false;
  else
     textVisible = true;
}

/*------------------------------------------------------------------------------
| IProgressIndicatorData::layoutNativeSlider                                   |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::layoutNativeSlider( )
{
  IMODTRACE_ALL("IProgressIndicatorData::layoutNativeSlider");

  IProgressIndicator *progressInd = (IProgressIndicator*)
        IWindow::windowWithHandle( IPARENTOF( trackbar ));


  long trackbarx, trackbary, trackbarcx, trackbarcy, spacing;
  IGraphicContext  gc( progressInd->handle() );
  ISize sliderSize( progressInd->size() );
  if (sliderSize == ISize())
     return;

  IProgressIndicator::Alignment align = progressInd->alignment();
  IProgressIndicator::HomePosition home = progressInd->homePosition();
  IProgressIndicator::Scale scale = progressInd->primaryScale();
  Boolean isVertical = progressInd->isVertical();
  unsigned long numTicks = progressInd->numberOfTicks( scale );
  if (numTicks == 0)
     return;

  ISlider::ButtonsPosition buttonPos;
  if (~ISTYLEOF( trackbar ) & TBS_NOTHUMB)
  {
     // This is a slider so we need to check for button position.
     ISlider *slider = (ISlider*)progressInd;
     buttonPos = slider->buttonsPosition();
  }
  else
     buttonPos = ISlider::none;

  long buttonSize = 0;
  if (buttonPos != ISlider::none)
     buttonSize = leftDown->size().width(); // Assume square bitmap.

  // Calculate size and position of trackbar within the slider static
  // control using top left origin.  Allow some additional room at the ends of
  // the slider to accomodate text if possible.

  ISize      firstSize(0,0), lastSize(0,0);
  IGString   *firstString, *lastString;
  if (!tickText[scale].isEmpty() && textVisible)
  {
     firstString = (IGString*)(tickText[scale].firstGraphic());
     if (firstString->id() == 0)
     {
        firstSize = ISize( firstString->boundingRect( gc ).size() );
     }

     lastString = (IGString*)(tickText[scale].lastGraphic());
     if (lastString->id() == numTicks-1)
     {
        lastSize = ISize( lastString->boundingRect( gc ).size() );
     }
  }

  if (isVertical)
  {
     // Vertical slider.
     if (tickSpacing[scale] != 0)
     {
        spacing = tickSpacing[scale];
     }
     else
     {
        if (firstSize.height() > 2*tickPos0 || lastSize.height() > 2*tickPos0)
        {
           // Using a large font here.  Allow a litte more room on the top/
           // bottom of the trackbar.  Allow no more room before the
           // first tick (or after the last tick) than half of the tick
           // spacing.

           unsigned long maxEndHeight = MAX( firstSize.height(), lastSize.height() );
           spacing = (sliderSize.height() - maxEndHeight - 2*buttonSize) / (numTicks-1);
           spacing = MAX( spacing,
                          (long)((sliderSize.height() - 2*buttonSize) / numTicks) );
        }
        else
        {
           spacing = (sliderSize.height() - 2*buttonSize - 2*tickPos0) / (numTicks-1);
        }
     }
     pixelRange = spacing * (numTicks-1);
     trackbarcx = IC_SLIDER_WIDTH;
     trackbarcy = pixelRange + 2*tickPos0;
     trackbary = MAX( 0, (long)((sliderSize.height() - trackbarcy)/2 - buttonSize) );
     if (buttonPos == ISlider::top)
        trackbary += 2*buttonSize;

     switch (align)
     {
       case IProgressIndicator::centered:
         trackbarx = (sliderSize.width() - IC_SLIDER_WIDTH) / 2;
         break;
       case IProgressIndicator::bottomLeft:
         trackbarx = 0;
         break;
       default:
         trackbarx = sliderSize.width() - IC_SLIDER_WIDTH;
     }
  }
  else
  {
     // Horizontal slider.
     if (tickSpacing[scale] != 0)
     {
        spacing = tickSpacing[scale];
     }
     else
     {
        // Check whether either the first or last tick text could use a little
        // more room.  Allow no more for the first and last tick text than
        // for any of the other ticks (ie, no more than half the spacing
        // will be left at either end).  Note that this calculation is a good
        // approximate of the optimal sizing since it doesn't consider the
        // space gain from text appearing above the buttons.

        if ((firstSize.width() > 2*tickPos0 || lastSize.width() > 2*tickPos0))
        {
           unsigned long maxEndWidth = MAX( firstSize.width(), lastSize.width() );
           spacing = (sliderSize.width() - maxEndWidth - 2*buttonSize) / (numTicks-1);
           // Don't leave more that half the spacing on each end:
           spacing = MAX( spacing, (sliderSize.width() - 2*buttonSize) / numTicks );
        }
        else
        {
           spacing = (sliderSize.width() - 2*buttonSize - 2*tickPos0) / (numTicks-1);
        }
     }
     pixelRange = spacing * (numTicks-1);
     trackbarcx = pixelRange + 2*tickPos0;
     trackbarx = MAX( 0, (long)((sliderSize.width() - trackbarcx)/2 - buttonSize));
     if (buttonPos == ISlider::left)
        trackbarx += 2*buttonSize;
     trackbarcy = IC_SLIDER_WIDTH;

     switch (align)
     {
       case IProgressIndicator::centered:
         trackbary = (sliderSize.height() - IC_SLIDER_WIDTH) / 2;
         break;
       case IProgressIndicator::bottomLeft:
         trackbary = sliderSize.height() - IC_SLIDER_WIDTH;
         break;
       default:
         trackbary = 0;
     }
  }

  // Set the size/position of the trackbar.
  MoveWindow( trackbar,
              trackbarx,
              trackbary,
              trackbarcx,
              trackbarcy,
              true );

  // Now we need to set the clip rectangle and position of
  // the tick text strings.
  if (!tickText[scale].isEmpty() && textVisible)
  {
     long             tick0xy, tickxy, stringId;
     IGString         *gText;
     IGList::Cursor   cursor( tickText[scale] );
     IRectangle       clipRect;
     ISize            textSize;
     IPoint           textPosition;

     // Set the value of offset0.  This value is the position of the
     // left or top most tick mark in the static control.
     tick0xy = (isVertical) ? (tickPos0 + trackbary) : (tickPos0 + trackbarx);

     // Calculate the new clipping rectangle and text position.
     for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
     {
         gText = (IGString*)&(tickText[scale].graphicAt( cursor ));
         stringId = gText->id();
         textSize = gText->boundingRect( gc ).size();
         tickxy = tick0xy + stringId*spacing;

         if (isVertical)
         {
            if (scale == IProgressIndicator::scale1)
            {
               // Text to right of trackbar.
               clipRect.moveTo( IPoint( trackbarx + trackbarcx,
                                        MAX( 0, (long)(tickxy - spacing/2 ))));
               if (stringId == 0)
                  clipRect.sizeTo( ISize( sliderSize.width() - trackbarcx - trackbarx,
                                          MIN( spacing, spacing/2 + tick0xy )));
               else if (stringId == numTicks-1)
                  clipRect.sizeTo( ISize( sliderSize.width() - trackbarcx - trackbarx,
                                          MIN( spacing,
                                               sliderSize.height() - tickxy + spacing/2 )));
               else
                  clipRect.sizeTo( ISize( sliderSize.width() - trackbarcx - trackbarx,
                                          spacing ));

               if (textSize.height() > clipRect.height())
                  textPosition = IPoint( trackbarx + trackbarcx,
                                         MAX( 0, (long)(tickxy - spacing/2 )));
               else
                  textPosition = IPoint( trackbarx + trackbarcx,
                                         tickxy - textSize.height()/2 );
            }
            else
            {
               // Text left of trackbar.
               clipRect.moveTo( IPoint( 0, MAX( 0, (long)(tickxy - spacing/2 ))));
               if (stringId == 0)
                  clipRect.sizeTo( ISize( trackbarx,
                                          MIN( spacing,
                                                spacing/2 + tick0xy )));
               else if (stringId == numTicks-1)
                  clipRect.sizeTo( ISize( trackbarx,
                                             MIN( spacing,
                                                  sliderSize.height() - tickxy + spacing/2)));
               else
                  clipRect.sizeTo( ISize( trackbarx, spacing ));

               if (textSize.height() > clipRect.height())
                  textPosition = IPoint( MAX(0, (long)(trackbarx - textSize.width())),
                                         tickxy - spacing/2 );
               else
                  textPosition = IPoint( MAX( 0, (long)(trackbarx - textSize.width())),
                                         tickxy - textSize.height()/2 );
            }
         }
         else
         {
            // Horizontal slider.
            if (scale == IProgressIndicator::scale1)
            {
               // Text on top of trackbar.
               clipRect.moveTo( IPoint( MAX( 0, (long)(tickxy-spacing/2) ), 0 ));

               if (stringId == 0)
                  clipRect.sizeTo( ISize( MIN( spacing, spacing/2 + tick0xy ),
                                          trackbary ));
               else if (stringId == numTicks-1)
                  clipRect.sizeTo( ISize( MIN( spacing,
                                               sliderSize.width() - tickxy + spacing/2 ),
                                          trackbary ));
               else
                  clipRect.sizeTo( ISize( spacing, trackbary ));

               if (textSize.width() > clipRect.width())
                  textPosition = IPoint( clipRect.left(),
                                         MAX( 0, (long)(trackbary - textSize.height()) ));
               else
                  textPosition = IPoint( tickxy - textSize.width()/2,
                                         MAX( 0, (long)(trackbary - textSize.height()) ));
            }
            else
            {
               // Text below trackbar.
               clipRect.moveTo( IPoint( MAX( 0, (long)(tickxy-spacing/2) ),
                                        trackbary + trackbarcy ));
               if (stringId == 0)
                  clipRect.sizeTo( ISize( MIN( spacing, spacing/2 + tick0xy ),
                                          sliderSize.height()-trackbary-trackbarcy ));
               else if (stringId == numTicks-1)
                  clipRect.sizeTo( ISize( MIN( spacing,
                                               sliderSize.width() - tickxy + spacing/2 ),
                                          sliderSize.height()-trackbary-trackbarcy ));
               else
                  clipRect.sizeTo( ISize( spacing,
                                          sliderSize.height() - trackbary - trackbarcy ));

               if (textSize.width() > clipRect.width())
                  textPosition = IPoint( clipRect.left(),
                                         trackbary + trackbarcy );
               else
                  textPosition = IPoint( tickxy - textSize.width()/2,
                                         trackbary + trackbarcy );
            }
         }
         gText->setClippingRect( clipRect );
         gText->moveTo( textPosition );
     }
  }

  // Set the positions of the buttons.
  if (buttonPos != ISlider::none)
  {
     unsigned long leftDownPos, rightUpPos;
     switch (buttonPos)
     {
       case ISlider::left:
         leftDownPos = trackbarx - 2*buttonSize;
         rightUpPos = leftDownPos + buttonSize;
         break;
       case ISlider::right:
         leftDownPos = trackbarx + trackbarcx;
         rightUpPos = leftDownPos + buttonSize;
         break;
       case ISlider::bottom:
         rightUpPos = trackbary + trackbarcy;
         leftDownPos = rightUpPos + buttonSize;
         break;
       case ISlider::top:
         rightUpPos = trackbary - 2*buttonSize;
         leftDownPos = rightUpPos + buttonSize;
         break;
     }

     unsigned long buttonx = trackbarx + (trackbarcx - buttonSize)/2,
                   buttony = trackbary + (trackbarcy - buttonSize)/2;
     if (isVertical)
     {
        rightUp->moveTo( IPoint( buttonx, rightUpPos ));
        rightUpPressed->moveTo( IPoint( buttonx, rightUpPos ));
        rightUpDisabled->moveTo( IPoint( buttonx, rightUpPos ));
        leftDown->moveTo( IPoint( buttonx, leftDownPos ));
        leftDownPressed->moveTo( IPoint( buttonx, leftDownPos ));
        leftDownDisabled->moveTo( IPoint( buttonx, leftDownPos ));
     }
     else
     {
        leftDown->moveTo( IPoint( leftDownPos, buttony ));
        leftDownPressed->moveTo( IPoint( leftDownPos, buttony ));
        leftDownDisabled->moveTo( IPoint( leftDownPos, buttony ));
        rightUp->moveTo( IPoint( rightUpPos, buttony ));
        rightUpPressed->moveTo( IPoint( rightUpPos, buttony ));
        rightUpDisabled->moveTo( IPoint( rightUpPos, buttony ));
     }
  }
  return;
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::trackbarRange                                        |
|                                                                              |
| Return the range extent of the trackbar.                                     |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IProgressIndicatorData::trackbarRange( )
{
  IMODTRACE_ALL("IProgressIndicatorData::trackbarRange");

  if (scaleFactor == 1)
     return (trackbar.sendEvent( TBM_GETRANGEMAX,
                                 IEventParameter1( 0 ),
                                 IEventParameter2( 0 )));
  else
     return (-(trackbar.sendEvent( TBM_GETRANGEMIN,
                                  IEventParameter1( 0 ),
                                  IEventParameter2( 0 ))));
}

/*------------------------------------------------------------------------------
| IProgressIndicatorData::armRangeOffset                                       |
|                                                                              |
| Return the position in the native trackbar.  This value will be scaled to a  |
| positive position since the actual position may be negative depending on the |
| control windows's home position.                                             |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IProgressIndicatorData::armRangeOffset( )
{
  IMODTRACE_ALL("IProgressIndicatorData::armRangeOffset");

  unsigned long ulStyle = ISTYLEOF( trackbar );
  long position;

  if (ulStyle & TBS_NOTHUMB)
     // This is a progress indicator (ie, read only).  In this case, only a
     // selection range is available.
     if (scaleFactor == 1)
        // We have a homeLeft horizontal or a homeTop vertical
        // trackbar.
        position = trackbar.sendEvent( TBM_GETSELEND,
                                       IEventParameter1( 0 ),
                                       IEventParameter2( 0 ) );
     else
        // We have a homeRight horizontal or a homeBottom vertical
        // trackbar.  In this case we want the (negative) start of the
        // selection range.
        position = trackbar.sendEvent( TBM_GETSELSTART,
                                       IEventParameter1( 0 ),
                                       IEventParameter2( 0 ) );
  else
     // This is a slider.  We can query the arm position.
     position = trackbar.sendEvent( TBM_GETPOS,
                                    IEventParameter1( 0 ),
                                    IEventParameter2( 0 ) );

  return ( position * scaleFactor );
}

/*------------------------------------------------------------------------------
| IProgressIndicatorData::setArmRangeOffset                                    |
|                                                                              |
| Set the position in the native trackbar.  The new position provided will be  |
| scaled appropriately to account for the control window's home position.  The |
| control will be redrawn.                                                     |
|                                                                              |
------------------------------------------------------------------------------*/
IProgressIndicatorData& IProgressIndicatorData::setArmRangeOffset(
                              unsigned long newPosition )
{
  IMODTRACE_ALL("IProgressIndicatorData::setArmRangeOffset");

  unsigned long ulStyle = ISTYLEOF( trackbar );

  // Set the arm position.
  trackbar.sendEvent( TBM_SETPOS,
                      IEventParameter1( true ),
                      IEventParameter2( scaleFactor * newPosition ) );

  if (~ulStyle & TBS_NOTHUMB)
  {
     // Set the button state.
     setButtonState();
  }

  if (ulStyle & TBS_ENABLESELRANGE)
  {
     drawRibbon( newPosition );
  }

  // Since Windows won't send an event, send one here to enable
  // handler and notification.
  IWindowHandle hwndParent = IPARENTOF( trackbar );
  hwndParent.sendEvent( IC_UM_VSCROLL,
                        IEventParameter1( TB_ENDTRACK, 0 ),
                        IEventParameter2( trackbar ));

  return *this;
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::drawRibbon                                           |
|                                                                              |
| Draw the ribbon in the native trackbar by adjusting the selection range.     |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::drawRibbon( unsigned long newPosition )
{
  IMODTRACE_ALL("IProgressIndicatorData::drawRibbon");

  unsigned long currentSelection;

  if (scaleFactor == 1)
  {
     // We have a homeLeft horizontal or a homeTop vertical
     // trackbar.
     currentSelection = trackbar.sendEvent( TBM_GETSELEND,
                                            IEventParameter1( 0 ),
                                            IEventParameter2( 0 ));
     if (currentSelection != newPosition)
        trackbar.sendEvent( TBM_SETSELEND,
                            IEventParameter1( true ),
                            IEventParameter2( newPosition ) );
  }
  else
  {
     // We have a homeRight horizontal or a homeBottom vertical
     // trackbar.  In this case we want to set the (negative) start of the
     // selection range.
     currentSelection = trackbar.sendEvent( TBM_GETSELSTART,
                                            IEventParameter1( 0 ),
                                            IEventParameter2( 0 ));
     if (currentSelection != -newPosition)
        trackbar.sendEvent( TBM_SETSELSTART,
                            IEventParameter1( true ),
                            IEventParameter2( -newPosition ) );
  }
  return;
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::setButtonState                                       |
|                                                                              |
| Draw the slider buttons in their correct state.                              |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::setButtonState( )
{
  IMODTRACE_ALL("IProgressIndicatorData::setButtonState");

  ISlider *slider;
  ISlider::ButtonsPosition buttonPos;
  if (~ISTYLEOF( trackbar ) & TBS_NOTHUMB)
  {
     slider = (ISlider*)IWindow::windowWithHandle( IPARENTOF( trackbar ));
     buttonPos = slider->buttonsPosition();
  }
  else
     // This is a progress indicator.
     buttonPos = ISlider::none;

  if (buttonPos == ISlider::none)
     return;

  Boolean isVertical = slider->isVertical();
  IProgressIndicator::HomePosition homePosition = slider->homePosition();
  unsigned long offset, range, rangeMin, rangeMax;

  // Get the max offset value.
  rangeMin = trackbar.sendEvent( TBM_GETRANGEMIN,
                                 IEventParameter1( 0 ),
                                 IEventParameter2( 0 ));

  rangeMax = trackbar.sendEvent( TBM_GETRANGEMAX,
                                 IEventParameter1( 0 ),
                                 IEventParameter2( 0 ));

  range = rangeMax - rangeMin;

  // Calculate the arm offset from the left/top.
  if (homePosition == IProgressIndicator::homeBottomLeft)
     offset = armRangeOffset();
  else
     offset = range - armRangeOffset();

  // Determine the button states.  Don't check for hit selected here; we do that
  // in our handler, but disable the timer if needed.
  IGBitmap *prevLeftDown = (IGBitmap*)(buttons.firstGraphic()),
           *prevRightUp = (IGBitmap*)(buttons.lastGraphic());
  if (offset == 0)
  {
     if (prevLeftDown == leftDownPressed)
        cancelTimer();
     buttons.removeFirst().addAsFirst( *leftDownDisabled );
  }
  else if (prevLeftDown != leftDownPressed)
     buttons.removeFirst().addAsFirst( *leftDown );

  if (offset == range)
  {
     if (prevRightUp == rightUpPressed)
        cancelTimer();
     buttons.removeLast().addAsLast( *rightUpDisabled );
  }
  else if (prevRightUp != rightUpPressed)
     buttons.removeLast().addAsLast( *rightUp );

  if (slider->isVisible())
  {
     IGraphicContext gc( slider->handle() );
     if (prevLeftDown != (IGBitmap*)(buttons.firstGraphic()))
        ((IGBitmap*)(buttons.firstGraphic()))->drawOn( gc );

     if (prevRightUp != (IGBitmap*)(buttons.lastGraphic()))
        ((IGBitmap*)(buttons.lastGraphic()))->drawOn( gc );
  }
  return;
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::buttonPressed                                        |
|                                                                              |
| Handle a button press.                                                       |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::buttonPressed( IGBitmap *button )
{
  IMODTRACE_ALL("IProgressIndicatorData::buttonPressed");

  IGraphicContext gc( (IWindowHandle)(IPARENTOF( trackbar )));

  // Determine which button was pressed and process it.
  if (button == (IGBitmap*)(buttons.firstGraphic()))
  {
     // The leftDown button was pressed.  Set the new button state.
     buttons.removeFirst().addAsFirst( *leftDownPressed );
     leftDownPressed->drawOn( gc );
  }
  else if (button == (IGBitmap*)(buttons.lastGraphic()))
  {
     // The rightUp button was pressed.  Set the new button state.
     buttons.removeLast().addAsLast( *rightUpPressed );
     rightUpPressed->drawOn( gc );
  }

  // Move the slider.
  moveArm();

  // Start the timer.
  if (!timer)
     startTimer();

  return;
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::buttonReleased                                       |
|                                                                              |
| Handle a button release.                                                     |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::buttonReleased( )
{
  IMODTRACE_ALL("IProgressIndicatorData::buttonReleased");

  // Cancel any outstanding timer.
  if (timer)
     cancelTimer();

  // Restore the pressed button to normal state.
  IGraphicContext gc( (IWindowHandle)(IPARENTOF( trackbar )));
  if ((IGBitmap*)(buttons.firstGraphic()) == leftDownPressed)
  {
     buttons.removeFirst().addAsFirst( *leftDown );
     leftDown->drawOn( gc );
  }
  else if ((IGBitmap*)(buttons.lastGraphic()) == rightUpPressed)
  {
     buttons.removeLast().addAsLast( *rightUp );
     rightUp->drawOn( gc );
  }
  return;
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::startTimer                                           |
|                                                                              |
| Start the timer for button repeat.                                           |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::startTimer( )
{
  IMODTRACE_ALL("IProgressIndicatorData::startTimer");

  if (!timer)
  {
     Boolean mouseCaptured( false );
     ISETCAPTURE( mouseCaptured, IPARENTOF( trackbar ));
     if (mouseCaptured)
     {
        timer = new ITimer( new ITimerMemberFn0<IProgressIndicatorData>(
                                   *this, IProgressIndicatorData::moveArm ),
                            500);
        // The following setFocus is a workaround to prevent the frame
        // from losing focus when the IObjectWindow used by the ITimer
        // object is created.
        SetFocus( trackbar );
     }
  }
  return;
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::cancelTimer                                          |
|                                                                              |
| Kill the button repeat timer.                                                |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::cancelTimer( )
{
  IMODTRACE_ALL("IProgressIndicatorData::cancelTimer");

  if (timer)
  {
     timer->stop();
     delete timer;
     timer = 0;
     IRELEASECAPTURE;
  }
  return;
}


/*------------------------------------------------------------------------------
| IProgressIndicatorData::moveArm                                              |
|                                                                              |
| Move the slider in response to button pressed or timer elapse.               |
|                                                                              |
------------------------------------------------------------------------------*/
void IProgressIndicatorData::moveArm( )
{
  IMODTRACE_ALL("IProgressIndicatorData::moveArm");

  ISlider *slider =
        (ISlider*)IWindow::windowWithHandle( IPARENTOF( trackbar ));

  // Get the home position so we know which way to move the slider.
  IProgressIndicator::HomePosition homePosition =
        slider->homePosition();

  // Get the current slider position.
  unsigned long armOffset = armRangeOffset();

  // Determine the amount to move the slider arm.
  unsigned long ulPageSize = trackbar.sendEvent( TBM_GETPAGESIZE,
                                                 IEventParameter1( 0 ),
                                                 IEventParameter2( 0 ));

  // Determine which button is pressed and move the slider arm.
  if ((IGBitmap*)(buttons.firstGraphic()) == leftDownPressed)
  {
     if (homePosition == IProgressIndicator::homeBottomLeft)
        setArmRangeOffset( armOffset - ulPageSize );
     else
        setArmRangeOffset( armOffset + ulPageSize );
  }
  else
  {
     if (homePosition == IProgressIndicator::homeBottomLeft)
        setArmRangeOffset( armOffset + ulPageSize );
     else
        setArmRangeOffset( armOffset - ulPageSize );
  }

  return;
}


/*------------------------------------------------------------------------------
| INativeSliderHandler::INativeSliderHandler                                   |
|                                                                              |
| A default handler is needed when using the Windows native control.  Since    |
| the Windows trackbar does not support arrow buttons, a default handler is    |
| used for positioning/sizing the trackbar and drawing its arrow bitmaps.      |
------------------------------------------------------------------------------*/
INativeSliderHandler::INativeSliderHandler( )
  : INativeSliderHandler::Inherited( )
{ }

/*------------------------------------------------------------------------------
| INativeSliderHandler::~INativeSliderHandler                            |
|                                                                              |
------------------------------------------------------------------------------*/
INativeSliderHandler::~INativeSliderHandler( )
{ }


/*------------------------------------------------------------------------------
| INativeSliderHandler::dispatchHandlerEvent                                   |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean INativeSliderHandler::dispatchHandlerEvent( IEvent& event )
{
  IMODTRACE_ALL("INativeSliderHandler::dispatchHandlerEvent");

  Boolean fStopHandling( false );

  IWindowHandle hwndControl(event.controlHandle());

  if ( hwndControl )
  {
     IWindowClassName className( hwndControl );

     if (className == WC_NATIVESLIDER)
     {
        switch (event.eventId())
        {
          case WM_HSCROLL:
          case WM_VSCROLL:
          case IC_UM_VSCROLL:
          {
            IMODTRACE_DEVELOP("INativeSliderHandler::dispatchHandlerEvent(WM_SCROLL)");

            IProgressIndicator *progressInd =
                  (IProgressIndicator*)event.controlWindow();
            if (progressInd)
            {
               IProgressIndicatorData *ppd = progressInd->fProgressIndicatorData;

               if (~ISTYLEOF( ppd->trackbar ) & TBS_NOTHUMB)
               {
                  if (progressInd->isRibbonStripEnabled())
                     // This event is for a slider with a ribbon strip, so
                     // the selection range may need to be updated.
                     ppd->drawRibbon( ppd->armRangeOffset() );
                  if (((ISlider*)progressInd)->buttonsPosition() != ISlider::none)
                     // This slider has buttons, so check the button state.
                     ppd->setButtonState();
               }

               // Send an SLN_ notification so handlers will work.
               unsigned long subMsg = event.parameter1().number1();
               IWindowHandle hwndParent = IPARENTOF( progressInd->handle() );
               if (subMsg == TB_THUMBTRACK)
                  hwndParent.sendEvent( WM_COMMAND,
                        IEventParameter1( (short)IIDOF( event.handle() ), SLN_SLIDERTRACK ),
                        IEventParameter2( event.handle().asUnsigned() ));
               else if (subMsg == TB_ENDTRACK)
                  hwndParent.sendEvent( WM_COMMAND,
                        IEventParameter1( (short)IIDOF( event.handle() ), SLN_CHANGE ),
                        IEventParameter2( event.handle().asUnsigned() ));
            }
            fStopHandling = true;
            break;
          }

          case WM_SETFOCUS:
          {
            IMODTRACE_DEVELOP(
                  "INativeSliderHandler::dispatchHandlerEvent(WM_SETFOCUS)");

            // If this is a progress indicator, prevent it from getting the input
            // focus.
            IProgressIndicator *progressInd =
                  (IProgressIndicator*)event.controlWindow();
            if (progressInd)
            {
               IProgressIndicatorData *ppd = progressInd->fProgressIndicatorData;
               unsigned long ulStyle = ISTYLEOF( ppd->trackbar );
               if (~ulStyle & TBS_NOTHUMB)
                  SetFocus( ppd->trackbar );
            }
          }
          break;

          case IC_UM_CANVAS_SETFOCUS:
          {
            IMODTRACE_DEVELOP(
                  "INativeSliderHandler::dispatchHandlerEvent(IC_UM_CANVAS_SETFOCUS)");

            IProgressIndicator *progressInd =
                  (IProgressIndicator*)event.controlWindow();
            if (progressInd)
            {
               IProgressIndicatorData *ppd = progressInd->fProgressIndicatorData;
               SetFocus( ppd->trackbar );
            }
          }
          break;

          case WM_WINDOWPOSCHANGED:
          {
            IMODTRACE_DEVELOP(
                  "INativeSliderHandler::dispatchHandlerEvent(WM_WINDOWPOSCHANGED)");

            // Move or size event.
            PWINDOWPOS pswp = (PWINDOWPOS)(void*)event.parameter2();
            if (!(pswp->flags & SWP_NOSIZE))
            {
               // Size event found.  The native trackbar needs to be positioned
               // within the static control.
               IProgressIndicator *progressInd =
                     (IProgressIndicator*)event.controlWindow();
               if (progressInd)
                  progressInd->fProgressIndicatorData->layoutNativeSlider();
            }
            break;
          }

          case WM_PAINT:
          {
            IMODTRACE_DEVELOP("INativeSliderHandler::dispatchHandlerEvent(WM_PAINT)");

            IProgressIndicator *progressInd =
                  (IProgressIndicator*)event.controlWindow();

            if (progressInd)
            {
               IProgressIndicatorData *ppd = progressInd->fProgressIndicatorData;
               PAINTSTRUCT rectlInvalid;
               IPresSpaceHandle hps = IBEGINPAINT( progressInd->handle(),
                                                   0,
                                                   &rectlInvalid );
               {
                 IGraphicContext gc( progressInd->handle() );

                 // Color the static control background with the background color.
                 IColor      penColor( progressInd->foregroundColor() ),
                             backgroundColor( progressInd->backgroundColor() );
                 IRectangle rectPaint( rectlInvalid.rcPaint );
                 IGRectangle colorRect( rectPaint );
                 IGRegion clipRegion( gc, rectPaint );
                 gc.setDrawOperation( IGraphicBundle::fillAndFrame );

                 // Clip out all children.
                 IWindow::ChildCursor cursor( *progressInd );
                 for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
                 {
                    IWindowHandle hwndChild = progressInd->childAt( cursor );
                    if (hwndChild)
                    {
                       IWindow *pChild = IWindow::windowWithHandle( hwndChild );
                       if (pChild && pChild->isVisible())
                       {
                          IRectangle visibleRect = pChild->visibleRectangle();
                          if (ICoordinateSystem::isConversionNeeded())
                          {
                             visibleRect =
                                   ICoordinateSystem::convertToNative(
                                         visibleRect,
                                         progressInd->size() );
                          }
                          IGRegion childRegion( gc, visibleRect );
                          clipRegion -= childRegion;
                       }
                    }
                 }

                 gc.setClipRegion( clipRegion );
                 gc.setPenColor( backgroundColor );
                 gc.setFillColor( backgroundColor );
                 gc.draw( colorRect );

                 gc.setPenColor( penColor );
                 if (ppd->textVisible)
                    ppd->tickText[progressInd->primaryScale()].drawOn( gc );

                 ppd->buttons.drawOn( gc );

                 gc.clearClipRegion();
               }  /* brackets to ensure ~IGraphicContext called before IENDPAINT */
               IENDPAINT( progressInd->handle(), hps, &rectlInvalid );

               event.setResult( true );
               fStopHandling = true;
            }
            break;
          }

          case WM_LBUTTONDOWN:
          case WM_LBUTTONDBLCLK:
          {
            IMODTRACE_DEVELOP("INativeSliderHandler::dispatchHandlerEvent(WM_LBUTTONDOWN)");

            IProgressIndicator *progressInd =
                  (IProgressIndicator*)event.controlWindow();
            if (progressInd)
            {
               IProgressIndicatorData *ppd = progressInd->fProgressIndicatorData;

               if (GetFocus() != ppd->trackbar)
                  SetFocus( ppd->trackbar );

               if ((ISTYLEOF( ppd->trackbar ) & TBS_NOTHUMB) ||
                   ((ISlider*)progressInd)->buttonsPosition() == ISlider::none)
                  // Either this is a progress indicator, or a slider with no buttons.
                  break;

               IGraphicContext  gc( progressInd->handle() );

               // Set a reasonable aperture size for button hit testing.
               gc.setHitApertureSize( ISize( 5, 5 ) );

               // Check for button pressed.
               IPoint pointer( event.parameter2().number1(),
                               event.parameter2().number2() );
               IGraphic *button = ppd->buttons.topGraphicUnderPoint( pointer, gc );
               if (button)
               {
                  ppd->buttonPressed( (IGBitmap*)button );
                  event.setResult( true );
               }
            }
            break;
          }

          case WM_LBUTTONUP:
          {
            IMODTRACE_DEVELOP("INativeSliderHandler::dispatchHandlerEvent(WM_LBUTTONUP)");

            IProgressIndicator *progressInd =
                  (IProgressIndicator*)event.controlWindow();
            if (progressInd)
            {
               IProgressIndicatorData *ppd = progressInd->fProgressIndicatorData;
               if ((ISTYLEOF( ppd->trackbar ) & TBS_NOTHUMB) ||
                   ((ISlider*)progressInd)->buttonsPosition() == ISlider::none)
                  // Either this is a progress indicator, or a slider with no buttons.
                  break;

               // Cancel any outstanding timer and reset the button state.
               if (ppd->timer)
               {
                  ppd->buttonReleased();
                  event.setResult( true );
               }
            }
            break;
          }

          case TBM_CLEARSEL:          // Clear selection range.
          case TBM_CLEARTICS:         // Remove tick marks.
          case TBM_GETCHANNELRECT:    // Get channel bounding rectangle.
          case TBM_GETLINESIZE:       // Get Up/Down/Left/Right key increment.
          case TBM_GETNUMTICS:        // Get number of tick marks.
          case TBM_GETPAGESIZE:       // Get PgUp/PgDown key increment.
          case TBM_GETPOS:            // Get current position of slider.
          case TBM_GETPTICS:          // Get tick array.
          case TBM_GETRANGEMIN:       // Get the current range of
          case TBM_GETRANGEMAX:       //    the slider.
          case TBM_GETSELSTART:       // Get the current selection range
          case TBM_GETSELEND:         //    of the slider.
          case TBM_GETTHUMBLENGTH:    // Get the slider length.
          case TBM_GETTHUMBRECT:      // Get the slider bounding rectangle.
          case TBM_GETTIC:            // Get tick mark position.
          case TBM_GETTICPOS:         // Get tick mark position in client coord.
          case TBM_SETLINESIZE:       // Set Up/Down/Left/Right key increment.
          case TBM_SETPAGESIZE:       // Set PgUp/PgDown key increment.
          case TBM_SETPOS:            // Set current position of slider.
          case TBM_SETRANGE:          // Set the min/max positions for slider.
          case TBM_SETRANGEMIN:       // Set the min position for slider.
          case TBM_SETRANGEMAX:       // Set the max position for slider.
          case TBM_SETSEL:            // Set the selection range in trackbar.
          case TBM_SETSELSTART:       // Set the start pos. of the selection range.
          case TBM_SETSELEND:         // Set the end position of the selection range.
          case TBM_SETTHUMBLENGTH:    // Set the slider length.
          case TBM_SETTIC:            // Set tick mark position.
          case TBM_SETTICFREQ:        // Set the tick frequency.
          {
            // Forward all trackbar messages to the native trackbar window.
            // This handles the case where a message is sent directly to the control
            // window whose handle is returned from handle() function for the
            // IProgressIndicator window.
            IProgressIndicator *progressInd =
                  (IProgressIndicator*)event.controlWindow();
            if (progressInd)
            {
               IProgressIndicatorData *ppd = progressInd->fProgressIndicatorData;
               IEventResult result =
                     ppd->trackbar.sendEvent( event.eventId(),
                                              event.parameter1(),
                                              event.parameter2() );
               event.setResult( result );
               fStopHandling = result;
            }
            break;
          }
        }
     }
     else if (className == WC_SLIDER)
     {
       if (event.eventId() == WM_CTLCOLORSLIDER)
       {
         // Instead of storing the brush in the controls private data
         // class, the brush is actually being stored in the underlying
         // controls window words.  This approach was taken so that
         // progressindicatordata object is not required for a ccl control.
         HBRUSH bkgrndbrush((HBRUSH)GetWindowLong(hwndControl, QWL_SLDBRUSH));

         IProgressIndicator *progressInd((IProgressIndicator*)event.controlWindow());

         IColor clrFgnd(progressInd->foregroundColor());
         IColor clrBgnd(progressInd->backgroundColor());

         // Set foreground and background text colors in device context
         SetBkColor( (HDC)event.parameter1(), clrBgnd.asRGBLong() );
         SetTextColor( (HDC)event.parameter1(), clrFgnd.asRGBLong() );

         // Return HBRUSH for painting controls background
         if (bkgrndbrush != 0)
           DeleteObject(bkgrndbrush);

         bkgrndbrush = CreateSolidBrush (NEARESTPALETTECOLOR(clrBgnd.asRGBLong()));
         SetWindowLong( hwndControl, QWL_SLDBRUSH, (long)bkgrndbrush );

         event.setResult( bkgrndbrush );

         fStopHandling = true;
       }
     }
  }

  return fStopHandling;
}


/*------------------------------------------------------------------------------
| isliderTrackbarWinProc                                                       |
| Window procedure for the trackbar part of a native slider.  This procedure   |
| intercepts the focus notifications for the trackbar and sends them to        |
| the combobox owner.  Other messages are sent on to the original procedure    |
| for the slider's trackbar.                                                   |
------------------------------------------------------------------------------*/
long  __stdcall isliderTrackbarWinProc( void*         hwnd,
                                        unsigned int  msg,
                                        unsigned int  mp1,
                                        long          mp2)
{
   IWindowHandle hwndSlider = IPARENTOF( hwnd );
   IWindowHandle hwndParent = IPARENTOF( hwndSlider );
   IProgressIndicator *progInd =
         (IProgressIndicator*)IWindow::windowWithHandle( hwndSlider, false );
   unsigned long ulStyle = ISTYLEOF( hwnd );
   if (progInd)
   {
      switch( msg )
      {
       case WM_CHAR:
         if (mp1 == VK_TAB)
         {
            // Send these to the combobox owner or parent.  We need
            // VK_TAB to handle keyboard navigation.  We need to pass on
            // VK_ESC and VK_ENTER for correct dialog behavior.
            hwndParent.sendEvent((unsigned long)msg,
                                 (unsigned long)mp1,
                                 (unsigned long)mp2);
         }
         break;

       case WM_SETFOCUS:
         if (ulStyle & TBS_NOTHUMB)
         {
            // This is a progress indicator.  We want to prevent it from
            // receiving input focus.
            SetFocus( (HWND)mp1 );
            return true;
         }
         else
         {
            // This is a slider so send a focus notification to the slider's
            // parent.
            hwndParent.sendEvent(
                  WM_COMMAND,
                  IEventParameter1( (short)IIDOF( hwndSlider ), SLN_SETFOCUS ),
                  IEventParameter2( hwndSlider.asUnsigned() ));
         }
         break;

       case WM_KILLFOCUS:
         hwndParent.sendEvent(
               WM_COMMAND,
               IEventParameter1( (short)IIDOF( hwndSlider ), SLN_KILLFOCUS ),
               IEventParameter2( hwndSlider.asUnsigned() ));
         break;

       default:
         break;
      }

      return CallWindowProc(
                (FARPROC)progInd->fProgressIndicatorData->fdefaultTrackbarProc,
                hwnd, msg, mp1, mp2 );
   }
   else
   {
      // We are basically busted if we can't find our parent IWindow.  We do
      // what is probably the most reasonable thing and call the
      // default procedure.
      return IDEFWINDOWPROC( hwnd, msg, mp1, mp2 );
   }
}


